Package: obr
Title: Access 'Office for Budget Responsibility' Data
Version: 0.6.2
Authors@R: person("Charles", "Coverdale", email = "charlesfcoverdale@gmail.com", role = c("aut", "cre"))
Description: Provides clean, tidy access to data published by the 'Office for
    Budget Responsibility' ('OBR'), the UK's independent fiscal watchdog. Covers
    the Public Finances Databank (outturn for PSNB, PSND, receipts, and
    expenditure since 1946), the Historical Official Forecasts Database (every
    'OBR' forecast since 2010), the Economic and Fiscal Outlook detailed forecast
    tables (five-year projections from the latest Budget), the monthly
    profiles for the public finances (the 'OBR' forecast apportioned across
    the months of the fiscal year), and the Welfare Trends Report
    (incapacity benefit spending and caseloads). All returned
    objects carry provenance metadata recording the source URL, publication
    vintage, retrieval time, and file fingerprint, so analyses can be audited
    and reproduced. Data is downloaded from the 'OBR' on first use and cached
    locally for subsequent calls. Data is sourced from the 'OBR' website
    <https://obr.uk>.
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-GB
URL: https://charlescoverdale.github.io/obr/,
        https://github.com/charlescoverdale/obr
BugReports: https://github.com/charlescoverdale/obr/issues
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: cli, httr2, readxl, tools
Suggests: testthat (>= 3.0.0), knitr, rmarkdown
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-08-23 18:22:50 UTC; charlescoverdale
Author: Charles Coverdale [aut, cre]
Maintainer: Charles Coverdale <charlesfcoverdale@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-24 07:40:02 UTC
Built: R 4.5.2; ; 2026-08-24 08:14:25 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",obr_tbl)
S3method(as.data.frame,obr_tbl)
S3method(print,obr_tbl)
S3method(summary,obr_tbl)
export(clear_cache)
export(get_efo_economy)
export(get_efo_fiscal)
export(get_efo_table)
export(get_expenditure)
export(get_forecast_revisions)
export(get_forecasts)
export(get_incapacity_caseloads)
export(get_incapacity_spending)
export(get_monthly_profiles)
export(get_pension_projections)
export(get_policy_measures)
export(get_psnb)
export(get_psnd)
export(get_public_finances)
export(get_receipts)
export(get_welfare_spending)
export(list_efo_economy_measures)
export(list_forecast_series)
export(obr_actual_vs_forecast)
export(obr_as_of)
export(obr_compare_vintages)
export(obr_efo_catalogue)
export(obr_efo_vintages)
export(obr_fiscal_rules)
export(obr_forecast_panel)
export(obr_headroom)
export(obr_pin)
export(obr_pinned)
export(obr_provenance)
export(obr_unpin)
export(policy_measures_summary)
importFrom(stats,aggregate)
importFrom(utils,head)
importFrom(utils,packageVersion)

# obr

[![CRAN status](https://www.r-pkg.org/badges/version/obr)](https://CRAN.R-project.org/package=obr) [![CRAN downloads](https://cranlogs.r-pkg.org/badges/obr)](https://cran.r-project.org/package=obr) [![Total Downloads](https://cranlogs.r-pkg.org/badges/grand-total/obr)](https://CRAN.R-project.org/package=obr) [![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable) [![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)

An R package for accessing data published by the [Office for Budget Responsibility](https://obr.uk/) (OBR).

## What is the OBR?

The Office for Budget Responsibility is the UK's independent fiscal watchdog. It was created in 2010 by the coalition government to provide an independent check on the government's fiscal plans - a role previously held by HM Treasury itself.

The distinction matters. HM Treasury is the government department that *sets* fiscal policy: it decides tax rates, spending plans, and how much the government intends to borrow. The OBR's job is to *scrutinise* those plans independently, producing its own economic and fiscal forecasts that are not influenced by ministers. Think of it as the equivalent of the Bank of England's independence for monetary policy, but applied to public finances.

In practice, this means the OBR publishes forecasts at each Budget and Autumn Statement showing whether it thinks the government is on track to meet its own fiscal rules - and it has no political incentive to be optimistic.

---

## Installation

```r
install.packages("obr")

# Or install the development version from GitHub
# install.packages("devtools")
devtools::install_github("charlescoverdale/obr")
```

---

## Key OBR datasets

| Dataset | What it contains | Frequency |
|---|---|---|
| [Public Finances Databank](https://obr.uk/public-finances-databank-2019-20/) | Outturn data on PSNB, PSND, receipts, and expenditure back to 1946-47 | Monthly |
| [Historical Official Forecasts Database](https://obr.uk/data/) | Every forecast the OBR (and pre-OBR Treasury) has published for key fiscal and economic variables since 1970 | Each fiscal event |
| [Economic and Fiscal Outlook](https://obr.uk/efo/economic-and-fiscal-outlook-march-2026/) | The flagship publication at each Budget - detailed projections across 5 years | Each Budget / Autumn Statement |
| [Fiscal Sustainability Report](https://obr.uk/frs/fiscal-risks-and-sustainability-july-2025/) | Long-run projections over 50 years, covering ageing, health, and debt dynamics | Annual |
| [Welfare Trends Report](https://obr.uk/wtr/welfare-trends-report-october-2024/) | Spending trends across the benefits system | Annual |

This package covers all five datasets listed above.

---

## Why does this package exist?

All OBR data is freely available at [obr.uk](https://obr.uk). The problem is *how* it is available: as Excel files with non-standard layouts, inconsistent headers, and footnote-laden sheets that require significant wrangling before they are usable in R.

For example, the Public Finances Databank has column headers buried in row 4 of the spreadsheet, data starting in row 8, and trailing footnote numbers appended to column names. The Historical Forecasts Database stores forecasts as a vintage matrix - rows are fiscal events, columns are fiscal years - which needs reshaping into a long format before it can be plotted or analysed.

This package handles all of that automatically. One function call returns a clean, tidy data frame. Data is cached locally so subsequent calls are instant.

```r
# Without this package
path <- "~/Downloads/Public_finances_databank_March_2025.xlsx"
raw  <- readxl::read_excel(path, sheet = "Aggregates (£bn)", col_names = FALSE)
series_names <- as.character(unlist(raw[4, ]))
# ... 30 more lines of wrangling ...

# With this package
library(obr)
get_psnb()
```

---

## Functions

### Public Finances Databank

| Function | Returns |
|---|---|
| `get_psnb()` | Annual Public Sector Net Borrowing in £bn |
| `get_psnd()` | Annual Public Sector Net Debt in £bn |
| `get_expenditure()` | Annual Total Managed Expenditure in £bn |
| `get_receipts()` | Tax receipts broken down by type, in £bn |
| `get_public_finances()` | All aggregate series in tidy long format |

### Historical Forecasts Database

| Function | Returns |
|---|---|
| `list_forecast_series()` | Data frame of available series (no download needed) |
| `get_forecasts(series)` | Every OBR forecast for a given series, in tidy long format |

### Economic and Fiscal Outlook (EFO)

| Function | Returns |
|---|---|
| `get_efo_fiscal()` | Five-year net borrowing component projections from the latest Budget (£bn, annual) |
| `get_efo_economy(measure)` | Quarterly economic projections: `"inflation"`, `"labour"`, or `"output_gap"` |
| `list_efo_economy_measures()` | Available economy measures (no download needed) |

### Welfare Trends Report (WTR)

| Function | Returns |
|---|---|
| `get_welfare_spending()` | Working-age welfare spending split by incapacity/non-incapacity (% GDP, from 1978-79) |
| `get_incapacity_spending()` | Incapacity benefit spending by benefit type (ESA, IB, etc.) as % GDP |
| `get_incapacity_caseloads()` | Combined incapacity caseloads and prevalence since 2008-09 |

### Fiscal Risks and Sustainability Report (FSR)

| Function | Returns |
|---|---|
| `get_pension_projections()` | 50-year state pension spending projections (% GDP) under demographic and triple-lock scenarios |

### Cache management

| Function | What it does |
|---|---|
| `clear_cache()` | Deletes all locally cached OBR files |

All download functions accept `refresh = TRUE` to force a fresh download from the OBR website.

---

## Examples

### 1. How much did COVID cost the UK?

```r
library(obr)

psnb <- get_psnb()
psnb[psnb$year %in% c("2018-19", "2019-20", "2020-21", "2021-22", "2022-23"), ]
#>       year  psnb_bn
#>  2018-19     42.5
#>  2019-20     57.1
#>  2020-21    317.8   # ← COVID year
#>  2021-22    144.8
#>  2022-23     87.6
```

The UK borrowed £318bn in 2020-21 - roughly seven times the pre-pandemic level - to fund furlough, bounce-back loans, and emergency NHS spending.

---

### 2. How has the OBR's borrowing forecast changed over time?

The OBR first forecast 2024-25 borrowing at **£37bn** (March 2022). By November 2025, that estimate had risen to **£149bn** - four times the original figure.

```r
psnb_fc <- get_forecasts("PSNB")
fc_2425 <- psnb_fc[psnb_fc$fiscal_year == "2024-25", c("forecast_date", "value")]
fc_2425
#>    forecast_date   value
#>     March 2022    36.5
#>  November 2022    84.3
#>     March 2023    85.4
#>  November 2023    84.6
#>     March 2024    87.2
#>   October 2024   127.5
#>     March 2025   137.3
#>  November 2025   149.5
```

The `get_forecasts()` function returns every published forecast across all fiscal events, making it straightforward to visualise forecast drift and assess how fiscal plans have evolved.

---

### 3. Has the UK ever run a surplus - and how likely is it to again?

```r
psnb <- get_psnb()

# Years with a surplus (negative PSNB = receipts exceed spending)
psnb[psnb$psnb_bn < 0, ]
#>       year  psnb_bn
#>  1969-70    -0.5
#>  1970-71    -1.3
#>  1971-72    -0.1
#>  1988-89    -9.0
#>  1989-90    -8.0
#>  1990-91    -0.1
#>  1997-98   -12.7
#>  1998-99   -14.5
#>  1999-00   -17.9
#>  2000-01    -0.5
```

The UK last ran a surplus in 2000-01. In the 24 years since, the government has borrowed every year. Combine with `get_forecasts("PSNB_pct")` to see whether the OBR projects any future surpluses.

---

### 4. Where does government revenue come from?

```r
receipts <- get_receipts()

# Top tax sources in 2023-24
r <- receipts[receipts$year == "2023-24", ]
r <- r[order(-r$value), ]
head(r[, c("series", "value")], 8)
#>                              series   value
#>   Public sector current receipts   1101.5
#>                       Income tax    290.4
#>                              VAT    183.1
#>    National insurance contributions 182.4
#>                    Corporation tax   88.4
#>                       Council tax   44.9
#>                         Fuel duty   24.5
#>                    Stamp duties     18.4
```

Income tax, VAT, and National Insurance together account for around 60% of all government receipts. Breaking this down over time reveals long-run shifts - such as the rising share of income tax as fiscal drag pulls more earners into higher bands.

---

### 5. What does the March 2026 Budget forecast for borrowing?

```r
efo <- get_efo_fiscal()
efo[efo$series == "Net borrowing", ]
#>   fiscal_year        series value_bn
#>       2025-26 Net borrowing    132.7
#>       2026-27 Net borrowing    115.5
#>       2027-28 Net borrowing     96.5
#>       2028-29 Net borrowing     86.0
#>       2029-30 Net borrowing     63.4
#>       2030-31 Net borrowing     59.0
```

The EFO detailed tables also include the full breakdown: current receipts, current expenditure, depreciation, net investment, and net borrowing - enabling you to see exactly how the deficit is projected to narrow.

---

### 6. Is the UK's incapacity benefits bill rising?

```r
welfare <- get_welfare_spending()
# Working-age incapacity spending, last 10 years
ic <- welfare[welfare$series == "Working-age incapacity benefits spending" &
              welfare$year >= "2014-15", ]
ic
#>        year                                      series value
#>     2014-15 Working-age incapacity benefits spending  1.44
#>     2015-16 Working-age incapacity benefits spending  1.33
#>     ...
#>     2023-24 Working-age incapacity benefits spending  1.78
#>     2024-25 Working-age incapacity benefits spending  2.02
#>     2025-26 Working-age incapacity benefits spending  2.16

# Number of people on incapacity benefits
cases <- get_incapacity_caseloads()
cases[cases$series == "Share of working age population", ]
#>  2008-09   Share of working age population  6.80
#>  ...
#>  2023-24   Share of working age population  6.82
```

Incapacity benefit spending and caseloads have risen sharply since the pandemic - a key driver of welfare reform debate in 2025.

---

### 7. What happens to the state pension bill as the UK ages?

```r
proj <- get_pension_projections()

# Central demographic projection: pension spending rises from 5% to 7.7% of GDP
central <- proj[proj$scenario_type == "Demographic scenarios" &
                proj$scenario == "Central projection", ]
head(central[, c("fiscal_year", "pct_gdp")], 5)
#>   fiscal_year pct_gdp
#>       2023-24    4.56
#>       2024-25    4.95
#>       2025-26    5.06
#>       2026-27    5.13
#>       2027-28    5.05

tail(central[, c("fiscal_year", "pct_gdp")], 5)
#>   fiscal_year pct_gdp
#>       2069-70    7.73
#>       2070-71    7.82
#>       2071-72    7.77
#>       2072-73    7.66
#>       2073-74    7.65
```

The OBR's central projection has the state pension rising from 4.6% of GDP today to 7.7% by 2073-74 as the UK population ages. The FSR also publishes scenarios for higher/lower life expectancy and different triple-lock uprating assumptions.

---

## Related packages

This package is part of a family of R packages for UK economic and fiscal data. They share a consistent interface - named functions, tidy data frames, local caching - and are designed to work together. All outputs join cleanly on date columns.

| Package | What it covers |
|---|---|
| [`ons`](https://github.com/charlescoverdale/ons) | ONS data (GDP, inflation, unemployment, wages, trade, house prices, population) |
| [`boe`](https://github.com/charlescoverdale/boe) | Bank of England data (Bank Rate, SONIA, gilt yields, exchange rates, mortgage rates) |
| [`hmrc`](https://github.com/charlescoverdale/hmrc) | HMRC tax receipts, corporation tax, stamp duty, R&D credits, and tax gap data |
| [`readecb`](https://github.com/charlescoverdale/readecb) | European Central Bank data (policy rates, HICP, exchange rates, yield curves) |
| [`readoecd`](https://github.com/charlescoverdale/readoecd) | OECD data (GDP, unemployment, inflation, trade across 38 member countries) |
| [`fred`](https://github.com/charlescoverdale/fred) | US Federal Reserve (FRED) data (800,000+ economic time series) |
| [`inflateR`](https://github.com/charlescoverdale/inflateR) | Adjust values for inflation using CPI or GDP deflator data |

---

## Keeping data up to date

The two core datasets - the Public Finances Databank and Historical Official Forecasts Database - are accessed via stable generic URLs that the OBR keeps pointed at the latest file. They update automatically whenever you call `refresh = TRUE`.

The EFO, WTR, and FSR functions work differently. The OBR publishes each edition at a new URL containing the publication date (e.g. `march-2026-economic-and-fiscal-outlook-...`), and does not maintain a generic redirect. This means the URLs in the package are hardcoded to a specific edition: calls to `get_efo_fiscal()` will always return the March 2026 Budget forecasts until the package is updated, even after the OBR publishes a new edition.

The OBR publishes on a roughly predictable schedule - the EFO twice a year (March and October/November), the FSR and WTR once a year each. This package will be updated to track each new edition, meaning **2–3 patch releases per year**. Check the [NEWS](https://github.com/charlescoverdale/obr/blob/main/NEWS.md) to see which edition each function currently reflects. If you need data from the absolute latest publication before a package update is available, download directly from [obr.uk](https://obr.uk).

---

## Issues

Please report bugs or requests at <https://github.com/charlescoverdale/obr/issues>.

## Keywords

Office for Budget Responsibility, OBR, UK fiscal forecasts, economic forecasts, GDP forecast, inflation forecast, public finances, government borrowing, fiscal policy, UK budget, R package

# obr

[![CRAN status](https://www.r-pkg.org/badges/version/obr)](https://CRAN.R-project.org/package=obr) [![CRAN downloads](https://cranlogs.r-pkg.org/badges/obr)](https://cran.r-project.org/package=obr) [![Total Downloads](https://cranlogs.r-pkg.org/badges/grand-total/obr)](https://CRAN.R-project.org/package=obr) [![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable) [![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)

An R package for accessing data published by the [Office for Budget Responsibility](https://obr.uk/) (OBR).

## What is the OBR?

The Office for Budget Responsibility is the UK's independent fiscal watchdog. It was created in 2010 by the coalition government to provide an independent check on the government's fiscal plans - a role previously held by HM Treasury itself.

The distinction matters. HM Treasury is the government department that *sets* fiscal policy: it decides tax rates, spending plans, and how much the government intends to borrow. The OBR's job is to *scrutinise* those plans independently, producing its own economic and fiscal forecasts that are not influenced by ministers. Think of it as the equivalent of the Bank of England's independence for monetary policy, but applied to public finances.

In practice, this means the OBR publishes forecasts at each Budget and Autumn Statement showing whether it thinks the government is on track to meet its own fiscal rules - and it has no political incentive to be optimistic.

---

## Installation

```r
install.packages("obr")

# Or install the development version from GitHub
# install.packages("devtools")
devtools::install_github("charlescoverdale/obr")
```

---

## Key OBR datasets

| Dataset | What it contains | Frequency |
|---|---|---|
| [Public Finances Databank](https://obr.uk/public-finances-databank-2019-20/) | Outturn data on PSNB, PSND, receipts, and expenditure back to 1946-47 | Monthly |
| [Historical Official Forecasts Database](https://obr.uk/data/) | Every forecast the OBR (and pre-OBR Treasury) has published for key fiscal and economic variables since 1970 | Each fiscal event |
| [Economic and Fiscal Outlook](https://obr.uk/efo/economic-and-fiscal-outlook-march-2026/) | The flagship publication at each Budget - detailed projections across 5 years | Each Budget / Autumn Statement |
| [Fiscal Sustainability Report](https://obr.uk/frs/fiscal-risks-and-sustainability-july-2025/) | Long-run projections over 50 years, covering ageing, health, and debt dynamics | Annual |
| [Welfare Trends Report](https://obr.uk/wtr/welfare-trends-report-october-2024/) | Spending trends across the benefits system | Annual |
| [Monthly profiles](https://obr.uk/monthly-public-finances-briefing/) | The EFO forecast for receipts, spending, and the CGNCR apportioned across the months of the fiscal year, against which each month's outturn is judged | Each EFO |

This package covers all six datasets listed above.

---

## Why does this package exist?

All OBR data is freely available at [obr.uk](https://obr.uk). The problem is *how* it is available: as Excel files with non-standard layouts, inconsistent headers, and footnote-laden sheets that require significant wrangling before they are usable in R.

For example, the Public Finances Databank has column headers buried in row 4 of the spreadsheet, data starting in row 8, and trailing footnote numbers appended to column names. The Historical Forecasts Database stores forecasts as a vintage matrix - rows are fiscal events, columns are fiscal years - which needs reshaping into a long format before it can be plotted or analysed.

This package handles all of that automatically. One function call returns a clean, tidy data frame. Data is cached locally so subsequent calls are instant.

```r
# Without this package
path <- "~/Downloads/Public_finances_databank_March_2025.xlsx"
raw  <- readxl::read_excel(path, sheet = "Aggregates (£bn)", col_names = FALSE)
series_names <- as.character(unlist(raw[4, ]))
# ... 30 more lines of wrangling ...

# With this package
library(obr)
get_psnb()
```

---

## Provenance and reproducibility

Every data-returning function returns an `obr_tbl`: a `data.frame` with attached metadata recording the OBR publication, the publication vintage, the source URL, when the data was retrieved, and the MD5 fingerprint of the underlying spreadsheet. The provenance prints automatically as a header above the data.

```r
library(obr)
get_efo_fiscal()
#> # obr_tbl: <rows> x <cols>
#> # Source:       OBR Economic and Fiscal Outlook, <vintage>
#> # URL:          https://obr.uk/download/<vintage-slug>/
#> # Retrieved:    <timestamp>
#> # File MD5:     <md5 prefix>
#> # Package:      obr <version>
#>
#> <data rows ...>
```

Use `obr_provenance()` to extract the metadata as a list, or `summary()` for the full provenance card. Provenance survives `[` subsetting and is stripped only on explicit `as.data.frame()`. This means an analysis pinned against an `obr_tbl` always carries the audit trail of which OBR publication produced the numbers.

```r
psnb <- get_psnb()
obr_provenance(psnb)
# Returns a list with: publication, vintage, source_url,
# retrieved (POSIXct), file_md5, package_version, notes
```

---

## Functions

### Public Finances Databank

| Function | Returns |
|---|---|
| `get_psnb()` | Annual Public Sector Net Borrowing in £bn |
| `get_psnd()` | Annual Public Sector Net Debt in £bn |
| `get_expenditure()` | Annual Total Managed Expenditure in £bn |
| `get_receipts()` | Tax receipts broken down by type, in £bn |
| `get_public_finances()` | All aggregate series in tidy long format |

### Historical Forecasts Database

| Function | Returns |
|---|---|
| `list_forecast_series()` | Data frame of available series (no download needed) |
| `get_forecasts(series)` | Every OBR forecast for a given series, in tidy long format |
| `obr_forecast_panel(series)` | Wide real-time panel: rows = forecast vintages, columns = fiscal years |
| `obr_compare_vintages(a, b, what)` | Tidy diff between two vintages of the same EFO table; returns `value_a`, `value_b`, and `revision = value_b - value_a` |
| `obr_actual_vs_forecast(series)` | Joins HFD forecasts against PFD outturn for the same series; returns `value_forecast`, `value_actual`, and `error` |
| `get_forecast_revisions(unit)` | EFO-to-EFO PSNB revisions decomposed into policy, classifications, and underlying components |

### Economic and Fiscal Outlook (EFO)

| Function | Returns |
|---|---|
| `obr_efo_catalogue()` | All 39 detailed-forecast tables this package recognises, with id, section, title, layout, and default `metric_type` / `unit`. No download needed. |
| `get_efo_table(id)` | Generic fetcher: pass a catalogue id (e.g. `"6.13"`, `"1.19"`, `"4.1"`) and get the parsed contents in the standard v0.4 schema. Covers ~35 tables across receipts, expenditure, debt, GDP, labour, inflation, housing, and household balance sheets. |
| `get_efo_fiscal()` | Net borrowing components (Table 6.5). Convenience wrapper over `get_efo_table("6.5")`. |
| `get_efo_economy(measure)` | Quarterly economic projections: `"inflation"`, `"labour"`, or `"output_gap"`. Convenience wrapper over `get_efo_table()`. |
| `list_efo_economy_measures()` | The three measures available via `get_efo_economy()` (no download needed). |
| `get_monthly_profiles()` | The EFO forecast apportioned across the twelve months of the fiscal year (receipts and spending profiles, or the CGNCR breakdown). The reference point for "borrowing so far this year vs the OBR profile" in the run-up to a fiscal event. |

### Welfare Trends Report (WTR)

| Function | Returns |
|---|---|
| `get_welfare_spending()` | Working-age welfare spending split by incapacity/non-incapacity (% GDP, from 1978-79) |
| `get_incapacity_spending()` | Incapacity benefit spending by benefit type (ESA, IB, etc.) as % GDP |
| `get_incapacity_caseloads()` | Combined incapacity caseloads and prevalence since 2008-09 |

### Fiscal Risks and Sustainability Report (FSR)

| Function | Returns |
|---|---|
| `get_pension_projections()` | **Deprecated in 0.5.1.** The OBR restructured the FSR workbooks in July 2026 and the state pension spending scenarios this returned are no longer published in that form. The series now lives in the FSR Chapter 3 workbook (Chart 3.11); see [obr.uk/frs](https://obr.uk/frs/fiscal-risks-and-sustainability-july-2026/). |

### Policy Measures Database (PMD)

| Function | Returns |
|---|---|
| `get_policy_measures()` | Every tax (since 1970) and spending (since 2010) measure scored at a UK fiscal event, with Exchequer effect by fiscal year in GBP million |
| `policy_measures_summary()` | Net Exchequer effect aggregated by fiscal event and fiscal year |

### Fiscal rules

| Function | Returns |
|---|---|
| `obr_fiscal_rules()` | The three Charter for Budget Responsibility rules (stability, investment, welfare cap) with metric, target description, and source Charter version. Numerical headroom is not shipped as a constant (changes at every fiscal event); derive it with `obr_headroom()`. |
| `obr_headroom()` | The current budget surplus path from EFO Table 6.5: the margin against the stability rule in every forecast year. Pass `target_year` to flag the year the rule bites on. |

### Cache management

| Function | What it does |
|---|---|
| `clear_cache()` | Deletes all locally cached OBR files |

All download functions accept `refresh = TRUE` to force a fresh download from the OBR website.

### Standard tidy long schema (v0.4.0)

The EFO, PFD, HFD, WTR, and FSR functions all return the same six columns so outputs from different publications can be `rbind()`'d, joined, and plotted the same way:

| Column | Type | Values |
|---|---|---|
| `period` | character | `"2024-25"`, `"2025Q1"`, etc. |
| `period_type` | character | `"fiscal_year"`, `"quarter"`, `"calendar_year"` |
| `series` | character | The variable name (e.g. `"CPI"`, `"PSNB"`, `"Net borrowing"`) |
| `metric_type` | character | `"level"`, `"yoy_pct"`, `"index"`, `"pct"`, `"pct_pts"` |
| `value` | double | The numeric value |
| `unit` | character | `"gbp_bn"`, `"pct"`, `"index"`, `"count_k"`, etc. |

`get_forecasts()` adds `forecast_date` as a leading column. `get_forecast_revisions()` and `get_policy_measures()` use specialised multi-dimensional schemas (migration to the standard layout queued for a later release).

The `metric_type` column resolves a class of bug from earlier versions where, for example, CPI Index values and CPI year-on-year growth values lived in the same `value` column with no machine-readable distinction. See `vignette("efo-forecasts")` for a walkthrough.

### Vignettes

- `vignette("efo-forecasts")`: working with the v0.4.0 schema, splitting Index from YoY, combining EFO with PFD outturn, comparing two vintages.
- `vignette("vintages")`: pinning to a specific EFO vintage for reproducible analysis.
- `vignette("forecast-revisions")`: decomposing PSNB revisions across vintages.
- `vignette("policy-measures")`: working with the Policy Measures Database.

### Provenance and vintage control

| Function | What it does |
|---|---|
| `obr_provenance(x)` | Extracts the source URL, vintage, retrieval time, and file fingerprint attached to any returned `obr_tbl` |
| `summary(x)` | Prints the full provenance card alongside the structural summary |
| `obr_efo_vintages()` | Lists every EFO published since June 2010, with publication dates and URL slugs |
| `obr_as_of(date)` | Returns the EFO that was current on a given calendar date |
| `obr_pin(vintage)` | Sets a session-wide EFO vintage; `get_efo_*` then defaults to that vintage |
| `obr_unpin()` | Clears any pin set by `obr_pin()` |
| `obr_pinned()` | Returns the currently pinned vintage, or `NULL` |

---

## Examples

### 1. How much did COVID cost the UK?

```r
library(obr)

psnb <- get_psnb()
psnb[psnb$period %in% c("2018-19", "2019-20", "2020-21", "2021-22", "2022-23"),
     c("period", "value", "unit")]
#>       period   value   unit
#>     2018-19    42.5  gbp_bn
#>     2019-20    57.1  gbp_bn
#>     2020-21   317.8  gbp_bn   # ← COVID year
#>     2021-22   144.8  gbp_bn
#>     2022-23    87.6  gbp_bn
```

The UK borrowed £318bn in 2020-21 - roughly seven times the pre-pandemic level - to fund furlough, bounce-back loans, and emergency NHS spending.

---

### 2. How has the OBR's borrowing forecast changed over time?

The OBR first forecast 2024-25 borrowing at **£37bn** (March 2022). By November 2025, that estimate had risen to **£149bn** - four times the original figure.

```r
psnb_fc <- get_forecasts("PSNB")
fc_2425 <- psnb_fc[psnb_fc$period == "2024-25", c("forecast_date", "value")]
fc_2425
#>    forecast_date   value
#>     March 2022    36.5
#>  November 2022    84.3
#>     March 2023    85.4
#>  November 2023    84.6
#>     March 2024    87.2
#>   October 2024   127.5
#>     March 2025   137.3
#>  November 2025   149.5
```

The `get_forecasts()` function returns every published forecast across all fiscal events, making it straightforward to visualise forecast drift and assess how fiscal plans have evolved.

---

### 3. Has the UK ever run a surplus - and how likely is it to again?

```r
psnb <- get_psnb()

# Years with a surplus (negative PSNB = receipts exceed spending)
psnb[psnb$value < 0, c("period", "value")]
#>       period   value
#>     1969-70    -0.5
#>     1970-71    -1.3
#>     1971-72    -0.1
#>     1988-89    -9.0
#>     1989-90    -8.0
#>     1990-91    -0.1
#>     1997-98   -12.7
#>     1998-99   -14.5
#>     1999-00   -17.9
#>     2000-01    -0.5
```

The UK last ran a surplus in 2000-01. In the 24 years since, the government has borrowed every year. Combine with `get_forecasts("PSNB_pct")` to see whether the OBR projects any future surpluses.

---

### 4. Where does government revenue come from?

```r
receipts <- get_receipts()

# Top tax sources in 2023-24
r <- receipts[receipts$period == "2023-24", ]
r <- r[order(-r$value), ]
head(r[, c("series", "value", "unit")], 8)
#>                              series   value    unit
#>   Public sector current receipts   1101.5  gbp_bn
#>                       Income tax    290.4  gbp_bn
#>                              VAT    183.1  gbp_bn
#>    National insurance contributions 182.4  gbp_bn
#>                    Corporation tax   88.4  gbp_bn
#>                       Council tax   44.9  gbp_bn
#>                         Fuel duty   24.5  gbp_bn
#>                    Stamp duties     18.4  gbp_bn
```

Income tax, VAT, and National Insurance together account for around 60% of all government receipts. Breaking this down over time reveals long-run shifts - such as the rising share of income tax as fiscal drag pulls more earners into higher bands.

---

### 5. What does the March 2026 Budget forecast for borrowing?

```r
efo <- get_efo_fiscal()
efo[efo$series == "Net borrowing", c("period", "series", "value", "unit")]
#>      period        series  value    unit
#>     2025-26 Net borrowing  132.7  gbp_bn
#>     2026-27 Net borrowing  115.5  gbp_bn
#>     2027-28 Net borrowing   96.5  gbp_bn
#>     2028-29 Net borrowing   86.0  gbp_bn
#>     2029-30 Net borrowing   63.4  gbp_bn
#>     2030-31 Net borrowing   59.0  gbp_bn
```

The EFO detailed tables also include the full breakdown: current receipts, current expenditure, depreciation, net investment, and net borrowing - enabling you to see exactly how the deficit is projected to narrow.

### 5a. Reproducing an analysis as it would have looked on a past date

Because the OBR revises its forecast at every Budget and Statement, an analysis run today and the same analysis run six months from now can return materially different numbers. The vintage layer pins to a specific EFO so the analysis is reproducible.

```r
# What did the OBR forecast for 2027-28 borrowing in October 2024 vs March 2026?
oct_2024 <- get_efo_fiscal(vintage = "October 2024")
mar_2026 <- get_efo_fiscal(vintage = "March 2026")

oct_2024[oct_2024$series == "Net borrowing" & oct_2024$period == "2027-28", ]
mar_2026[mar_2026$series == "Net borrowing" & mar_2026$period == "2027-28", ]

# Or in one step, with the v0.4.0 helper:
obr_compare_vintages("October 2024", "March 2026", what = "fiscal")
#>   period period_type        series metric_type   unit value_a value_b revision
#>  2027-28 fiscal_year Net borrowing       level gbp_bn    81.7    96.5     14.8
#>  ...

# Or pin once and let every subsequent EFO call use that vintage
obr_pin("October 2024")
get_efo_fiscal()                  # uses October 2024
get_efo_economy("inflation")      # also uses October 2024
obr_unpin()

# Find which EFO was current on a given date
obr_as_of("2024-12-15")
#> [1] "October 2024"
```

---

### 6. Is the UK's incapacity benefits bill rising?

```r
welfare <- get_welfare_spending()
# Working-age incapacity spending, last 10 years
ic <- welfare[welfare$series == "Working-age incapacity benefits spending" &
              welfare$period >= "2014-15", ]
ic[, c("period", "series", "value", "unit")]
#>      period                                   series  value  unit
#>     2014-15 Working-age incapacity benefits spending   1.44   pct
#>     2015-16 Working-age incapacity benefits spending   1.33   pct
#>     ...
#>     2023-24 Working-age incapacity benefits spending   1.78   pct
#>     2024-25 Working-age incapacity benefits spending   2.02   pct
#>     2025-26 Working-age incapacity benefits spending   2.16   pct

# Number of people on incapacity benefits (in thousands)
cases <- get_incapacity_caseloads()
cases[cases$series == "Claimants", c("period", "value", "unit")]
#>     2008-09  2360  count_k
#>     ...
#>     2024-25  3140  count_k
```

Incapacity benefit spending and caseloads have risen sharply since the pandemic - a key driver of welfare reform debate in 2025.

---

### 7. Every tax measure in a Budget

```r
# All tax measures scored from 2025-26 onwards
pmd <- get_policy_measures(type = "tax", since = "2025-26")

# Filter to measures from a specific event, ordered by 2025-26 effect
oct24 <- pmd[grepl("October 2024", pmd$event) &
             pmd$fiscal_year == "2025-26", ]
oct24 <- oct24[order(-oct24$value_mn), ]
head(oct24[, c("measure", "head", "value_mn")])
```

The PMD covers every measure scored at a UK fiscal event: 1970 onwards for tax, 2010 onwards for spending. Combine `search =` and `since =` to pull a thematic time series, e.g. all alcohol-duty measures since 2010.

```r
get_policy_measures(type = "tax", search = "alcohol", since = "2010-11")
```

### 7a. What does the OBR say about the fiscal rules?

```r
obr_fiscal_rules()
```

Returns the three Charter for Budget Responsibility rules in force (stability rule, investment rule, welfare cap), with their target metric, direction of pass, and the source Charter version. Numerical headroom is not shipped as a constant because it changes at every fiscal event; derive it from `get_efo_fiscal()` or consult the EFO press release for the relevant vintage.

---

### 8. Browsing the full EFO databank

```r
# Every detailed-forecast table the package recognises
cat <- obr_efo_catalogue()
nrow(cat)  # 39 (35 currently parseable, 2 cross-references, 2 complex layouts)

# All Debt-section tables
cat[cat$section == "Debt", c("table_id", "title")]
#>   table_id                                                 title
#>      6.11  Public sector net debt year-on-year changes
#>      6.12  Total gross financing
#>      6.13  Composition of public sector net debt
#>      6.14  Composition of public sector net worth
#>      6.16  Central government debt interest by financing component
#>      6.17  Outstanding stocks, debt interest payments and effective rates

# Pull any one
get_efo_table("6.13")             # composition of net debt, % of GDP
get_efo_table("1.19")             # CPI category inflation, quarterly
get_efo_table("6.16")             # debt interest by financing component
get_efo_table("6.5", vintage = "October 2024")
```

The `get_efo_table()` dispatcher routes each id to a layout-specific parser and returns the standard v0.4 schema. The headline wrappers (`get_efo_fiscal()`, `get_efo_economy()`) remain as convenience entry points; under the hood they call the dispatcher.

---

## Limitations

A few things `obr` deliberately does not do, and a few caveats worth knowing.

- **Fiscal Risks and Sustainability coverage is thin.** The EFO detailed-forecast tables are complete: v0.5.0 closed the last gaps and `obr` now reaches all 39 in the Aggregates and Economy workbooks, discoverable through `obr_efo_catalogue()` and fetchable with `get_efo_table()`. The FSR long-run projections are the exception. `get_pension_projections()` was deprecated in v0.5.1 after the OBR restructured the FSR workbooks in July 2026, and nothing has replaced it yet.
- **EFO and PFD use the standard v0.4.0 schema; FRD and PMD do not yet.** `get_forecast_revisions()` and `get_policy_measures()` return their existing multi-dimensional schemas (forecast date x revision component, fiscal event x measure x head). Migration to the v0.4.0 layout is queued for a later release.
- **Vintage table is hardcoded and needs maintenance.** `obr_efo_vintages()` returns a static list of EFOs. When a new EFO publishes, the package needs a release to recognise it as a pinnable vintage. Until then, the dynamic resolver will still find the live file but `obr_pin("November 2026")` will error.
- **The `classify_metric_type()` heuristic is regex-based.** It correctly handles the OBR's current naming conventions (Index, deflator, inflation, growth, rate). New series with unusual names may default to `metric_type = "level"`; the per-measure default supplied by `get_efo_economy()` corrects this for the inflation sheet but may need tuning if OBR adds a new sheet.
- **UK only.** The package wraps OBR publications; it does not cover the US Congressional Budget Office, the Australian Parliamentary Budget Office, or other fiscal-watchdog equivalents.
- **No Python equivalent.** A port to Python is not planned; users wanting the same data in Python should use the OBR's published Excel files directly or contribute a port.
- **Network access is required on first use.** Cached files persist across sessions but `clear_cache()` or a refresh argument forces a re-download. There is no offline mode.

---

## Related packages

| Package | Description |
|---|---|
| [`ons`](https://github.com/charlescoverdale/ons) | UK Office for National Statistics data |
| [`hmrc`](https://github.com/charlescoverdale/hmrc) | HM Revenue & Customs tax data |
| [`boe`](https://github.com/charlescoverdale/boe) | Bank of England data |
| [`fred`](https://github.com/charlescoverdale/fred) | US Federal Reserve (FRED) data |
| [`debtkit`](https://github.com/charlescoverdale/debtkit) | Debt sustainability analysis |
| [`yieldcurves`](https://github.com/charlescoverdale/yieldcurves) | Yield curve fitting (Nelson-Siegel, Svensson) |
| [`inflateR`](https://github.com/charlescoverdale/inflateR) | Inflation adjustment for UK price series |
| [`inflationkit`](https://github.com/charlescoverdale/inflationkit) | Inflation analysis |
| [`greenbook`](https://github.com/charlescoverdale/greenbook) | HM Treasury Green Book CBA primitives |
| [`magentabook`](https://github.com/charlescoverdale/magentabook) | HM Treasury Magenta Book evaluation primitives |

---

## Keeping data up to date

The Public Finances Databank is accessed via a stable URL that the OBR keeps pointed at the latest file. The Historical Official Forecasts Database, EFO, WTR, and FSR functions all use a dynamic URL resolver that probes the OBR's recent fiscal events (most recent first) and uses whichever URL is live, so the package automatically picks up new editions without a code change.

If the resolver cannot find a live URL (for example because of a temporary network outage, or because the OBR has moved to an unfamiliar slug pattern), it falls back to the last-known-good URL and emits an explicit warning rather than failing silently. The returned `obr_tbl` always records the URL that was actually used, so any analysis carries its own audit trail.

The OBR publishes on a roughly predictable schedule: the EFO twice a year (March and October/November), the FSR each summer, the WTR annually. The package's fallback URLs are refreshed at each release; check the [NEWS](https://github.com/charlescoverdale/obr/blob/main/NEWS.md) for the current fallback edition.

---

## Issues

Please report bugs or requests at <https://github.com/charlescoverdale/obr/issues>.

## Keywords

Office for Budget Responsibility, OBR, UK fiscal forecasts, economic forecasts, GDP forecast, inflation forecast, public finances, government borrowing, fiscal policy, UK budget, R package

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# library(obr)
# 
# rev <- get_forecast_revisions(unit = "gbp_bn")
# unique(rev$component)
# # "Total"
# # "Policy"
# # "of which: receipts"
# # "of which: spending"
# # "Classifications and one-offs"
# # "Underlying"
# # "of which: receipts"
# # "of which: debt interest"
# # "of which: non-interest spending"

## ----eval = FALSE-------------------------------------------------------------
# top <- rev[rev$component %in% c(
#   "Total", "Policy",
#   "Classifications and one-offs", "Underlying"
# ), ]
# 
# # Most recent vintage's revision for each fiscal year
# latest <- max(top$forecast_date)
# top[top$forecast_date == latest, ]

## ----eval = FALSE-------------------------------------------------------------
# rev_pct <- get_forecast_revisions(unit = "pct_gdp")

## ----eval = FALSE-------------------------------------------------------------
# panel <- obr_forecast_panel("PSNB")
# # Every PSNB forecast for 2027-28
# panel[, c("forecast_date", "2027-28")]

## ----eval = FALSE-------------------------------------------------------------
# panel <- obr_forecast_panel("PSNB")
# 
# # Forecast value for FY 2024-25 across every vintage
# psnb_2425 <- panel[, c("forecast_date", "2024-25")]
# psnb_2425 <- psnb_2425[!is.na(psnb_2425[["2024-25"]]), ]
# psnb_2425

## ----eval = FALSE-------------------------------------------------------------
# prov <- obr_provenance(rev)
# sprintf("OBR Forecast Revisions Database, vintage %s. File MD5: %s. Retrieved %s.",
#         prov$vintage,
#         prov$file_md5,
#         format(prov$retrieved, "%Y-%m-%d"))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# library(obr)
# 
# inf <- get_efo_economy("inflation")
# table(inf$metric_type, inf$unit)
# #         index  pct
# #   index   372    0
# #   yoy_pct   0 1844

## ----eval = FALSE-------------------------------------------------------------
# inf_yoy <- inf[inf$metric_type == "yoy_pct", ]
# head(inf_yoy[inf_yoy$series == "CPI", ])

## ----eval = FALSE-------------------------------------------------------------
# inf_idx <- inf[inf$metric_type == "index", ]

## ----eval = FALSE-------------------------------------------------------------
# forecast <- get_efo_fiscal()                # 5-year forecast, gbp_bn
# forecast <- forecast[forecast$series == "Net borrowing", ]
# 
# outturn <- get_psnb()                       # historical outturn, series = "PSNB", gbp_bn
# outturn <- outturn[outturn$period >= "2020-21", ]
# 
# # Both have period (fiscal_year), value (gbp_bn), unit. Stack them.
# combined <- rbind(
#   data.frame(source = "outturn",  outturn[,  c("period", "value", "unit")]),
#   data.frame(source = "forecast", forecast[, c("period", "value", "unit")])
# )

## ----eval = FALSE-------------------------------------------------------------
# oct24 <- get_efo_fiscal(vintage = "October 2024")
# mar26 <- get_efo_fiscal(vintage = "March 2026")
# 
# # Net borrowing forecast for 2027-28 from each vintage
# oct24[oct24$series == "Net borrowing" & oct24$period == "2027-28", "value"]
# mar26[mar26$series == "Net borrowing" & mar26$period == "2027-28", "value"]

## ----eval = FALSE-------------------------------------------------------------
# obr_provenance(get_efo_fiscal())
# # $publication: "EFO"
# # $vintage:     "March 2026"
# # $source_url:  ...
# # $retrieved:   timestamp
# # $file_md5:    fingerprint of the underlying spreadsheet
# # $package_version: obr version

# UK fiscal rules per the Charter for Budget Responsibility (Autumn 2024,
# updated Autumn 2025) and the Budget Responsibility Act 2024. The rules
# themselves are encoded statically; numerical headroom is intentionally
# not hardcoded because the value changes at every fiscal event and any
# value shipped in the package would go stale within months. Users should
# derive headroom from the current EFO via get_efo_fiscal() or consult the
# OBR's press release for the relevant vintage.

fiscal_rules_table <- function() {
  data.frame(
    rule = c("stability", "investment", "welfare_cap"),
    metric = c(
      "Current budget balance, % of GDP",
      "Public Sector Net Financial Liabilities (PSNFL), % of GDP, year-on-year change",
      "AME-capped welfare spending, GBP bn (cash)"
    ),
    target_description = c(
      "Current budget in balance or surplus by the target year",
      "PSNFL as a share of GDP falling year-on-year by the target year",
      "AME-capped welfare spending below the cap level by the target year"
    ),
    target_year_rule = c(
      "5th forecast year (rolls to 3rd from 2026-27 onwards)",
      "5th forecast year (rolls to 3rd from 2026-27 onwards)",
      "Set in nominal terms for a specific fiscal year by the Treasury"
    ),
    direction_of_pass = c(
      "current budget surplus larger -> pass",
      "PSNFL/GDP falling more steeply -> pass",
      "AME-capped spending further below cap -> pass"
    ),
    source_charter = "Charter for Budget Responsibility, Autumn 2024 (updated Autumn 2025)",
    source_act     = "Budget Responsibility Act 2024 (fiscal lock)",
    stringsAsFactors = FALSE
  )
}

#' Get the current UK fiscal rules
#'
#' Returns a data frame describing the Charter for Budget Responsibility
#' fiscal rules currently in force, as encoded in the package at the time
#' of release. Numerical headroom against each rule is *not* shipped as a
#' constant because it is updated at every fiscal event and would go stale
#' within months; use [obr_headroom()] to derive the stability-rule margin
#' from the current EFO, or consult the OBR's EFO press release for the
#' relevant vintage.
#'
#' The current Charter (Autumn 2024, with an Autumn 2025 update) sets three
#' numerical rules:
#' \itemize{
#'   \item \strong{Stability rule}: current budget in balance or surplus by
#'     the target year. The target year is the 5th forecast year and rolls
#'     to the 3rd forecast year from 2026-27 onwards.
#'   \item \strong{Investment rule}: Public Sector Net Financial Liabilities
#'     (PSNFL) as a share of GDP falling year-on-year by the target year.
#'   \item \strong{Welfare cap}: AME-capped welfare spending below the
#'     statutory cap level by the target year.
#' }
#' The Budget Responsibility Act 2024 also adds a non-numerical
#' \strong{fiscal lock} requiring an OBR forecast for any fiscally
#' significant measure.
#'
#' @return A data frame with columns:
#' \describe{
#'   \item{rule}{Short rule code: `"stability"`, `"investment"`, `"welfare_cap"`.}
#'   \item{metric}{The metric the rule binds on.}
#'   \item{target_description}{Plain-English target.}
#'   \item{target_year_rule}{How the target year is set under the Charter.}
#'   \item{direction_of_pass}{Sign convention for headroom.}
#'   \item{source_charter}{Charter version that defines the rule.}
#'   \item{source_act}{Act of Parliament backing the rule.}
#' }
#'
#' @examples
#' obr_fiscal_rules()
#'
#' @family fiscal rules
#' @export
obr_fiscal_rules <- function() {
  fiscal_rules_table()
}

# Tests for obr_fiscal_rules() - offline.

test_that("obr_fiscal_rules() returns the three rules in the current Charter", {
  rules <- obr_fiscal_rules()
  expect_s3_class(rules, "data.frame")
  expect_setequal(rules$rule, c("stability", "investment", "welfare_cap"))
  expect_named(rules, c(
    "rule", "metric", "target_description", "target_year_rule",
    "direction_of_pass", "source_charter", "source_act"
  ))
  # All character columns are populated
  for (col in setdiff(names(rules), "rule")) {
    expect_true(all(nzchar(rules[[col]])),
                info = paste("column", col))
  }
})

test_that("obr_fiscal_rules() cites the current Charter and the Act", {
  rules <- obr_fiscal_rules()
  expect_match(rules$source_charter, "Autumn 2024", all = TRUE)
  expect_match(rules$source_act,     "Budget Responsibility Act 2024",
               all = TRUE)
})

test_that("obr_fiscal_rules() does not ship hardcoded numerical headroom", {
  rules <- obr_fiscal_rules()
  # Numerical headroom values are intentionally not shipped because they
  # change at every fiscal event. Guard against accidental re-introduction.
  expect_false("headroom_published_gbp_bn" %in% names(rules))
  expect_false("headroom" %in% names(rules))
})

#' @keywords internal
#' @concept OBR
#' @concept Office for Budget Responsibility
#' @concept fiscal policy
#' @concept public finances
#' @concept PSNB
#' @concept national debt
#' @concept UK budget
#' @concept economic forecasts
#' @importFrom utils head packageVersion
#' @importFrom stats aggregate
"_PACKAGE"

# Generic EFO detailed-forecast-table fetcher.
#
# get_efo_table(table_id) reads the catalogue, fetches the right workbook,
# dispatches to a layout-specific parser, and returns the standard v0.4.0
# schema. All EFO tables exposed by the package share this entry point.
# get_efo_fiscal() and get_efo_economy() are thin wrappers over it, kept
# for back-compat and as headline entry points.

# ---------------------------------------------------------------------------
# Layout-specific parsers
# ---------------------------------------------------------------------------

# Wide quarterly layout. Period in col 2 ("2008Q1"), series headers in the
# row immediately above the first data row, values in cols 3+.
#
# Header detection picks the row above first_data_row that has the most
# non-empty, non-numeric cells across cols 3:ncol. Single-column probing
# (the v0.4.0 implementation) failed on sheets like 1.15 (Potential output)
# where the canonical header row leaves col 3 empty.
parse_quarterly_wide <- function(path, sheet,
                                 default_metric_type = NA_character_,
                                 default_unit        = NA_character_,
                                 period_regex        = "^[0-9]{4}Q[1-4]$",
                                 period_type         = "quarter") {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  col2 <- as.character(unlist(raw[, 2L]))
  is_period <- grepl(period_regex, col2)
  first_data_row <- which(is_period)[1L]
  if (is.na(first_data_row)) return(NULL)

  header_row <- find_header_row(raw, first_data_row)
  if (is.na(header_row)) return(NULL)

  series <- trimws(gsub("\r\n", " ",
                        as.character(unlist(raw[header_row, 3L:ncol(raw)]))))
  valid_series <- !is.na(series) & series != ""

  data_idx <- which(is_period)
  periods  <- col2[data_idx]

  metric <- classify_metric_type(series)
  if (!is.na(default_metric_type)) {
    metric[!is.na(metric) & metric == "level"] <- default_metric_type
  }
  unit <- default_unit_for_metric(metric)
  if (!is.na(default_unit)) {
    unit[is.na(unit)] <- default_unit
  }

  result_list <- list()
  for (j in which(valid_series)) {
    col_idx <- j + 2L
    if (col_idx > ncol(raw)) next
    vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[data_idx, col_idx])))
    )
    if (all(is.na(vals))) next
    result_list[[length(result_list) + 1L]] <- obr_long(
      period      = periods,
      period_type = period_type,
      series      = series[j],
      value       = vals,
      unit        = unit[j],
      metric_type = metric[j]
    )
  }

  if (length(result_list) == 0L) return(NULL)
  result <- do.call(rbind, result_list)
  result[!is.na(result$value), ]
}

# Sub-sector matrix layout: a single fiscal year, transactions as rows,
# sub-sectors (Central government / Local authorities / etc) as columns.
# Used by Table 6.4 (Public sector transactions by sub-sector). Returns the
# standard v0.4 schema plus a `sub_sector` column carrying the column header.
parse_subsector_matrix <- function(path, sheet,
                                   default_metric_type = "level",
                                   default_unit        = "gbp_bn") {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  # Locate the single fiscal-year cell (typically row 4 col 3)
  fy_row <- NA_integer_
  for (i in seq_len(nrow(raw))) {
    cells <- as.character(unlist(raw[i, ]))
    if (any(grepl("^[0-9]{4}-[0-9]{2}$", cells), na.rm = TRUE)) {
      fy_row <- i
      break
    }
  }
  if (is.na(fy_row)) return(NULL)
  fy_cells <- as.character(unlist(raw[fy_row, ]))
  fy_idx   <- which(grepl("^[0-9]{4}-[0-9]{2}$", fy_cells))[1L]
  fiscal_year <- fy_cells[fy_idx]

  # Sub-sector header row is below the fiscal-year row; first non-empty
  # row whose cells in cols 3+ are mostly non-numeric strings.
  subsector_row <- NA_integer_
  for (i in (fy_row + 1L):min(fy_row + 4L, nrow(raw))) {
    cells <- as.character(unlist(raw[i, 3L:ncol(raw)]))
    n_chr <- sum(!is.na(cells) & nzchar(cells) &
                   is.na(suppressWarnings(as.numeric(cells))))
    if (n_chr >= 2L) { subsector_row <- i; break }
  }
  if (is.na(subsector_row)) return(NULL)

  subsectors <- trimws(gsub("\r\n", " ",
                            as.character(unlist(raw[subsector_row, ]))))
  ss_cols    <- which(!is.na(subsectors) & nzchar(subsectors) &
                        seq_along(subsectors) >= 3L)
  if (length(ss_cols) == 0L) return(NULL)

  # Walk data rows: series name in col 2, values in ss_cols.
  results <- list()
  if (subsector_row >= nrow(raw)) return(NULL)
  for (i in (subsector_row + 1L):nrow(raw)) {
    series <- as.character(unlist(raw[i, 2L]))
    if (is.na(series) || !nzchar(series)) next
    for (j in ss_cols) {
      cell <- as.character(unlist(raw[i, j]))
      val  <- suppressWarnings(as.numeric(cell))
      if (is.na(val)) next
      out <- obr_long(
        period      = fiscal_year,
        period_type = "fiscal_year",
        series      = series,
        value       = val,
        unit        = default_unit,
        metric_type = default_metric_type
      )
      out$sub_sector <- subsectors[j]
      results[[length(results) + 1L]] <- out
    }
  }
  if (length(results) == 0L) return(NULL)
  do.call(rbind, results)
}

# Indented quarterly layout: periods in col 3 in "Q1 2016" format, single
# value column in col 4. Section markers ("Outturn" / "Forecast") sit in
# col 2 sparsely (only on first row of each section). Used by 6.10.
parse_quarterly_indented <- function(path, sheet,
                                     default_metric_type = "pct",
                                     default_unit        = "pct") {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  if (ncol(raw) < 4L) return(NULL)

  col2 <- as.character(unlist(raw[, 2L]))
  col3 <- as.character(unlist(raw[, 3L]))
  col4 <- as.character(unlist(raw[, 4L]))

  # Periods: "Q1 2016" -> canonical "2016Q1"
  q_match <- regexec("^Q([1-4])\\s+([0-9]{4})$", col3)
  is_period <- vapply(q_match, function(m) m[[1L]][1L] != -1L, logical(1L))
  data_idx  <- which(is_period)
  if (length(data_idx) == 0L) return(NULL)

  periods <- vapply(regmatches(col3[data_idx], q_match[data_idx]),
                    function(parts) sprintf("%sQ%s", parts[3L], parts[2L]),
                    character(1L))

  vals <- suppressWarnings(as.numeric(col4[data_idx]))

  series <- efo_sheet_series_name(raw, sheet)

  obr_long(
    period      = periods,
    period_type = "quarter",
    series      = series,
    value       = vals,
    unit        = default_unit,
    metric_type = default_metric_type
  )
}

# Cross-reference follow-through: a sheet whose only content is a redirect
# like "See Table 6.2 of our November 2025 Economic and fiscal outlook".
# Resolves the redirect, fetches the named table from the named vintage's
# Aggregates workbook, and parses it with parse_fiscal_year_wide.
follow_cross_reference <- function(path, sheet, refresh = FALSE) {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")
  txt <- paste(stats::na.omit(as.character(unlist(raw))), collapse = " ")
  m <- regmatches(
    txt,
    regexec(paste0("See Table ([0-9]+\\.[0-9]+[a-z]?) of our ",
                   "([A-Z][a-z]+) ([0-9]{4})"), txt)
  )[[1L]]
  if (length(m) < 4L) return(NULL)
  target_table <- m[2L]
  vintage      <- paste(m[3L], m[4L])

  src <- tryCatch(
    efo_aggregates_source(refresh = refresh, vintage = vintage),
    error = function(e) NULL
  )
  if (is.null(src)) return(NULL)

  data <- parse_fiscal_year_wide(
    src$path, target_table,
    default_metric_type = "level",
    default_unit        = "gbp_bn"
  )
  if (is.null(data)) return(NULL)

  list(
    data         = data,
    src          = src,
    target_table = target_table,
    vintage      = vintage
  )
}

# Internal: pick the row above first_data_row most likely to be the header.
# Score = number of non-empty, non-numeric cells in cols 3:ncol. The row with
# the highest score becomes the header. This works for sheets where the
# canonical name row leaves the early columns blank (sheet 1.15 etc).
find_header_row <- function(raw, first_data_row) {
  if (first_data_row < 2L) return(NA_integer_)
  best_row   <- NA_integer_
  best_score <- 0L
  for (i in (first_data_row - 1L):1L) {
    cells <- as.character(unlist(raw[i, 3L:ncol(raw)]))
    score <- sum(
      !is.na(cells) & nzchar(cells) &
        is.na(suppressWarnings(as.numeric(cells)))
    )
    if (score > best_score) {
      best_score <- score
      best_row   <- i
    }
  }
  best_row
}

# Wide annual layout where calendar YEARS are in col 2 (rows) and series are
# column headers (rows above the first data row). This is the inverse of
# parse_annual_year_wide, where years are column headers. Used by 1.19b
# (CPI category weights) for example.
parse_annual_period_wide <- function(path, sheet,
                                     default_metric_type = NA_character_,
                                     default_unit        = NA_character_) {
  parse_quarterly_wide(
    path, sheet,
    default_metric_type = default_metric_type,
    default_unit        = default_unit,
    period_regex        = "^[0-9]{4}$",
    period_type         = "calendar_year"
  )
}

# Single-series quarterly layout. Period in col 2, single value column to
# the right. Series name derives from the sheet's title row (row 2, col 2).
parse_quarterly_single <- function(path, sheet,
                                   default_metric_type = NA_character_,
                                   default_unit        = NA_character_) {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  col2 <- as.character(unlist(raw[, 2L]))
  is_period <- grepl("^[0-9]{4}Q[1-4]$", col2)
  data_idx  <- which(is_period)
  if (length(data_idx) == 0L) return(NULL)

  # Pick the column with the most numeric values aligned to the period rows.
  best_col <- NA_integer_
  best_n   <- -1L
  for (j in 3L:ncol(raw)) {
    vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[data_idx, j])))
    )
    n <- sum(!is.na(vals))
    if (n > best_n) {
      best_n   <- n
      best_col <- j
    }
  }
  if (is.na(best_col)) return(NULL)

  series <- efo_sheet_series_name(raw, sheet)

  metric <- if (!is.na(default_metric_type)) default_metric_type else "level"
  derived_unit <- default_unit_for_metric(metric)
  unit <- if (is.na(derived_unit)) default_unit else derived_unit

  obr_long(
    period      = col2[data_idx],
    period_type = "quarter",
    series      = series,
    value       = suppressWarnings(
      as.numeric(as.character(unlist(raw[data_idx, best_col])))
    ),
    unit        = unit,
    metric_type = metric
  )
}

# Wide annual layout with calendar years as columns. Series in col 2,
# data in cols matching ^[0-9]{4}$.
parse_annual_year_wide <- function(path, sheet,
                                   default_metric_type = NA_character_,
                                   default_unit        = NA_character_) {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  year_row <- NA_integer_
  for (i in seq_len(nrow(raw))) {
    cells <- as.character(unlist(raw[i, ]))
    if (sum(grepl("^[0-9]{4}$", cells), na.rm = TRUE) >= 2L) {
      year_row <- i
      break
    }
  }
  if (is.na(year_row)) return(NULL)

  year_cells <- as.character(unlist(raw[year_row, ]))
  year_cols  <- which(grepl("^[0-9]{4}$", year_cells))
  years      <- year_cells[year_cols]

  results <- list()
  if (year_row >= nrow(raw)) return(NULL)
  for (i in (year_row + 1L):nrow(raw)) {
    row  <- as.character(unlist(raw[i, ]))
    name <- if (length(row) >= 2L) row[2L] else NA_character_
    if (is.na(name) || !nzchar(name) || grepl("^Note", name, ignore.case = TRUE)) next

    vals <- suppressWarnings(as.numeric(row[year_cols]))
    if (all(is.na(vals))) next

    metric <- classify_metric_type(name)
    if (!is.na(default_metric_type) &&
        !is.na(metric) && metric == "level") {
      metric <- default_metric_type
    }
    derived_unit <- default_unit_for_metric(metric)
    unit <- if (is.na(derived_unit)) default_unit else derived_unit

    results[[length(results) + 1L]] <- obr_long(
      period      = years,
      period_type = "calendar_year",
      series      = name,
      value       = vals,
      unit        = unit,
      metric_type = metric
    )
  }

  if (length(results) == 0L) return(NULL)
  out <- do.call(rbind, results)
  out[!is.na(out$value), ]
}

# Wide fiscal-year layout. Year row contains fiscal-year strings ("2024-25").
# Series name comes from col 2 (top-level row) or col 3 (sub-row beneath an
# "of which:" marker). Empty rows and "of which:" markers are skipped.
parse_fiscal_year_wide <- function(path, sheet,
                                   default_metric_type = NA_character_,
                                   default_unit        = NA_character_) {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  year_row <- NA_integer_
  for (i in seq_len(nrow(raw))) {
    cells <- as.character(unlist(raw[i, ]))
    if (sum(grepl("^[0-9]{4}-[0-9]{2}$", cells), na.rm = TRUE) >= 2L) {
      year_row <- i
      break
    }
  }
  if (is.na(year_row)) return(NULL)

  year_cells   <- as.character(unlist(raw[year_row, ]))
  year_cols    <- which(grepl("^[0-9]{4}-[0-9]{2}$", year_cells))
  fiscal_years <- year_cells[year_cols]

  results <- list()
  if (year_row >= nrow(raw)) return(NULL)
  for (i in (year_row + 1L):nrow(raw)) {
    row <- as.character(unlist(raw[i, ]))

    # Pull series name: prefer col 2 (top-level), fall back to col 3 (sub-row)
    name_2 <- if (length(row) >= 2L) row[2L] else NA_character_
    name_3 <- if (length(row) >= 3L) row[3L] else NA_character_

    series <- NA_character_
    if (!is.na(name_2) && nzchar(name_2) && name_2 != "of which:") {
      series <- name_2
    } else if (!is.na(name_3) && nzchar(name_3) && name_3 != "of which:") {
      series <- name_3
    }
    if (is.na(series)) next

    vals <- suppressWarnings(as.numeric(row[year_cols]))
    if (all(is.na(vals))) next

    metric <- classify_metric_type(series)
    if (!is.na(default_metric_type) &&
        !is.na(metric) && metric == "level") {
      metric <- default_metric_type
    }
    derived_unit <- default_unit_for_metric(metric)
    unit <- if (is.na(derived_unit)) default_unit else derived_unit

    results[[length(results) + 1L]] <- obr_long(
      period      = fiscal_years,
      period_type = "fiscal_year",
      series      = series,
      value       = vals,
      unit        = unit,
      metric_type = metric
    )
  }

  if (length(results) == 0L) return(NULL)
  out <- do.call(rbind, results)
  out[!is.na(out$value), ]
}

# Internal: derive a clean series name from the sheet's title row.
# Title row 2 col 2 is typically "1.4 Nominal GDP (£ billion)"; we strip
# the leading "1.4 " and any trailing parenthesis to get "Nominal GDP".
efo_sheet_series_name <- function(raw, sheet, fallback = NULL) {
  title <- tryCatch(as.character(unlist(raw[2L, 2L])),
                    error = function(e) NA_character_)
  if (is.na(title) || !nzchar(title)) {
    return(if (is.null(fallback)) sheet else fallback)
  }
  out <- gsub("^[0-9]+\\.[0-9]+[a-z]?\\s*[:.]?\\s*", "", title)
  out <- gsub("\\s*\\([^)]*\\)\\s*$", "", out)
  trimws(out)
}

# Hand-curated short series names for single-series tables (preserves the
# v0.4.0 series name for output_gap and gives stable names elsewhere).
efo_single_series_overrides <- function() {
  c(
    "1.4"  = "Nominal GDP",
    "1.14" = "Output gap",
    "1.20" = "Electricity price"
  )
}

# ---------------------------------------------------------------------------
# Dispatcher
# ---------------------------------------------------------------------------

#' Get any EFO detailed-forecast table by id
#'
#' Generic fetcher that takes an EFO table identifier (e.g. `"6.5"`,
#' `"1.7"`, `"4.1"`) and returns the parsed contents in the standard
#' v0.4.0 schema. Use [obr_efo_catalogue()] to discover which tables
#' are available.
#'
#' @details
#' Internally, this function looks up `table_id` in the catalogue, fetches
#' the right workbook (Aggregates or Economy), dispatches to a layout-
#' specific parser, and tags every row with `metric_type` and `unit`
#' according to the catalogue's defaults plus per-row classification.
#'
#' Coverage today: 17 fiscal Aggregates tables + 22 macro Economy tables.
#' One sheet (6.11 PSND year-on-year changes) is a cross-reference to a
#' previous EFO and returns `NULL` with a warning rather than data; OBR
#' itself directs users to the previous EFO for that table.
#'
#' Headline functions [get_efo_fiscal()] and [get_efo_economy()] are kept
#' as thin wrappers over this dispatcher.
#'
#' @param table_id Character. The EFO table identifier, e.g. `"6.5"`,
#'   `"1.7"`, `"6.13"`. See [obr_efo_catalogue()] for the full list.
#' @param vintage Optional EFO vintage label (e.g. `"October 2024"`). If
#'   `NULL`, uses any pin set via [obr_pin()] or falls back to the latest
#'   live EFO.
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema columns
#' (`period`, `period_type`, `series`, `metric_type`, `value`, `unit`).
#' Returns `NULL` (with a warning) for cross-reference sheets.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'
#'   # Net borrowing components (same data as get_efo_fiscal())
#'   get_efo_table("6.5")
#'
#'   # CPI category inflation by year
#'   get_efo_table("1.19")
#'
#'   # Composition of public sector net debt
#'   get_efo_table("6.13")
#'
#'   # Pin to a specific vintage
#'   get_efo_table("6.5", vintage = "October 2024")
#' })
#' options(op)
#' }
#'
#' @family EFO
#' @export
get_efo_table <- function(table_id, vintage = NULL, refresh = FALSE) {
  if (!is.character(table_id) || length(table_id) != 1L) {
    cli::cli_abort("{.arg table_id} must be a single character string.")
  }

  meta <- efo_catalogue_lookup(table_id)

  src <- if (meta$file == "aggregates") {
    efo_aggregates_source(refresh = refresh, vintage = vintage)
  } else {
    efo_economy_source(refresh = refresh, vintage = vintage)
  }

  # Cross-reference: follow the redirect to the previous EFO and return its
  # equivalent table. Provenance points at the previous vintage; notes
  # explain the redirect.
  if (meta$layout == "cross_reference") {
    redirect <- follow_cross_reference(src$path, table_id, refresh = refresh)
    if (is.null(redirect)) {
      cli::cli_warn(c(
        "EFO table {.val {table_id}} ({meta$title}) is a cross-reference.",
        "i" = "Could not resolve the redirect to a previous EFO."
      ))
      return(NULL)
    }
    notes <- sprintf(
      "Cross-reference: this current-EFO sheet points at Table %s of the %s EFO. Data sourced from there.",
      redirect$target_table, redirect$vintage
    )
    return(new_obr_tbl(
      data        = redirect$data,
      publication = "EFO",
      vintage     = redirect$vintage,
      source_url  = redirect$src$url,
      retrieved   = redirect$src$retrieved,
      file_md5    = redirect$src$file_md5,
      notes       = notes
    ))
  }

  data <- switch(
    meta$layout,
    "quarterly_wide"      = parse_quarterly_wide(
      src$path, table_id,
      meta$default_metric_type, meta$default_unit
    ),
    "quarterly_single"    = parse_quarterly_single(
      src$path, table_id,
      meta$default_metric_type, meta$default_unit
    ),
    "annual_year_wide"    = parse_annual_year_wide(
      src$path, table_id,
      meta$default_metric_type, meta$default_unit
    ),
    "annual_period_wide"  = parse_annual_period_wide(
      src$path, table_id,
      meta$default_metric_type, meta$default_unit
    ),
    "fiscal_year_wide"    = parse_fiscal_year_wide(
      src$path, table_id,
      meta$default_metric_type, meta$default_unit
    ),
    "subsector_matrix"    = parse_subsector_matrix(
      src$path, table_id,
      meta$default_metric_type %||% "level",
      meta$default_unit        %||% "gbp_bn"
    ),
    "quarterly_indented"  = parse_quarterly_indented(
      src$path, table_id,
      meta$default_metric_type %||% "pct",
      meta$default_unit        %||% "pct"
    ),
    cli::cli_abort(c(
      "Unknown layout {.val {meta$layout}} for table {.val {table_id}}.",
      "i" = "This is a package bug; please file an issue."
    ))
  )

  # Apply single-series name overrides (for output_gap etc.) so v0.4.0
  # series names are preserved.
  if (meta$layout == "quarterly_single") {
    overrides <- efo_single_series_overrides()
    if (table_id %in% names(overrides) && !is.null(data)) {
      data$series <- overrides[[table_id]]
    }
  }

  if (is.null(data)) {
    cli::cli_warn(c(
      "EFO table {.val {table_id}} could not be parsed (layout {meta$layout}).",
      "i" = "Cached file may be from a vintage that did not include this table."
    ))
    return(NULL)
  }

  efo_obr_tbl(data, src)
}

# obr_headroom() tests. Validation is offline; the derivation test hits
# the live EFO aggregates workbook and is skipped on CRAN.

test_that("obr_headroom() validates target_year", {
  expect_error(obr_headroom(target_year = "2029"),
               regexp = "fiscal-year string")
  expect_error(obr_headroom(target_year = 2029),
               regexp = "fiscal-year string")
})

test_that("obr_headroom() flips the sign of the current budget deficit", {
  skip_on_cran()
  skip_if_offline()
  op <- options(obr.cache_dir = tempdir())
  on.exit(options(op), add = TRUE)

  hr  <- obr_headroom()
  efo <- get_efo_table("6.5")

  expect_s3_class(hr, "obr_tbl")
  expect_true(all(hr$series == "Current budget surplus"))
  expect_true(all(hr$unit == "gbp_bn"))

  cbd <- efo[efo$series == "Current budget deficit", ]
  joined <- merge(hr, cbd, by = "period")
  expect_gt(nrow(joined), 0L)
  expect_equal(joined$value.x, -joined$value.y)
})

test_that("obr_headroom() flags the target year and warns when absent", {
  skip_on_cran()
  skip_if_offline()
  op <- options(obr.cache_dir = tempdir())
  on.exit(options(op), add = TRUE)

  years <- obr_headroom()$period
  ty    <- years[length(years)]
  hr    <- obr_headroom(target_year = ty)
  expect_true("is_target_year" %in% names(hr))
  expect_equal(sum(hr$is_target_year), 1L)
  expect_equal(hr$period[hr$is_target_year], ty)

  expect_warning(obr_headroom(target_year = "1999-00"),
                 regexp = "not a year in this EFO forecast")
})

# Convert the OBR Historical Forecasts Database from tidy long format
# (rows = forecast vintage x fiscal year) into a wide real-time panel
# (rows = forecast vintage, columns = fiscal year). Useful for plotting
# how successive OBR forecasts evolved for the same fiscal year, or for
# computing forecast errors against outturn.

# Convert an OBR-style forecast date string ("March 2024", "October 2010")
# into an integer index for chronological ordering. NA on unrecognised input.
forecast_date_to_index <- function(date_str) {
  parts <- strsplit(date_str, " ", fixed = TRUE)
  vapply(parts, function(p) {
    if (length(p) < 2L) return(NA_integer_)
    m <- match(tolower(p[1]), tolower(month.name))
    y <- suppressWarnings(as.integer(p[2]))
    if (is.na(m) || is.na(y)) return(NA_integer_)
    as.integer(y * 12L + m)
  }, integer(1))
}

#' Build a wide real-time panel of OBR forecasts
#'
#' Takes the long-format Historical Forecasts Database output of
#' [get_forecasts()] and pivots it to a wide panel with one row per
#' forecast vintage (e.g. `"March 2024"`) and one column per fiscal year
#' being forecast (e.g. `"2024-25"`). The first column is the forecast
#' vintage; remaining columns are numeric forecast values.
#'
#' This format mirrors how the OBR's own Historical Forecasts Database is
#' laid out and how forecast-error studies (e.g. the OBR's own Forecast
#' Evaluation Report) decompose performance: each row is one forecast
#' published at a fiscal event, each column is the target year being
#' forecast, and the diagonal-like structure lets you read the h-step
#' ahead forecast for any vintage.
#'
#' @param series Character. Series to return; see [list_forecast_series()].
#'   Defaults to `"PSNB"`.
#' @param refresh Logical. If `TRUE`, re-download the underlying database.
#' @param order_chronologically Logical. If `TRUE` (the default), rows are
#'   ordered from earliest to latest forecast vintage. If `FALSE`, rows
#'   appear in the order returned by [get_forecasts()].
#'
#' @return An `obr_tbl` whose first column (`forecast_date`) is character
#'   and whose remaining columns are numeric forecast values, one per
#'   fiscal year. Provenance is inherited from the underlying call to
#'   [get_forecasts()] and a `notes` field describes the panel.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   panel <- obr_forecast_panel("PSNB")
#'   # PSNB forecast for 2024-25 across every vintage
#'   panel[, c("forecast_date", "2024-25")]
#' })
#' options(op)
#' }
#'
#' @family forecasts
#' @export
obr_forecast_panel <- function(series = "PSNB",
                               refresh = FALSE,
                               order_chronologically = TRUE) {
  fc   <- get_forecasts(series = series, refresh = refresh)
  prov <- attr(fc, "obr_provenance")
  long <- as.data.frame(fc)

  if (order_chronologically) {
    long$.idx <- forecast_date_to_index(long$forecast_date)
    long <- long[order(long$.idx), , drop = FALSE]
    long$.idx <- NULL
  }

  # Use stats::reshape for the long-to-wide pivot.
  # `period` replaced `fiscal_year` in v0.4.0; the source column is still
  # the fiscal year being forecast.
  wide <- stats::reshape(
    long[, c("forecast_date", "period", "value")],
    idvar     = "forecast_date",
    timevar   = "period",
    direction = "wide"
  )
  rownames(wide) <- NULL
  names(wide) <- sub("^value\\.", "", names(wide))

  # Sort columns: forecast_date first, then fiscal years ascending
  yr_cols <- setdiff(names(wide), "forecast_date")
  yr_cols <- yr_cols[order(yr_cols)]
  wide    <- wide[, c("forecast_date", yr_cols), drop = FALSE]

  new_obr_tbl(
    data        = wide,
    publication = prov$publication,
    vintage     = prov$vintage,
    source_url  = prov$source_url,
    retrieved   = prov$retrieved,
    file_md5    = prov$file_md5,
    notes       = sprintf(
      "Wide real-time panel for %s: rows = forecast vintage, columns = fiscal year.",
      series
    )
  )
}

test_that("forecast_date_to_index() orders OBR-style dates correctly", {
  d <- c("November 2010", "March 2011", "July 2015",
         "December 2014", "March 2024")
  idx <- forecast_date_to_index(d)
  expect_type(idx, "integer")
  expect_equal(length(idx), length(d))
  # Ordering matches calendar order
  ordered <- d[order(idx)]
  expect_equal(ordered,
               c("November 2010", "March 2011", "December 2014",
                 "July 2015", "March 2024"))
})

test_that("forecast_date_to_index() returns NA on un-recognised input", {
  expect_true(is.na(forecast_date_to_index("not a date")))
  expect_true(is.na(forecast_date_to_index("Foo 2024")))
})

test_that("obr_forecast_panel() returns a wide obr_tbl with provenance", {
  skip_on_cran()
  skip_if_offline()

  panel <- obr_forecast_panel("PSNB")
  expect_s3_class(panel, "obr_tbl")
  expect_true("forecast_date" %in% names(panel))
  # At least one fiscal year column
  yr_cols <- setdiff(names(panel), "forecast_date")
  expect_gt(length(yr_cols), 5L)
  expect_true(all(grepl("^[0-9]{4}-[0-9]{2}$", yr_cols)))
  # Numeric values
  expect_true(is.numeric(panel[[yr_cols[1L]]]))

  prov <- obr_provenance(panel)
  expect_equal(prov$publication, "HFD")
  expect_match(prov$notes, "real-time panel")
})

# Welfare Trends Report: charts and tables.
# URL resolved dynamically; resolver warns if it falls through to the
# hardcoded fallback.
WTR_FALLBACK <- "https://obr.uk/download/welfare-trends-report-october-2024-charts-and-tables/"
WTR_FILENAME <- "welfare_trends.xlsx"

wtr_source <- function(refresh = FALSE) {
  obr_get_xlsx(
    candidates = wtr_url_candidates(),
    fallback   = WTR_FALLBACK,
    filename   = WTR_FILENAME,
    refresh    = refresh,
    label      = "Welfare Trends Report"
  )
}

wtr_obr_tbl <- function(data, src) {
  new_obr_tbl(
    data        = data,
    publication = "WTR",
    vintage     = obr_url_vintage(src$url),
    source_url  = src$url,
    retrieved   = src$retrieved,
    file_md5    = src$file_md5
  )
}

# Generic parser for WTR chart data sheets.
# Layout: col 1 = NA (except "Back to contents" in row 1),
#         col 2 = chart title (row 2) then series names in last data rows,
#         cols 3+ = fiscal year labels (year_row) then values (data rows).
#
# Returns the v0.4.0 schema: period, period_type, series, metric_type, value, unit.
# Most WTR data is denominated as a percentage of GDP; the unit defaults to
# "pct" when the heuristic classifier cannot assign a more specific unit.
# Caller can override by post-processing the returned frame (e.g. caseload
# series that are denominated in thousands of claimants).
parse_wtr_chart <- function(path, sheet, unit_default = "pct") {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")
  col2 <- as.character(unlist(raw[, 2]))

  non_na    <- which(!is.na(col2) & col2 != "")
  data_rows <- non_na[non_na > 2L]
  if (length(data_rows) == 0L) return(NULL)

  year_row_idx <- data_rows[1L] - 1L
  all_yr       <- as.character(unlist(raw[year_row_idx, ]))
  year_cols    <- which(grepl("^[0-9]{4}-[0-9]{2}$", all_yr))
  fiscal_years <- all_yr[year_cols]
  if (length(year_cols) == 0L) return(NULL)

  series_names <- col2[data_rows]

  result_list <- vector("list", length(data_rows))
  for (j in seq_along(data_rows)) {
    vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[data_rows[j], year_cols])))
    )
    metric <- classify_metric_type(series_names[j])
    derived_unit <- default_unit_for_metric(metric)
    unit <- if (is.na(derived_unit)) unit_default else derived_unit
    result_list[[j]] <- obr_long(
      period      = fiscal_years,
      period_type = "fiscal_year",
      series      = series_names[j],
      value       = vals,
      unit        = unit,
      metric_type = metric
    )
  }

  result <- do.call(rbind, result_list)
  result[!is.na(result$value), ]
}

#' Get working-age welfare spending
#'
#' Downloads (and caches) the OBR Welfare Trends Report charts and tables
#' workbook and returns annual working-age welfare spending as a share of
#' GDP, split into incapacity-related and non-incapacity spending.
#'
#' Data cover fiscal years from 1978-79 through the current forecast horizon.
#' The exact vintage is recorded in the returned object's provenance.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema (columns:
#' `period`, `period_type`, `series`, `metric_type`, `value`, `unit`).
#' Values are spending as a percentage of GDP; `metric_type` is `"pct"`,
#' `unit` is `"pct"`.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   welfare <- get_welfare_spending()
#'   welfare[welfare$series == "Working-age incapacity benefits spending" &
#'   welfare$period >= "2000-01", ]
#' })
#' options(op)
#' }
#'
#' @family welfare
#' @export
get_welfare_spending <- function(refresh = FALSE) {
  src <- wtr_source(refresh)
  wtr_obr_tbl(parse_wtr_chart(src$path, "C1.3"), src)
}

#' Get incapacity benefits spending by type
#'
#' Downloads (and caches) the OBR Welfare Trends Report charts and tables
#' workbook and returns annual spending on each incapacity benefit as a
#' share of GDP, from 1978-79 to the current forecast horizon.
#'
#' Series include: Invalidity Benefit, Incapacity Benefit, Employment and
#' Support Allowance (ESA), Sickness Benefit, and Severe Disablement
#' Allowance.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema. `series` is the
#' benefit name, values are spending as a percentage of GDP, `metric_type`
#' is `"pct"`, `unit` is `"pct"`. See [get_public_finances()] for full
#' column docs.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   ib <- get_incapacity_spending()
#'   unique(ib$series)
#' })
#' options(op)
#' }
#'
#' @family welfare
#' @export
get_incapacity_spending <- function(refresh = FALSE) {
  src <- wtr_source(refresh)
  wtr_obr_tbl(parse_wtr_chart(src$path, "C1.1"), src)
}

#' Get incapacity benefit caseloads
#'
#' Downloads (and caches) the OBR Welfare Trends Report charts and tables
#' workbook and returns the combined incapacity benefit caseload since
#' 2008-09, in both absolute terms (thousands of claimants) and as a share
#' of the working-age population.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema. The two series
#' (`"Claimants"` and `"Share of working age population"`) carry different
#' units: claimants are in thousands and the share is a percentage. After
#' calling, the caller may want to overwrite `unit` to `"count_k"` for the
#' claimants series, since the heuristic classifier cannot infer the
#' "thousands" denomination from the series name alone.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   cases <- get_incapacity_caseloads()
#'   cases[cases$series == "Claimants", ]
#' })
#' options(op)
#' }
#'
#' @family welfare
#' @export
get_incapacity_caseloads <- function(refresh = FALSE) {
  src <- wtr_source(refresh)
  out <- parse_wtr_chart(src$path, "C3.1")
  if (!is.null(out)) {
    is_claimants <- out$series == "Claimants"
    out$unit[is_claimants] <- "count_k"
    out$metric_type[is_claimants] <- "level"
  }
  wtr_obr_tbl(out, src)
}

# v0.4.0 standard tidy long schema.
v04_long_cols <- c("period", "period_type", "series",
                   "metric_type", "value", "unit")

test_that("get_welfare_spending() returns the v0.4.0 schema", {
  skip_on_cran()
  skip_if_offline()

  result <- get_welfare_spending()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_type(result$period, "character")
  expect_type(result$series, "character")
  expect_type(result$value,  "double")
  expect_gt(nrow(result), 50)
  expect_true(any(grepl("incapacity", result$series, ignore.case = TRUE)))
  expect_true(all(grepl("^[0-9]{4}-[0-9]{2}$", result$period)))
  # All values are percentages of GDP
  expect_true(all(result$value >= 0 & result$value < 20))
  expect_true(all(result$period_type == "fiscal_year"))
})

test_that("get_welfare_spending() covers from 1978-79", {
  skip_on_cran()
  skip_if_offline()

  result <- get_welfare_spending()
  expect_true("1978-79" %in% result$period)
})

test_that("get_incapacity_spending() returns the v0.4.0 schema", {
  skip_on_cran()
  skip_if_offline()

  result <- get_incapacity_spending()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_gt(nrow(result), 100)
  expect_true(any(grepl("ESA|Employment and support",
                        result$series, ignore.case = TRUE)))
  expect_true(any(grepl("Invalidity", result$series)))
})

test_that("get_incapacity_caseloads() returns the v0.4.0 schema with mixed units", {
  skip_on_cran()
  skip_if_offline()

  result <- get_incapacity_caseloads()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_gt(nrow(result), 20)
  expect_true("Claimants" %in% result$series)
  expect_true(any(grepl("working age", result$series, ignore.case = TRUE)))
  claimants <- result[result$series == "Claimants", ]
  expect_true("2008-09" %in% claimants$period)
  # Claimants in thousands; roughly 2,000 to 3,500 for incapacity
  expect_true(all(claimants$value > 500 & claimants$value < 10000))
  # Wrapper retags the claimants series with the right unit
  expect_true(all(claimants$unit == "count_k"))
})

test_that("WTR functions return obr_tbl with WTR provenance", {
  skip_on_cran()
  skip_if_offline()

  for (fn in list(get_welfare_spending, get_incapacity_spending,
                  get_incapacity_caseloads)) {
    res  <- fn()
    expect_s3_class(res, "obr_tbl")
    prov <- obr_provenance(res)
    expect_equal(prov$publication, "WTR")
    expect_match(prov$source_url, "welfare-trends-report")
    expect_match(prov$vintage, "^[A-Z][a-z]+ [0-9]{4}$")
  }
})

# Fiscal Risks and Sustainability Report.
#
# get_pension_projections() was deprecated in 0.5.1. In July 2026 the OBR
# restructured the FSR workbook set: the executive-summary sheet this
# function parsed ("C1.2", state pension spending under demographic and
# triple-lock scenarios) no longer exists. The equivalent series now sits in
# the Chapter 3 "Long-term spending projections" workbook (Chart 3.11,
# "State pension spending under alternative uprating assumptions"), with a
# different scenario structure. The function is kept as a stub so existing
# scripts do not fail with "object not found"; the download and parsing code
# has been removed.

#' Get long-run state pension spending projections (deprecated)
#'
#' @description
#' **Deprecated since obr 0.5.1.**
#'
#' In July 2026 the OBR restructured the Fiscal Risks and Sustainability
#' Report workbook set. The executive-summary sheet this function read
#' (`C1.2`, state pension spending split into demographic and triple-lock
#' scenarios) no longer exists, so the function could no longer return data.
#'
#' The equivalent series is now published in the FSR Chapter 3 "Long-term
#' spending projections" workbook, as Chart 3.11 "State pension spending
#' under alternative uprating assumptions", with a different scenario
#' structure (triple-lock, CPI, and average-earnings uprating rather than the
#' old demographic vs triple-lock split). See
#' <https://obr.uk/frs/fiscal-risks-and-sustainability-july-2026/>.
#'
#' This stub is retained so existing scripts do not error. It emits a
#' deprecation warning and returns `NULL`. It will be removed in a future
#' release.
#'
#' @param refresh Ignored. Retained so existing calls do not error.
#'
#' @return `NULL`, invisibly.
#'
#' @examples
#' # Deprecated since 0.5.1: emits a warning and returns NULL.
#' suppressWarnings(get_pension_projections())
#'
#' @family long-term fiscal
#' @export
get_pension_projections <- function(refresh = FALSE) {
  .Deprecated(
    msg = paste0(
      "`get_pension_projections()` is deprecated and now returns NULL.\n",
      "The OBR restructured the Fiscal Risks and Sustainability Report in ",
      "July 2026; the state pension spending scenarios this function read ",
      "(executive-summary sheet 'C1.2') are no longer published in that ",
      "form. The equivalent series now sits in the FSR Chapter 3 workbook ",
      "(Chart 3.11, 'State pension spending under alternative uprating ",
      "assumptions'). See ",
      "https://obr.uk/frs/fiscal-risks-and-sustainability-july-2026/."
    )
  )
  invisible(NULL)
}

# get_pension_projections() was deprecated in 0.5.1 after the OBR restructured
# the FSR workbooks (July 2026). It no longer accesses the network; it warns
# and returns NULL. See R/fiscal_sustainability.R.

test_that("get_pension_projections() is deprecated and returns NULL", {
  expect_warning(
    result <- get_pension_projections(),
    "deprecated",
    class = "deprecatedWarning"
  )
  expect_null(result)
})

test_that("get_pension_projections() ignores its refresh argument", {
  # The stub must not error on the old signature and must not hit the network.
  expect_warning(result <- get_pension_projections(refresh = TRUE), "deprecated")
  expect_null(result)
})

# Stable OBR download URL (server redirects to the latest vintage).
PFD_URL      <- "https://obr.uk/download/public-finances-databank/"
PFD_FILENAME <- "public_finances_databank.xlsx"

pfd_source <- function(refresh = FALSE) {
  obr_get_xlsx(
    candidates = PFD_URL,
    fallback   = PFD_URL,
    filename   = PFD_FILENAME,
    refresh    = refresh,
    label      = "Public Finances Databank"
  )
}

# Strip a trailing footnote marker (1 or 2 trailing digits) from a label, but
# only when preceded by a letter, period, or whitespace, so genuinely numeric
# trailing chars (e.g. years) are preserved. This protects series names that
# legitimately end in digits from being silently corrupted.
strip_footnote <- function(x) {
  trimws(gsub("(?<=[A-Za-z\\.\\s])[0-9]{1,2}$", "", x, perl = TRUE))
}

# Parse "Aggregates (£bn)" sheet in tidy long format.
# Row 4 = series names; data starts at row 8.
parse_aggregates_bn <- function(path) {
  sheet <- resolve_sheet_name(
    path,
    primary          = "Aggregates (\u00a3bn)",
    fallback_pattern = "^Aggregates \\(.{1,2}bn\\)$"
  )
  raw <- readxl::read_excel(path, sheet = sheet,
                             col_names = FALSE, .name_repair = "minimal")

  series_names      <- as.character(unlist(raw[4, ]))
  series_names[1]   <- "year"
  series_names[series_names %in% c("NA", "NULL")] <- NA

  keep <- !is.na(series_names)
  data <- as.data.frame(raw[8:nrow(raw), keep], stringsAsFactors = FALSE)
  names(data) <- series_names[keep]

  data <- data[grepl("^[0-9]{4}-[0-9]{2}", data$year), ]

  value_cols <- setdiff(names(data), "year")
  result <- do.call(rbind, lapply(value_cols, function(col) {
    metric <- classify_metric_type(col)
    derived_unit <- default_unit_for_metric(metric)
    unit <- if (is.na(derived_unit)) "gbp_bn" else derived_unit
    obr_long(
      period      = data$year,
      period_type = "fiscal_year",
      series      = col,
      value       = suppressWarnings(as.numeric(data[[col]])),
      unit        = unit,
      metric_type = metric
    )
  }))
  result[!is.na(result$value), ]
}

# Parse "Receipts (£bn)" sheet in tidy long format.
# Row 4 = series names; data starts at row 7.
parse_receipts_bn <- function(path) {
  sheet <- resolve_sheet_name(
    path,
    primary          = "Receipts (\u00a3bn)",
    fallback_pattern = "^Receipts \\(.{1,2}bn\\)$"
  )
  raw <- readxl::read_excel(path, sheet = sheet,
                             col_names = FALSE, .name_repair = "minimal")

  series_names    <- as.character(unlist(raw[4, ]))
  series_names[1] <- "year"
  series_names[series_names %in% c("NA", "NULL")] <- NA
  series_names    <- strip_footnote(series_names)
  series_names[series_names == ""] <- NA

  keep <- !is.na(series_names)
  data <- as.data.frame(raw[7:nrow(raw), keep], stringsAsFactors = FALSE)
  names(data) <- series_names[keep]

  data <- data[grepl("^[0-9]{4}-[0-9]{2}", data$year), ]

  value_cols <- setdiff(names(data), "year")
  result <- do.call(rbind, lapply(value_cols, function(col) {
    obr_long(
      period      = data$year,
      period_type = "fiscal_year",
      series      = col,
      value       = suppressWarnings(as.numeric(data[[col]])),
      unit        = "gbp_bn",
      metric_type = "level"
    )
  }))
  result[!is.na(result$value), ]
}

# Internal: build an obr_tbl from a parsed PFD frame and a source descriptor.
pfd_obr_tbl <- function(data, src) {
  new_obr_tbl(
    data        = data,
    publication = "PFD",
    vintage     = obr_url_vintage(src$final_url %||% src$url),
    source_url  = src$url,
    retrieved   = src$retrieved,
    file_md5    = src$file_md5
  )
}

#' Get all Public Finances Databank aggregates
#'
#' Downloads (and caches) the OBR Public Finances Databank and returns all
#' aggregate fiscal series in tidy long format. Covers outturn from 1946-47
#' and OBR projections through the current forecast horizon.
#'
#' Series include: Public sector net borrowing, Public sector net debt, Total
#' managed expenditure, Public sector current receipts, Nominal GDP, GDP
#' deflator, and more.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy exists.
#'   Defaults to `FALSE`.
#'
#' @return An `obr_tbl` (a `data.frame` with attached provenance) with the
#' standard v0.4.0 schema (columns: `period`, `period_type`, `series`,
#' `metric_type`, `value`, `unit`):
#' \describe{
#'   \item{period}{Fiscal year (character, e.g. `"2024-25"`)}
#'   \item{period_type}{Always `"fiscal_year"` for this function}
#'   \item{series}{Series name (character)}
#'   \item{metric_type}{Usually `"level"`; ratio or index series get a more
#'     specific value derived from the series name}
#'   \item{value}{Numeric value in units described by `unit`}
#'   \item{unit}{Usually `"gbp_bn"`; ratios and indices override}
#' }
#' Use [obr_provenance()] to extract source URL, vintage, and retrieval time.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   pf <- get_public_finances()
#'   unique(pf$series)
#'   obr_provenance(pf)
#' })
#' options(op)
#' }
#'
#' @family public finances
#' @export
get_public_finances <- function(refresh = FALSE) {
  src <- pfd_source(refresh)
  pfd_obr_tbl(parse_aggregates_bn(src$path), src)
}

#' Get Public Sector Net Borrowing (PSNB)
#'
#' Downloads (and caches) the OBR Public Finances Databank and returns
#' annual Public Sector Net Borrowing in \enc{£}{GBP} billion. A positive
#' value means the government is borrowing (deficit); a negative value means
#' a surplus.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy exists.
#'   Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema. `series` is
#' `"PSNB"`, `metric_type` is `"level"`, `unit` is `"gbp_bn"`. See
#' [get_public_finances()] for column definitions.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   psnb <- get_psnb()
#'   tail(psnb)
#' })
#' options(op)
#' }
#'
#' @family public finances
#' @export
get_psnb <- function(refresh = FALSE) {
  src <- pfd_source(refresh)
  agg <- parse_aggregates_bn(src$path)
  out <- agg[agg$series == "Public sector net borrowing", ]
  out$series <- "PSNB"
  rownames(out) <- NULL
  pfd_obr_tbl(out, src)
}

#' Get Public Sector Net Debt (PSND)
#'
#' Downloads (and caches) the OBR Public Finances Databank and returns
#' annual Public Sector Net Debt in \enc{£}{GBP} billion.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy exists.
#'   Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema. `series` is
#' `"PSND"`, `metric_type` is `"level"`, `unit` is `"gbp_bn"`. See
#' [get_public_finances()] for column definitions.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   psnd <- get_psnd()
#'   tail(psnd)
#' })
#' options(op)
#' }
#'
#' @family public finances
#' @export
get_psnd <- function(refresh = FALSE) {
  src <- pfd_source(refresh)
  agg <- parse_aggregates_bn(src$path)
  out <- agg[agg$series == "Public sector net debt", ]
  out$series <- "PSND"
  rownames(out) <- NULL
  pfd_obr_tbl(out, src)
}

#' Get Total Managed Expenditure
#'
#' Downloads (and caches) the OBR Public Finances Databank and returns
#' annual Total Managed Expenditure (TME) in \enc{£}{GBP} billion. TME is
#' the broadest measure of UK government spending.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy exists.
#'   Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema. `series` is
#' `"TME"`, `metric_type` is `"level"`, `unit` is `"gbp_bn"`. See
#' [get_public_finances()] for column definitions.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   tme <- get_expenditure()
#'   tail(tme)
#' })
#' options(op)
#' }
#'
#' @family public finances
#' @export
get_expenditure <- function(refresh = FALSE) {
  src <- pfd_source(refresh)
  agg <- parse_aggregates_bn(src$path)
  out <- agg[agg$series == "Total managed expenditure", ]
  out$series <- "TME"
  rownames(out) <- NULL
  pfd_obr_tbl(out, src)
}

#' Get public sector receipts by tax type
#'
#' Downloads (and caches) the OBR Public Finances Databank and returns
#' public sector current receipts broken down by individual tax type,
#' in tidy long format. Coverage begins in 1999-00.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy exists.
#'   Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema (columns:
#' `period`, `period_type`, `series`, `metric_type`, `value`, `unit`).
#' `series` is the tax or receipt category, `metric_type` is `"level"`,
#' `unit` is `"gbp_bn"`. See [get_public_finances()] for full column docs.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   receipts <- get_receipts()
#'   receipts[grepl("income tax", receipts$series, ignore.case = TRUE), ]
#' })
#' options(op)
#' }
#'
#' @family public finances
#' @export
get_receipts <- function(refresh = FALSE) {
  src <- pfd_source(refresh)
  pfd_obr_tbl(parse_receipts_bn(src$path), src)
}

# v0.4.0 standard tidy long schema for all data-fetching functions.
v04_long_cols <- c("period", "period_type", "series",
                   "metric_type", "value", "unit")

test_that("get_psnb() returns the v0.4.0 schema with series='PSNB'", {
  skip_on_cran()
  skip_if_offline()

  result <- get_psnb()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_type(result$period, "character")
  expect_type(result$value,  "double")
  expect_gt(nrow(result), 50)
  expect_true(all(grepl("^[0-9]{4}-[0-9]{2}$", result$period)))
  expect_true(all(result$series == "PSNB"))
  expect_true(all(result$period_type == "fiscal_year"))
  expect_true(all(result$metric_type == "level"))
  expect_true(all(result$unit == "gbp_bn"))
})

test_that("get_psnd() returns the v0.4.0 schema with series='PSND'", {
  skip_on_cran()
  skip_if_offline()

  result <- get_psnd()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_type(result$value, "double")
  expect_gt(nrow(result), 30)
  expect_true(all(result$series == "PSND"))
})

test_that("get_expenditure() returns the v0.4.0 schema with series='TME'", {
  skip_on_cran()
  skip_if_offline()

  result <- get_expenditure()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_type(result$value, "double")
  expect_gt(nrow(result), 50)
  expect_true(all(result$series == "TME"))
})

test_that("get_receipts() returns the v0.4.0 schema with multiple tax series", {
  skip_on_cran()
  skip_if_offline()

  result <- get_receipts()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_type(result$series, "character")
  expect_type(result$value,  "double")
  expect_gt(length(unique(result$series)), 10)
  expect_true(any(grepl("income tax", result$series, ignore.case = TRUE)))
  expect_true(any(grepl("VAT",        result$series, ignore.case = TRUE)))
  expect_true(all(result$unit == "gbp_bn"))
})

test_that("get_public_finances() returns the v0.4.0 schema with all aggregates", {
  skip_on_cran()
  skip_if_offline()

  result <- get_public_finances()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  series <- unique(result$series)
  expect_true("Public sector net borrowing" %in% series)
  expect_true("Public sector net debt"      %in% series)
  expect_true("Total managed expenditure"   %in% series)
})

test_that("COVID-19 spike visible in PSNB data", {
  skip_on_cran()
  skip_if_offline()

  psnb <- get_psnb()
  covid_year  <- psnb$value[psnb$period == "2020-21"]
  normal_year <- psnb$value[psnb$period == "2018-19"]

  expect_gt(covid_year, normal_year * 5)
})

test_that("PFD-backed functions return obr_tbl with PFD provenance", {
  skip_on_cran()
  skip_if_offline()

  for (fn in list(get_psnb, get_psnd, get_expenditure,
                  get_receipts, get_public_finances)) {
    res  <- fn()
    expect_s3_class(res, "obr_tbl")
    prov <- obr_provenance(res)
    expect_equal(prov$publication, "PFD")
    expect_match(prov$source_url, "obr\\.uk")
    expect_true(inherits(prov$retrieved, "POSIXt"))
    expect_match(prov$file_md5, "^[a-f0-9]{32}$")
    expect_match(prov$package_version, "^[0-9]+\\.[0-9]+\\.[0-9]+")
  }
})

# EFO Detailed Forecast Tables
# URLs are resolved dynamically; the fallback is the latest known good URL
# at package release. Silent fallbacks were removed in 0.3.0; the resolver
# warns when it falls through.
EFO_AGGREGATES_FALLBACK <- "https://obr.uk/download/march-2026-economic-and-fiscal-outlook-detailed-forecast-tables-aggregates/"
EFO_AGGREGATES_FILENAME <- "efo_aggregates.xlsx"
EFO_ECONOMY_FALLBACK    <- "https://obr.uk/download/march-2026-economic-and-fiscal-outlook-detailed-forecast-tables-economy/"
EFO_ECONOMY_FILENAME    <- "efo_economy.xlsx"

efo_aggregates_source <- function(refresh = FALSE, vintage = NULL) {
  vintage <- resolve_efo_vintage(vintage)
  if (!is.null(vintage)) {
    return(efo_pinned_source(
      vintage  = vintage,
      suffix   = "detailed-forecast-tables-aggregates",
      stem     = "efo_aggregates",
      refresh  = refresh
    ))
  }
  obr_get_xlsx(
    candidates = efo_url_candidates("detailed-forecast-tables-aggregates"),
    fallback   = EFO_AGGREGATES_FALLBACK,
    filename   = EFO_AGGREGATES_FILENAME,
    refresh    = refresh,
    label      = "EFO Aggregates"
  )
}

efo_economy_source <- function(refresh = FALSE, vintage = NULL) {
  vintage <- resolve_efo_vintage(vintage)
  if (!is.null(vintage)) {
    return(efo_pinned_source(
      vintage  = vintage,
      suffix   = "detailed-forecast-tables-economy",
      stem     = "efo_economy",
      refresh  = refresh
    ))
  }
  obr_get_xlsx(
    candidates = efo_url_candidates("detailed-forecast-tables-economy"),
    fallback   = EFO_ECONOMY_FALLBACK,
    filename   = EFO_ECONOMY_FILENAME,
    refresh    = refresh,
    label      = "EFO Economy"
  )
}

# Download an EFO file pinned to a specific vintage. No URL probing: the
# vintage table determines the slug. Cache filenames carry the vintage tag
# so different vintages do not overwrite each other.
efo_pinned_source <- function(vintage, suffix, stem, refresh = FALSE) {
  url <- efo_url_for_vintage(vintage, suffix)
  filename <- sprintf("%s_%s.xlsx", stem, vintage_cache_tag(vintage))
  path <- obr_fetch(url, filename, refresh = refresh)
  list(
    path      = path,
    url       = url,
    final_url = url,
    source    = "pinned",
    retrieved = tryCatch(file.info(path)$mtime, error = function(e) Sys.time()),
    file_md5  = tryCatch(unname(tools::md5sum(path)), error = function(e) NA_character_)
  )
}

# Build an obr_tbl from a parsed EFO frame and a source descriptor.
efo_obr_tbl <- function(data, src) {
  new_obr_tbl(
    data        = data,
    publication = "EFO",
    vintage     = obr_url_vintage(src$url),
    source_url  = src$url,
    retrieved   = src$retrieved,
    file_md5    = src$file_md5
  )
}

# Parse sheet 6.5 (Components of Net Borrowing) from EFO aggregates file.
# Row 5 has fiscal year labels in columns that contain year-like strings.
# Data rows are those where col 2 is non-NA and at least one value is numeric.
# Returns the v0.4.0 schema: period, period_type, series, metric_type, value, unit.
parse_efo_fiscal <- function(path) {
  raw <- readxl::read_excel(path, sheet = "6.5",
                            col_names = FALSE, .name_repair = "minimal")

  all_row5 <- as.character(unlist(raw[5, ]))
  year_cols <- which(grepl("^[0-9]{4}-[0-9]{2}$", all_row5))
  fiscal_years <- all_row5[year_cols]

  col2 <- as.character(unlist(raw[, 2]))

  result_list <- list()
  for (i in seq_len(nrow(raw))) {
    nm <- col2[i]
    if (is.na(nm) || nm == "") next
    vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[i, year_cols])))
    )
    if (all(is.na(vals))) next
    result_list[[length(result_list) + 1]] <- obr_long(
      period      = fiscal_years,
      period_type = "fiscal_year",
      series      = nm,
      value       = vals,
      unit        = "gbp_bn",
      metric_type = "level"
    )
  }

  result <- do.call(rbind, result_list)
  result[!is.na(result$value), ]
}

# Generic parser for EFO economy sheets (quarterly, wide format).
# Finds the first row where col 2 has a quarterly period (e.g. "2008Q1"),
# then takes the row immediately before it as the series-name header.
#
# Returns the v0.4.0 schema: period, period_type, series, metric_type, value, unit.
#
# metric_type is set per series, in order:
#   1. heuristic classifier: detects "Index" / "deflator" → "index",
#      "rate"/"share" → "pct", "growth"/"inflation" → "yoy_pct"
#   2. for series the classifier returns "level" on, the caller's
#      `default_metric_type` is applied (e.g. "yoy_pct" for the inflation sheet,
#      where bare names like "CPI" / "RPI" denote annual rates by convention)
#
# This solves the v0.3.x bug where (e.g.) CPI YoY values and any CPI Index
# values lived in the same `value` column with no machine-readable distinction.
# The per-measure default also fixes the related issue that bare names like
# "CPI" do not contain "inflation" / "growth" / "rate" in the source sheet,
# so the classifier alone cannot tell what units they are in.
parse_efo_economy_sheet <- function(path, sheet,
                                    default_metric_type = NULL,
                                    default_unit        = NA_character_) {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  col2 <- as.character(unlist(raw[, 2]))
  is_period <- grepl("^[0-9]{4}Q[1-4]$", col2)
  first_data_row <- which(is_period)[1]
  if (is.na(first_data_row)) return(NULL)

  header_row <- NA
  for (i in (first_data_row - 1):1) {
    v <- as.character(unlist(raw[i, 3]))
    if (!is.na(v) && is.na(suppressWarnings(as.numeric(v)))) {
      header_row <- i
      break
    }
  }
  if (is.na(header_row)) return(NULL)

  series <- trimws(gsub("\r\n", " ",
                        as.character(unlist(raw[header_row, 3:ncol(raw)]))))
  valid_series <- !is.na(series) & series != ""

  data_idx <- which(is_period)
  periods  <- col2[data_idx]

  metric <- classify_metric_type(series)
  if (!is.null(default_metric_type)) {
    metric[!is.na(metric) & metric == "level"] <- default_metric_type
  }
  unit <- default_unit_for_metric(metric)
  if (!is.na(default_unit)) {
    unit[is.na(unit)] <- default_unit
  }

  result_list <- list()
  for (j in which(valid_series)) {
    col_idx <- j + 2L
    if (col_idx > ncol(raw)) next
    vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[data_idx, col_idx])))
    )
    if (all(is.na(vals))) next
    result_list[[length(result_list) + 1]] <- obr_long(
      period      = periods,
      period_type = "quarter",
      series      = series[j],
      value       = vals,
      unit        = unit[j],
      metric_type = metric[j]
    )
  }

  if (length(result_list) == 0) return(NULL)
  result <- do.call(rbind, result_list)
  result[!is.na(result$value), ]
}

# Special-case parser for sheet 1.14 (output gap).
# Fix in 0.3.0: locate the value column by scanning each column for the one
# containing the most numeric entries aligned with the quarterly periods,
# rather than hardcoding column 3. This protects us from OBR moving the
# column or inserting a leading column.
parse_efo_output_gap <- function(path) {
  raw <- readxl::read_excel(path, sheet = "1.14",
                            col_names = FALSE, .name_repair = "minimal")
  col2 <- as.character(unlist(raw[, 2]))
  is_period <- grepl("^[0-9]{4}Q[1-4]$", col2)
  data_idx  <- which(is_period)
  if (length(data_idx) == 0L) return(NULL)

  best_col <- NA_integer_
  best_n   <- -1L
  for (j in 3:ncol(raw)) {
    vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[data_idx, j])))
    )
    n <- sum(!is.na(vals))
    if (n > best_n) {
      best_n   <- n
      best_col <- j
    }
  }
  if (is.na(best_col)) return(NULL)

  obr_long(
    period      = col2[data_idx],
    period_type = "quarter",
    series      = "Output gap",
    value       = suppressWarnings(
      as.numeric(as.character(unlist(raw[data_idx, best_col])))
    ),
    unit        = "pct",
    metric_type = "pct"
  )
}

#' List available EFO economy measures
#'
#' Returns a data frame of the economy measures available via
#' [get_efo_economy()], showing the `measure` name to pass and a
#' short description of what each covers.
#'
#' @return A data frame with columns `measure`, `sheet`, and `description`.
#'
#' @examples
#' list_efo_economy_measures()
#'
#' @family EFO
#' @export
list_efo_economy_measures <- function() {
  data.frame(
    measure = c("labour", "inflation", "output_gap"),
    sheet   = c("1.6", "1.7", "1.14"),
    description = c(
      "Labour market: employment, unemployment rate, participation rate, hours worked",
      "Inflation: CPI, CPIH, RPI, RPIX, mortgage rates, rents",
      "OBR central estimate of the output gap (% of potential output)"
    ),
    stringsAsFactors = FALSE
  )
}

#' Get EFO fiscal projections (net borrowing components)
#'
#' Downloads (and caches) the OBR \emph{Economic and Fiscal Outlook} Detailed
#' Forecast Tables - Aggregates file and returns the components of net
#' borrowing (Table 6.5) in tidy long format.
#'
#' Covers the five-year forecast horizon published at the most recent
#' fiscal event. Key series include current receipts, current expenditure,
#' depreciation, net investment, and net borrowing (PSNB). The exact
#' vintage is recorded in the returned object's provenance attribute and
#' visible in the printed header.
#'
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#' @param vintage Optional EFO vintage label such as `"October 2024"`. If
#'   supplied, the function downloads the file for that specific EFO. If
#'   `NULL` (the default), the function uses any vintage set via [obr_pin()],
#'   or falls back to the latest live EFO via the dynamic URL resolver. See
#'   [obr_efo_vintages()] for the full list of supported vintages.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema (columns:
#' `period`, `period_type`, `series`, `metric_type`, `value`, `unit`):
#' \describe{
#'   \item{period}{Fiscal year being forecast, e.g. `"2025-26"` (character)}
#'   \item{period_type}{Always `"fiscal_year"` for this function (character)}
#'   \item{series}{Component name, e.g. `"Net borrowing"` (character)}
#'   \item{metric_type}{Always `"level"` for this function (character)}
#'   \item{value}{Projected value (numeric)}
#'   \item{unit}{Always `"gbp_bn"` for this function (character)}
#' }
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   efo <- get_efo_fiscal()
#'   efo[efo$series == "Net borrowing", ]
#'   obr_provenance(efo)$vintage
#'
#'   # Pin to a specific EFO for reproducibility
#'   october_2024 <- get_efo_fiscal(vintage = "October 2024")
#' })
#' options(op)
#' }
#'
#' @family EFO
#' @export
get_efo_fiscal <- function(refresh = FALSE, vintage = NULL) {
  # v0.5.0: thin wrapper over the generic dispatcher.
  get_efo_table("6.5", vintage = vintage, refresh = refresh)
}

#' Get EFO economy projections
#'
#' Downloads (and caches) the OBR \emph{Economic and Fiscal Outlook} Detailed
#' Forecast Tables - Economy file and returns quarterly economic projections
#' for a chosen measure in tidy long format.
#'
#' Data run from 2008 Q1 through the current forecast horizon. Use
#' [list_efo_economy_measures()] to see all available measures.
#'
#' @param measure Character. Which economy table to return. One of
#'   `"inflation"`, `"labour"`, or `"output_gap"`. Defaults to
#'   `"inflation"`.
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#' @param vintage Optional EFO vintage label such as `"October 2024"`. If
#'   supplied, the function downloads the file for that specific EFO. If
#'   `NULL` (the default), the function uses any vintage set via [obr_pin()],
#'   or falls back to the latest live EFO via the dynamic URL resolver.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema (columns:
#' `period`, `period_type`, `series`, `metric_type`, `value`, `unit`):
#' \describe{
#'   \item{period}{Calendar quarter, e.g. `"2025Q1"` (character)}
#'   \item{period_type}{Always `"quarter"` for this function (character)}
#'   \item{series}{Variable name, e.g. `"CPI"` (character)}
#'   \item{metric_type}{One of `"index"`, `"yoy_pct"`, `"pct"`, `"level"`,
#'     classified from the series name. This is the v0.4.0 fix for the v0.3.x
#'     issue where, e.g., CPI Index values and CPI YoY values shared a single
#'     `value` column with no machine-readable distinction.}
#'   \item{value}{Numeric value in units described by `unit`}
#'   \item{unit}{One of `"index"`, `"pct"`, etc., paired with `metric_type`}
#' }
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   inf <- get_efo_economy("inflation")
#'   inf[inf$series == "CPI", ]
#'
#'   lab <- get_efo_economy("labour")
#'
#'   # Compare CPI projections from two different EFOs
#'   inf_oct24 <- get_efo_economy("inflation", vintage = "October 2024")
#'   inf_mar26 <- get_efo_economy("inflation", vintage = "March 2026")
#' })
#' options(op)
#' }
#'
#' @family EFO
#' @export
get_efo_economy <- function(measure = c("inflation", "labour", "output_gap"),
                            refresh = FALSE,
                            vintage = NULL) {
  # v0.5.0: thin wrapper over the generic dispatcher. Per-measure metric
  # defaults (e.g. inflation = yoy_pct/pct) live in the catalogue.
  measure <- match.arg(measure)
  table_id <- switch(measure,
                     "inflation"  = "1.7",
                     "labour"     = "1.6",
                     "output_gap" = "1.14")
  get_efo_table(table_id, vintage = vintage, refresh = refresh)
}

test_that("list_efo_economy_measures() works without network", {
  result <- list_efo_economy_measures()

  expect_s3_class(result, "data.frame")
  expect_named(result, c("measure", "sheet", "description"))
  expect_equal(nrow(result), 3)
  expect_true("inflation"  %in% result$measure)
  expect_true("labour"     %in% result$measure)
  expect_true("output_gap" %in% result$measure)
})

test_that("get_efo_economy() errors on invalid measure", {
  expect_error(
    get_efo_economy("NOT_A_MEASURE"),
    regexp = "should be one of"
  )
})

# v0.4.0 schema: period, period_type, series, metric_type, value, unit.
v04_long_cols <- c("period", "period_type", "series",
                   "metric_type", "value", "unit")

test_that("get_efo_fiscal() returns the v0.4.0 long schema", {
  skip_on_cran()
  skip_if_offline()

  result <- get_efo_fiscal()

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_type(result$period,      "character")
  expect_type(result$period_type, "character")
  expect_type(result$series,      "character")
  expect_type(result$metric_type, "character")
  expect_type(result$value,       "double")
  expect_type(result$unit,        "character")
  expect_gt(nrow(result), 10)
  expect_true("Net borrowing"    %in% result$series)
  expect_true("Current receipts" %in% result$series)
  expect_true(all(result$period_type == "fiscal_year"))
  expect_true(all(result$metric_type == "level"))
  expect_true(all(result$unit == "gbp_bn"))
})

test_that("get_efo_fiscal() covers 5-year forecast horizon", {
  skip_on_cran()
  skip_if_offline()

  result <- get_efo_fiscal()
  years  <- unique(result$period)

  expect_gte(length(years), 5)
  expect_true(all(grepl("^[0-9]{4}-[0-9]{2}$", years)))
})

test_that("get_efo_economy('inflation') returns the v0.4.0 schema", {
  skip_on_cran()
  skip_if_offline()

  result <- get_efo_economy("inflation")

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_gt(nrow(result), 200)
  expect_true(any(grepl("CPI",  result$series)))
  expect_true(any(grepl("CPIH", result$series)))
  expect_true(all(grepl("^[0-9]{4}Q[1-4]$", result$period)))
  expect_true(all(result$period_type == "quarter"))
})

test_that("get_efo_economy('inflation') splits Index vs YoY into metric_type", {
  # This is the v0.4.0 fix for OBR feedback (Ben Northcott, 2026-04-29):
  # previously, CPI Index values (~135) and CPI YoY values (~2.1) sat in the
  # same `value` column with no machine-readable distinction. Now they are
  # tagged via metric_type ("index" vs "yoy_pct") and unit ("index" vs "pct").
  skip_on_cran()
  skip_if_offline()

  result <- get_efo_economy("inflation")
  types  <- unique(result$metric_type)

  expect_true(any(types %in% c("index", "yoy_pct", "pct", "level")))
  # Index rows and YoY rows should each be filterable independently
  if ("index" %in% types) {
    idx <- result[result$metric_type == "index", ]
    expect_gt(nrow(idx), 0)
    expect_true(all(idx$unit == "index"))
  }
  if ("yoy_pct" %in% types) {
    yoy <- result[result$metric_type == "yoy_pct", ]
    expect_gt(nrow(yoy), 0)
    expect_true(all(yoy$unit == "pct"))
  }
})

test_that("get_efo_economy('labour') returns the v0.4.0 schema", {
  skip_on_cran()
  skip_if_offline()

  result <- get_efo_economy("labour")

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_gt(nrow(result), 200)
  expect_true(any(grepl("Employment", result$series)))
  expect_true(any(grepl("unemployment", result$series, ignore.case = TRUE)))
  expect_true(all(result$period_type == "quarter"))
})

test_that("get_efo_economy('output_gap') returns the v0.4.0 schema", {
  skip_on_cran()
  skip_if_offline()

  result <- get_efo_economy("output_gap")

  expect_s3_class(result, "data.frame")
  expect_named(result, v04_long_cols)
  expect_gt(nrow(result), 100)
  expect_true(all(result$series == "Output gap"))
  expect_true(any(grepl("^197", result$period)))
  expect_true(all(result$period_type == "quarter"))
  expect_true(all(result$metric_type == "pct"))
  expect_true(all(result$unit == "pct"))
})

test_that("EFO functions return obr_tbl with EFO provenance", {
  skip_on_cran()
  skip_if_offline()

  for (call in list(quote(get_efo_fiscal()),
                    quote(get_efo_economy("inflation")),
                    quote(get_efo_economy("output_gap")))) {
    res  <- eval(call)
    expect_s3_class(res, "obr_tbl")
    prov <- obr_provenance(res)
    expect_equal(prov$publication, "EFO")
    expect_match(prov$source_url, "economic-and-fiscal-outlook")
    expect_match(prov$vintage, "^[A-Z][a-z]+ [0-9]{4}$")
  }
})

v04_long_cols <- c("period", "period_type", "series",
                   "metric_type", "value", "unit")

# ----- Catalogue -----------------------------------------------------------

test_that("obr_efo_catalogue() returns a populated catalogue", {
  cat <- obr_efo_catalogue()
  expect_s3_class(cat, "data.frame")
  expect_named(cat, c("table_id", "file", "section", "title",
                      "layout", "default_metric_type", "default_unit"))
  # 17 aggregates + 22 economy = 39 rows
  expect_gte(nrow(cat), 39)
  expect_true("6.5" %in% cat$table_id)
  expect_true("1.7" %in% cat$table_id)
  expect_true("1.14" %in% cat$table_id)
  expect_true(all(cat$file %in% c("aggregates", "economy")))
  expect_true(all(cat$layout %in%
                    c("quarterly_wide", "quarterly_single",
                      "annual_year_wide", "annual_period_wide",
                      "fiscal_year_wide",
                      "subsector_matrix", "quarterly_indented",
                      "cross_reference")))
})

# ----- Argument validation -------------------------------------------------

test_that("get_efo_table() errors on unknown table id", {
  expect_error(get_efo_table("99.99"), regexp = "Unknown EFO table id")
})

test_that("get_efo_table() errors on non-string table id", {
  expect_error(get_efo_table(c("6.5", "1.7")), regexp = "single character")
  expect_error(get_efo_table(NULL),             regexp = "single character")
})

# ----- Cross-reference and complex-layout sheets ---------------------------

test_that("get_efo_table() follows cross-references to a previous EFO", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("6.11")
  expect_s3_class(out, "obr_tbl")
  expect_named(out, v04_long_cols)
  prov <- obr_provenance(out)
  # Resolves to a previous EFO vintage, not the current one
  expect_match(prov$vintage, "^[A-Z][a-z]+ [0-9]{4}$")
  expect_match(prov$notes, "Cross-reference")
})

test_that("get_efo_table('6.4') returns subsector_matrix shape", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("6.4")
  expect_s3_class(out, "obr_tbl")
  expect_true("sub_sector" %in% names(out))
  expect_true(all(c(v04_long_cols, "sub_sector") %in% names(out)))
  expect_true("Central government" %in% out$sub_sector)
  expect_true(all(out$period_type == "fiscal_year"))
  expect_true(all(out$unit == "gbp_bn"))
})

test_that("get_efo_table('6.10') returns quarterly_indented data", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("6.10")
  expect_s3_class(out, "obr_tbl")
  expect_named(out, v04_long_cols)
  expect_true(all(out$period_type == "quarter"))
  expect_true(all(grepl("^[0-9]{4}Q[1-4]$", out$period)))
  expect_true(all(out$unit == "pct"))
})

# ----- Each layout returns the standard schema -----------------------------

test_that("get_efo_table('6.5') returns standard schema (fiscal_year_wide)", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("6.5")
  expect_s3_class(out, "obr_tbl")
  expect_named(out, v04_long_cols)
  expect_true(all(out$period_type == "fiscal_year"))
  expect_true(all(out$unit == "gbp_bn"))
  expect_true("Net borrowing" %in% out$series)
})

test_that("get_efo_table('1.7') returns standard schema (quarterly_wide)", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("1.7")
  expect_s3_class(out, "obr_tbl")
  expect_named(out, v04_long_cols)
  expect_true(all(out$period_type == "quarter"))
  expect_true(all(out$metric_type == "yoy_pct"))
  expect_true(all(out$unit == "pct"))
  expect_true("CPI" %in% out$series)
})

test_that("get_efo_table('1.4') returns standard schema (quarterly_single)", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("1.4")
  expect_s3_class(out, "obr_tbl")
  expect_named(out, v04_long_cols)
  expect_true(all(out$period_type == "quarter"))
  expect_true(all(out$series == "Nominal GDP"))
})

test_that("get_efo_table('1.13') returns standard schema (annual_year_wide)", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("1.13")
  expect_s3_class(out, "obr_tbl")
  expect_named(out, v04_long_cols)
  expect_true(all(out$period_type == "calendar_year"))
})

test_that("get_efo_table('1.19b') returns standard schema (annual_period_wide)", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("1.19b")
  expect_s3_class(out, "obr_tbl")
  expect_named(out, v04_long_cols)
  expect_true(all(out$period_type == "calendar_year"))
})

# ----- Backward compatibility: headline wrappers still work ----------------

test_that("get_efo_fiscal() == get_efo_table('6.5')", {
  skip_on_cran()
  skip_if_offline()
  via_fn  <- get_efo_fiscal()
  via_id  <- get_efo_table("6.5")
  # Same data shape, same series, same values
  expect_identical(names(via_fn), names(via_id))
  expect_equal(nrow(via_fn), nrow(via_id))
  expect_setequal(via_fn$series, via_id$series)
})

test_that("get_efo_economy('inflation') == get_efo_table('1.7')", {
  skip_on_cran()
  skip_if_offline()
  via_fn <- get_efo_economy("inflation")
  via_id <- get_efo_table("1.7")
  expect_identical(names(via_fn), names(via_id))
  expect_equal(nrow(via_fn), nrow(via_id))
})

test_that("get_efo_economy('output_gap') keeps series='Output gap'", {
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_economy("output_gap")
  expect_true(all(out$series == "Output gap"))
})

# ----- Index-linked gilts edge case (the v0.5.0 classifier fix) ------------

test_that("Net debt composition (6.13) classifies all rows as pct", {
  # Regression test for the v0.5.0 classifier fix. "Index-linked gilts" was
  # being misclassified as metric_type=index because the bare \\bindex\\b
  # pattern matched. Now the pattern requires Index at end of string.
  skip_on_cran()
  skip_if_offline()
  out <- get_efo_table("6.13")
  expect_true(all(out$metric_type == "pct"))
  expect_true(all(out$unit == "pct"))
  expect_true("Index-linked gilts" %in% out$series)
})

# Try a sequence of OBR download URLs until one responds with status < 400.
# Returns a list with: $url        (the working candidate URL or NA),
#                      $final_url  (URL after redirects, used to recover the
#                                  publication vintage when the slug is stable),
#                      $source     ("live" if a candidate worked, "none" if not).
# OBR URLs follow the pattern: https://obr.uk/download/{month}-{year}-{publication}/
# Stable slugs (like public-finances-databank/) redirect to the latest file;
# capturing the redirect target lets us recover the vintage.
obr_resolve_url <- function(url_candidates) {
  for (url in url_candidates) {
    # GET (the OBR's WordPress endpoints don't reliably support HEAD).
    # Redirects are followed so the final URL can be recovered for
    # publications that use a stable slug.
    resp <- tryCatch(
      httr2::request(url) |>
        httr2::req_user_agent("obr R package (https://github.com/charlescoverdale/obr)") |>
        httr2::req_throttle(rate = 5 / 10) |>
        # The OBR CDN sometimes 403s bursts of probes (each vintage probe is
        # a fresh GET). Those are rate-limit responses, not missing files, so
        # back off and retry them; genuine 404s fail fast.
        httr2::req_retry(
          max_tries    = 3,
          # Cap the total time spent retrying. On a machine the CDN refuses
          # outright, every probe would otherwise sit through the full
          # backoff before failing, and a vintage lookup probes several
          # URLs. Measured against a always-403 endpoint, a refused request
          # costs 15.2s uncapped and 6.6s at this setting. Caps of 8 or
          # above make no difference: the natural 2+4+8 backoff finishes
          # inside them.
          max_seconds  = 5,
          is_transient = function(resp) {
            httr2::resp_status(resp) %in% c(403L, 429L, 503L)
          },
          backoff      = function(i) min(10, 2 ^ i)
        ) |>
        httr2::req_error(is_error = function(resp) FALSE) |>
        httr2::req_perform(),
      error = function(e) NULL
    )
    if (!is.null(resp) && httr2::resp_status(resp) < 400L) {
      # Soft-404 guard: a candidate slug that resolves to an HTML page is a
      # WordPress error/landing page, not a data file. Only accept responses
      # whose content type is not text/html.
      ctype <- tryCatch(httr2::resp_content_type(resp),
                        error = function(e) NA_character_)
      if (!is.na(ctype) && grepl("text/html", ctype, fixed = TRUE)) next
      final_url <- tryCatch(httr2::resp_url(resp), error = function(e) url)
      return(list(url = url, final_url = final_url, source = "live"))
    }
  }
  list(url = NA_character_, final_url = NA_character_, source = "none")
}

# Build EFO URL candidates for recent publication cycles (most recent first).
# Months are ordered latest-in-year first so that, once an autumn EFO
# publishes, it wins over that year's spring EFO. With march probed first
# (the pre-0.6.0 order), the resolver kept returning the spring EFO for the
# rest of the year: exactly the failure mode a Budget-day user would hit.
efo_url_candidates <- function(suffix) {
  current_year <- as.integer(format(Sys.Date(), "%Y"))
  months <- c("november", "october", "march")
  candidates <- character(0)
  for (yr in seq(current_year, current_year - 2L)) {
    for (mn in months) {
      candidates <- c(candidates, paste0(
        "https://obr.uk/download/", mn, "-", yr,
        "-economic-and-fiscal-outlook-", suffix, "/"
      ))
    }
  }
  candidates
}

# Build WTR URL candidates.
wtr_url_candidates <- function() {
  current_year <- as.integer(format(Sys.Date(), "%Y"))
  months <- c("october", "june", "march")
  candidates <- character(0)
  for (yr in seq(current_year, current_year - 2L)) {
    for (mn in months) {
      candidates <- c(candidates, paste0(
        "https://obr.uk/download/welfare-trends-report-", mn, "-", yr,
        "-charts-and-tables/"
      ))
    }
  }
  candidates
}

# Build Historical Official Forecasts Database URL candidates.
# OBR sometimes re-uses an older slug across vintages, so we try recent slugs
# first and fall through to the known-stable slug as a fallback.
forecasts_url_candidates <- function() {
  current_year <- as.integer(format(Sys.Date(), "%Y"))
  months <- c("november", "october", "july", "march")
  candidates <- character(0)
  for (yr in seq(current_year, current_year - 3L)) {
    for (mn in months) {
      candidates <- c(candidates, paste0(
        "https://obr.uk/download/historical-official-forecasts-database-",
        mn, "-", yr, "/"
      ))
    }
  }
  candidates
}

# Cache directory (platform-aware, base R).
obr_cache_dir <- function() {
  d <- getOption("obr.cache_dir", default = tools::R_user_dir("obr", "cache"))
  if (!dir.exists(d)) dir.create(d, recursive = TRUE)
  d
}

# Download a file and cache it; return local path.
obr_fetch <- function(url, filename, refresh = FALSE) {
  if (!is.logical(refresh) || length(refresh) != 1L || is.na(refresh)) {
    cli::cli_abort("{.arg refresh} must be a single {.cls logical} value.")
  }

  path <- file.path(obr_cache_dir(), filename)

  if (file.exists(path) && !refresh) {
    cli::cli_inform(c("i" = "Loading from cache. Use {.code refresh = TRUE} to re-download."))
    return(path)
  }

  cli::cli_inform(c("i" = "Downloading {.file {filename}} from OBR..."))

  resp <- tryCatch(
    httr2::request(url) |>
      httr2::req_user_agent("obr R package (https://github.com/charlescoverdale/obr)") |>
      httr2::req_throttle(rate = 5 / 10) |>
      httr2::req_retry(
        # 403 is included alongside 429/503 because the OBR's CDN occasionally
        # returns 403 to legitimate requests when probing many URLs in quick
        # succession; backing off and retrying typically clears it.
        max_tries    = 4,
        # See the note on the probe request above. Same budget: a larger
        # one measured no better, because the uncapped backoff already
        # completes within it.
        max_seconds  = 5,
        is_transient = function(resp) {
          httr2::resp_status(resp) %in% c(403L, 429L, 503L)
        },
        backoff      = function(i) min(30, 2 ^ i)
      ) |>
      httr2::req_perform(),
    error = function(e) {
      cli::cli_abort(
        c("Failed to download {.url {url}}.",
          "x" = conditionMessage(e)),
        call = NULL
      )
    }
  )

  writeBin(httr2::resp_body_raw(resp), path)
  cli::cli_inform(c("v" = "Saved to cache."))
  path
}

# Resolve URL with fallback, download, return list with full source metadata.
# Raises a clear warning if the live candidates all failed and a fallback was
# used: this is the audit fix for the silent-fallback issue.
obr_get_xlsx <- function(candidates, fallback, filename, refresh = FALSE,
                         label = "OBR publication") {
  resolved <- obr_resolve_url(candidates)
  if (resolved$source == "live") {
    url       <- resolved$url
    final_url <- resolved$final_url
    source    <- "live"
  } else {
    cli::cli_warn(c(
      "Could not resolve a current {label} URL from {length(candidates)} candidate{?s}.",
      "i" = "Falling back to {.url {fallback}}.",
      "!" = "Returned data may be older than expected. Run with internet access, or pin a vintage explicitly when that feature ships."
    ))
    url       <- fallback
    final_url <- fallback
    source    <- "fallback"
  }

  path <- obr_fetch(url, filename, refresh = refresh)

  retrieved <- tryCatch(file.info(path)$mtime,
                        error = function(e) Sys.time())
  md5 <- tryCatch(unname(tools::md5sum(path)),
                  error = function(e) NA_character_)

  list(
    path      = path,
    url       = url,
    final_url = final_url,
    source    = source,
    retrieved = retrieved,
    file_md5  = md5
  )
}

# Tolerant sheet-name resolver. Pass a primary expected name and an optional
# regex fallback; returns the first match found in the file. Errors clearly
# (with the available sheet list) if neither matches.
resolve_sheet_name <- function(path, primary, fallback_pattern = NULL) {
  available <- readxl::excel_sheets(path)
  if (primary %in% available) return(primary)
  if (!is.null(fallback_pattern)) {
    m <- grep(fallback_pattern, available, value = TRUE)
    if (length(m) >= 1L) return(m[1L])
  }
  cli::cli_abort(c(
    "Could not find sheet matching {.val {primary}} in {.file {basename(path)}}.",
    "i" = "Available sheets: {.val {available}}.",
    "!" = "OBR may have changed the sheet name. Please file an issue at https://github.com/charlescoverdale/obr/issues."
  ))
}

# Allowed values for the schema metadata columns.
# These are the controlled vocabularies that all data-returning functions
# populate from 0.4.0 onwards.
OBR_PERIOD_TYPES <- c("fiscal_year", "quarter", "calendar_year", "month")
OBR_METRIC_TYPES <- c("level", "yoy_pct", "index", "pct", "pct_pts")
OBR_UNITS        <- c("gbp_bn", "gbp_mn", "pct", "pct_pts", "index",
                      "count_mn", "count_k", "count", "hours", "ratio")

# Internal: classify a series name into a metric_type.
# Heuristic; returns one of OBR_METRIC_TYPES. Defaults to "level".
# Used so users can tell apart, e.g. CPI Index (135.2) vs CPI YoY (2.1) which
# previously sat in the same `value` column with no machine-readable distinction.
classify_metric_type <- function(series) {
  if (length(series) == 0L) return(character(0))
  s <- tolower(as.character(series))
  vapply(s, function(x) {
    if (is.na(x) || x == "") return(NA_character_)
    # YoY: explicit growth-rate signals only. Plain "change" is too eager
    # because OBR uses it for level-difference series too (e.g. "Adjustment
    # for the change in pension entitlements", which is in £bn).
    if (grepl(paste0("\\binflation\\b|\\bgrowth\\b|",
                     "\\by[/-]y\\b|\\byoy\\b|year[ -]on[ -]year|",
                     "annual\\s*%|%\\s*change"), x)) {
      "yoy_pct"
    # Index: "Index" must be a standalone word at end of string, or an
    # explicit "(2015=100)" / "(2010=100)" base-year tag. "Index-linked
    # gilts" or "Index of X" must NOT match because the values are levels
    # in £bn or % of GDP, not index points.
    } else if (grepl("\\bindex\\s*$|\\(20[12][05][ =]100\\)", x)) {
      "index"
    } else if (grepl("percentage points?|\\bpp\\b", x)) {
      "pct_pts"
    } else if (grepl("\\brate\\b|\\bratio\\b|\\bshare\\b|\\(%\\)|\\bper cent\\b|\\bpercent\\b|\\bpct\\b|%\\s*$", x)) {
      "pct"
    } else {
      "level"
    }
  }, character(1), USE.NAMES = FALSE)
}

# Internal: default unit for a metric_type. For "level" the caller must supply
# the actual unit (gbp_bn, count_mn, etc.) since level can be many things.
default_unit_for_metric <- function(metric_type) {
  vapply(metric_type, function(m) {
    if (is.na(m)) return(NA_character_)
    switch(m,
           "index"   = "index",
           "yoy_pct" = "pct",
           "pct"     = "pct",
           "pct_pts" = "pct_pts",
           NA_character_)
  }, character(1), USE.NAMES = FALSE)
}

# Internal: build a standard tidy-long observation frame using the v0.4.0
# schema. All data-returning functions use this so outputs can be rbind()'d
# across publications.
#
# Columns: period, period_type, series, metric_type, value, unit
obr_long <- function(period, period_type, series, value,
                     unit = NA_character_, metric_type = NULL) {
  if (is.null(metric_type)) {
    metric_type <- classify_metric_type(series)
  }
  if (length(period_type) == 1L) period_type <- rep(period_type, length(period))
  if (length(unit) == 1L)        unit        <- rep(unit,        length(period))
  data.frame(
    period      = as.character(period),
    period_type = as.character(period_type),
    series      = as.character(series),
    metric_type = as.character(metric_type),
    value       = as.numeric(value),
    unit        = as.character(unit),
    stringsAsFactors = FALSE
  )
}

#' Clear cached OBR files
#'
#' Deletes all files downloaded and cached by the obr package. The next
#' function call will re-download fresh data from the OBR website.
#'
#' @return Invisibly returns `NULL`.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' clear_cache()
#' options(op)
#' }
#'
#' @family data access
#' @export
clear_cache <- function() {
  files <- list.files(obr_cache_dir(), full.names = TRUE)
  n <- length(files)
  if (n > 0) file.remove(files)
  cli::cli_inform("Removed {n} cached file{?s}.")
  invisible(NULL)
}

test_that("clear_cache() runs without error", {
  op <- options(obr.cache_dir = tempfile("obr_test_cache_"))
  on.exit(options(op))
  dir.create(getOption("obr.cache_dir"), recursive = TRUE)
  expect_invisible(clear_cache())
})

test_that("obr_cache_dir() creates and returns a directory", {
  op <- options(obr.cache_dir = tempfile("obr_test_dir_"))
  on.exit(options(op))
  d <- obr_cache_dir()
  expect_type(d, "character")
  expect_true(dir.exists(d))
})

test_that("classify_metric_type() splits Index from YoY series names", {
  expect_equal(classify_metric_type("CPI Index"),     "index")
  expect_equal(classify_metric_type("CPIH Index"),    "index")
  expect_equal(classify_metric_type("CPI inflation"), "yoy_pct")
  expect_equal(classify_metric_type("Real GDP growth"), "yoy_pct")
  expect_equal(classify_metric_type("Y/Y change"),    "yoy_pct")
  expect_equal(classify_metric_type("Unemployment rate"), "pct")
  expect_equal(classify_metric_type("Participation rate"), "pct")
  expect_equal(classify_metric_type("Net borrowing"),  "level")
  # "deflator" is intentionally NOT classified as index because OBR sometimes
  # reports it as YoY % change rather than as a level index. Caller (e.g.
  # the inflation sheet parser) supplies a default_metric_type for these.
  expect_equal(classify_metric_type("GDP deflator"),   "level")
  expect_equal(classify_metric_type("Output gap (pp)"), "pct_pts")
  # v0.5.0: tighter rules so EFO Aggregates Section 6 series classify cleanly.
  # "Index-linked gilts" should NOT be index (it's a % of GDP value); "Index"
  # only matches at end of string or with explicit base-year tag.
  expect_equal(classify_metric_type("Index-linked gilts"), "level")
  expect_equal(classify_metric_type("CPI (2015=100)"),     "index")
  # Bare "change" must not trigger yoy_pct: OBR has many "change in X"
  # series that are level differences in £bn.
  expect_equal(
    classify_metric_type("Adjustment for the change in pension entitlements"),
    "level"
  )
  # Vectorised
  expect_equal(
    classify_metric_type(c("CPI Index", "CPI inflation", "Net borrowing")),
    c("index", "yoy_pct", "level")
  )
})

test_that("default_unit_for_metric() maps metric_type to unit", {
  expect_equal(default_unit_for_metric("index"),   "index")
  expect_equal(default_unit_for_metric("yoy_pct"), "pct")
  expect_equal(default_unit_for_metric("pct"),     "pct")
  expect_equal(default_unit_for_metric("pct_pts"), "pct_pts")
  expect_true(is.na(default_unit_for_metric("level")))
})

test_that("obr_long() builds the standard schema", {
  out <- obr_long(
    period      = c("2024-25", "2025-26"),
    period_type = "fiscal_year",
    series      = "PSNB",
    value       = c(120.5, 110.2),
    unit        = "gbp_bn",
    metric_type = "level"
  )
  expect_named(out, c("period", "period_type", "series",
                      "metric_type", "value", "unit"))
  expect_equal(nrow(out), 2)
  expect_equal(out$series, c("PSNB", "PSNB"))
  expect_equal(out$unit, c("gbp_bn", "gbp_bn"))
})

test_that("obr_long() infers metric_type from series when omitted", {
  out <- obr_long(
    period      = c("2025Q1", "2025Q1"),
    period_type = "quarter",
    series      = c("CPI Index", "CPI inflation"),
    value       = c(135.1, 2.1),
    unit        = c("index", "pct")
  )
  expect_equal(out$metric_type, c("index", "yoy_pct"))
})

test_that("URL candidates probe autumn slugs before spring within a year", {
  cands <- efo_url_candidates("detailed-forecast-tables-aggregates")
  yr <- format(Sys.Date(), "%Y")
  nov <- grep(paste0("november-", yr), cands)
  oct <- grep(paste0("october-", yr), cands)
  mar <- grep(paste0("march-", yr), cands)
  expect_true(nov < oct && oct < mar)

  fc <- forecasts_url_candidates()
  expect_true(grep(paste0("november-", yr), fc) <
                grep(paste0("march-", yr), fc))
})

# Stability-rule margin ("headroom") derived from EFO Table 6.5.
#
# The Charter's stability rule requires the current budget to be in balance
# or surplus by the target year. Table 6.5 (Components of net borrowing)
# publishes the current budget deficit for every year of the forecast, so
# the margin against the rule in any year is simply the current budget
# surplus (the deficit with its sign flipped). This function returns that
# path. It deliberately does not hardcode which year is the target year:
# the Charter's target-year convention changes over time (5th forecast
# year, rolling to 3rd from 2026-27), so the caller supplies it.

#' Compute the stability-rule margin from the EFO forecast
#'
#' Returns the current budget surplus for every year of an EFO forecast,
#' derived from Table 6.5 (Components of net borrowing). Under the Charter
#' for Budget Responsibility's stability rule the current budget must be in
#' balance or surplus by the target year, so the value in the target year
#' is the margin, or "headroom", against the rule: positive means the rule
#' is met with room to spare, negative means it is missed.
#'
#' @details
#' The value returned is `-1 *` the published "Current budget deficit"
#' series, i.e. a surplus is positive. Pass `target_year` to flag the year
#' the rule currently bites on; the function does not guess it, because the
#' Charter's target-year convention changes over time (see
#' [obr_fiscal_rules()] for the rules as encoded at release).
#'
#' Note the OBR's own published headroom figure at a fiscal event can
#' differ slightly from the Table 6.5 arithmetic (rounding, and any
#' rule-specific adjustments described in the EFO text). Treat this as the
#' published forecast path for the rule metric, not a reproduction of the
#' OBR's press-notice headroom number.
#'
#' @param vintage Optional EFO vintage label such as `"March 2026"`. If
#'   `NULL` (the default), uses any pin set via [obr_pin()] or the latest
#'   live EFO.
#' @param target_year Optional fiscal-year string (e.g. `"2029-30"`). If
#'   supplied, an `is_target_year` column flags that year.
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard schema columns (`period`,
#' `period_type`, `series`, `metric_type`, `value`, `unit`), where `series`
#' is `"Current budget surplus"` and `value` is in GBP billion (positive =
#' surplus = headroom under the stability rule). If `target_year` is
#' supplied, an additional logical `is_target_year` column is included.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#'
#' hr <- tryCatch(obr_headroom(), error = function(e) NULL)
#' if (!is.null(hr)) hr
#'
#' # Flag the target year and read off the margin
#' hr <- tryCatch(obr_headroom(target_year = "2029-30"),
#'                error = function(e) NULL)
#' if (!is.null(hr)) hr[hr$is_target_year, ]
#'
#' options(op)
#' }
#'
#' @family fiscal rules
#' @export
obr_headroom <- function(vintage = NULL, target_year = NULL, refresh = FALSE) {
  if (!is.null(target_year) &&
      (!is.character(target_year) || length(target_year) != 1L ||
       !grepl("^[0-9]{4}-[0-9]{2}$", target_year))) {
    cli::cli_abort("{.arg target_year} must be a fiscal-year string like {.val 2029-30}.")
  }

  efo  <- get_efo_table("6.5", vintage = vintage, refresh = refresh)
  prov <- obr_provenance(efo)

  df <- as.data.frame(efo)
  cb <- df[df$series == "Current budget deficit", , drop = FALSE]
  if (nrow(cb) == 0L) {
    cli::cli_abort(c(
      "Could not find the {.val Current budget deficit} series in EFO Table 6.5.",
      "!" = "The OBR may have renamed the series. Please file an issue at https://github.com/charlescoverdale/obr/issues."
    ))
  }

  out <- data.frame(
    period      = cb$period,
    period_type = cb$period_type,
    series      = "Current budget surplus",
    metric_type = "level",
    value       = -cb$value,
    unit        = "gbp_bn",
    stringsAsFactors = FALSE
  )
  out <- out[order(out$period), ]
  rownames(out) <- NULL

  notes <- paste0("Derived from EFO Table 6.5: value = -(Current budget ",
                  "deficit). Positive = surplus = margin against the ",
                  "Charter stability rule.")
  if (!is.null(target_year)) {
    out$is_target_year <- out$period == target_year
    if (!any(out$is_target_year)) {
      cli::cli_warn(c(
        "{.arg target_year} {.val {target_year}} is not a year in this EFO forecast.",
        "i" = "Forecast years: {.val {unique(out$period)}}."
      ))
    }
    notes <- paste0(notes, sprintf(" Target year flagged: %s.", target_year))
  }

  new_obr_tbl(
    data        = out,
    publication = prov$publication,
    vintage     = prov$vintage,
    source_url  = prov$source_url,
    retrieved   = prov$retrieved,
    file_md5    = prov$file_md5,
    notes       = notes
  )
}

# Workflow helpers added in v0.4.0.
# obr_compare_vintages():    diff two EFO vintages on the standard schema.
# obr_actual_vs_forecast():  pair OBR forecasts with ONS outturn from the PFD.

# Internal: dispatch a "what" key to the underlying data function. Named
# shortcuts cover the four original v0.4.0 comparisons; any other value is
# treated as an EFO catalogue table id (v0.6.0), so all 39 detailed-forecast
# tables can be diffed across vintages.
.compare_fn <- function(what) {
  switch(
    what,
    "fiscal"     = function(...) get_efo_fiscal(...),
    "inflation"  = function(...) get_efo_economy("inflation", ...),
    "labour"     = function(...) get_efo_economy("labour", ...),
    "output_gap" = function(...) get_efo_economy("output_gap", ...),
    {
      cat_ids <- efo_catalogue_table()$table_id
      if (!what %in% cat_ids) {
        cli::cli_abort(c(
          "Unknown {.arg what} value: {.val {what}}.",
          "i" = paste0("Use one of {.val fiscal}, {.val inflation}, ",
                       "{.val labour}, {.val output_gap}, or any table id ",
                       "from {.fn obr_efo_catalogue} (e.g. {.val 6.13}).")
        ))
      }
      function(...) get_efo_table(what, ...)
    }
  )
}

#' Compare two EFO vintages
#'
#' Pulls the same EFO table from two vintages and returns a tidy diff with
#' a revision column (`value_b - value_a`). Useful for quantifying how the
#' OBR's view changed between fiscal events.
#'
#' @details
#' Rows are the **inner join** of the two vintages on the schema keys
#' (`period`, `period_type`, `series`, `metric_type`, `unit`). Periods or
#' series that are present in only one vintage are silently dropped. If
#' you need to see what was added or removed between vintages, compare
#' `obr_efo_vintages()` row counts or call the underlying functions
#' directly with each vintage and `setdiff()` on the keys.
#'
#' Calling the function with `vintage_a == vintage_b` is allowed and
#' returns an all-zero `revision` column. There is no special handling
#' beyond that.
#'
#' @param vintage_a,vintage_b EFO vintage labels (e.g. `"October 2024"`,
#'   `"March 2026"`). Use [obr_efo_vintages()] to see all valid labels.
#' @param what Which EFO table to compare. Either one of the named
#'   shortcuts `"fiscal"` (Table 6.5, the default), `"inflation"` (sheet
#'   1.7), `"labour"` (sheet 1.6), `"output_gap"` (sheet 1.14), or any
#'   table id from [obr_efo_catalogue()] (e.g. `"6.13"`, `"1.19"`), so all
#'   detailed-forecast tables can be diffed across vintages.
#' @param refresh Logical. If `TRUE`, re-download even if cached files
#'   exist. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema columns
#' (`period`, `period_type`, `series`, `metric_type`, `unit`) plus
#' `value_a`, `value_b`, and `revision` (`value_b - value_a`).
#' Provenance points at the second vintage; the first vintage URL is
#' recorded in the `notes` field.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   diff <- obr_compare_vintages("October 2024", "March 2026")
#'   diff[diff$series == "Net borrowing", ]
#'
#'   # Compare the inflation forecast across two vintages
#'   inf_diff <- obr_compare_vintages("October 2024", "March 2026",
#'   what = "inflation")
#'
#'   # Any catalogue table works too, e.g. debt interest (Table 6.16)
#'   di_diff <- obr_compare_vintages("November 2025", "March 2026",
#'   what = "6.16")
#' })
#' options(op)
#' }
#'
#' @family forecasts
#' @export
obr_compare_vintages <- function(vintage_a, vintage_b,
                                 what = "fiscal",
                                 refresh = FALSE) {
  if (!is.character(what) || length(what) != 1L || is.na(what)) {
    cli::cli_abort("{.arg what} must be a single character string.")
  }
  fn <- .compare_fn(what)

  a <- fn(vintage = vintage_a, refresh = refresh)
  b <- fn(vintage = vintage_b, refresh = refresh)
  if (is.null(a) || is.null(b)) {
    cli::cli_abort(c(
      "Table {.val {what}} could not be fetched for both vintages.",
      "i" = "Cross-reference sheets (see {.fn obr_efo_catalogue}) cannot be compared directly."
    ))
  }

  prov_b  <- obr_provenance(b)
  prov_a  <- obr_provenance(a)

  keys <- c("period", "period_type", "series", "metric_type", "unit")
  # Some tables carry extra identifying columns (e.g. sub_sector for 6.4);
  # include any that appear in both vintages so the join stays one-to-one.
  extra <- setdiff(intersect(names(a), names(b)), c(keys, "value"))
  keys  <- c(keys, extra)
  out  <- merge(
    as.data.frame(a)[, c(keys, "value")],
    as.data.frame(b)[, c(keys, "value")],
    by        = keys,
    suffixes  = c("_a", "_b"),
    all       = FALSE
  )
  out$revision <- out$value_b - out$value_a
  rownames(out) <- NULL

  new_obr_tbl(
    data        = out,
    publication = prov_b$publication,
    vintage     = prov_b$vintage,
    source_url  = prov_b$source_url,
    retrieved   = prov_b$retrieved,
    file_md5    = prov_b$file_md5,
    notes       = sprintf(
      paste0("Vintage diff: %s (a) -> %s (b). ",
             "revision = value_b - value_a. ",
             "Earlier vintage URL: %s"),
      prov_a$vintage, prov_b$vintage, prov_a$source_url
    )
  )
}

# Internal: map a get_forecasts() series name to the matching PFD outturn fn.
.actuals_fn_for <- function(series) {
  switch(
    series,
    "PSNB"        = function(...) get_psnb(...),
    "PSND"        = function(...) get_psnd(...),
    "expenditure" = function(...) get_expenditure(...),
    NULL
  )
}

#' Pair OBR forecasts with PFD outturn
#'
#' Joins the long-format Historical Forecasts Database for a given series
#' against the Public Finances Databank outturn for the same series.
#' Returns one row per (forecast vintage, fiscal year) where both an OBR
#' forecast value and an ONS outturn value exist, with the forecast error
#' (`value_forecast - value_actual`) computed.
#'
#' Useful for forecast-evaluation studies, similar in shape to the OBR's
#' own Forecast Evaluation Report decomposition.
#'
#' Currently supports series for which a clean `gbp_bn` outturn function
#' exists in this package: `"PSNB"`, `"PSND"`, `"expenditure"`. Other
#' series (CPI, GDP, percentages of GDP) need outturn from external
#' packages and will error.
#'
#' @param series Forecast series. One of `"PSNB"`, `"PSND"`, `"expenditure"`.
#'   Defaults to `"PSNB"`.
#' @param refresh Logical. If `TRUE`, re-download underlying files.
#'
#' @return An `obr_tbl` with columns `forecast_date`, `period` (fiscal year
#' being forecast), `period_type`, `series`, `unit`, `value_forecast`
#' (from HFD), `value_actual` (from PFD outturn), and `error`
#' (`value_forecast - value_actual`). Provenance points at the HFD;
#' the PFD source URL is recorded in `notes`.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   eval <- obr_actual_vs_forecast("PSNB")
#'
#'   # 1-year-ahead forecast errors only:
#'   # take the forecast vintage closest to the start of each fiscal year
#'   eval2425 <- eval[eval$period == "2024-25", ]
#'   eval2425[order(eval2425$forecast_date), ]
#' })
#' options(op)
#' }
#'
#' @family forecasts
#' @export
obr_actual_vs_forecast <- function(series = c("PSNB", "PSND", "expenditure"),
                                   refresh = FALSE) {
  series <- match.arg(series)

  fn <- .actuals_fn_for(series)
  if (is.null(fn)) {
    cli::cli_abort(c(
      "No outturn function in this package for series {.val {series}}.",
      "i" = "Supported: {.val PSNB}, {.val PSND}, {.val expenditure}."
    ))
  }

  fc      <- get_forecasts(series, refresh = refresh)
  actuals <- fn(refresh = refresh)

  prov_fc  <- obr_provenance(fc)
  prov_act <- obr_provenance(actuals)

  fc_df  <- as.data.frame(fc)[,  c("forecast_date", "period", "period_type",
                                   "series", "unit", "value")]
  act_df <- as.data.frame(actuals)[, c("period", "value")]

  out <- merge(fc_df, act_df, by = "period",
               suffixes = c("_forecast", "_actual"))
  out$error <- out$value_forecast - out$value_actual
  out <- out[, c("forecast_date", "period", "period_type",
                 "series", "unit",
                 "value_forecast", "value_actual", "error")]
  rownames(out) <- NULL

  new_obr_tbl(
    data        = out,
    publication = prov_fc$publication,
    vintage     = prov_fc$vintage,
    source_url  = prov_fc$source_url,
    retrieved   = prov_fc$retrieved,
    file_md5    = prov_fc$file_md5,
    notes       = sprintf(
      paste0("Forecast vs outturn for %s. ",
             "error = value_forecast - value_actual. ",
             "Outturn source: %s"),
      series, prov_act$source_url
    )
  )
}

test_that("obr_compare_vintages() errors on unknown 'what'", {
  expect_error(
    obr_compare_vintages("October 2024", "March 2026", what = "NOT_A_TABLE"),
    regexp = "Unknown"
  )
})

test_that("obr_actual_vs_forecast() errors on unsupported series", {
  expect_error(
    obr_actual_vs_forecast("CPI"),
    regexp = "should be one of"
  )
})

test_that("obr_compare_vintages('fiscal') returns the expected diff schema", {
  skip_on_cran()
  skip_if_offline()

  diff <- obr_compare_vintages("October 2024", "March 2026", what = "fiscal")

  expect_s3_class(diff, "obr_tbl")
  expect_named(
    diff,
    c("period", "period_type", "series", "metric_type", "unit",
      "value_a", "value_b", "revision")
  )
  expect_type(diff$value_a,  "double")
  expect_type(diff$value_b,  "double")
  expect_type(diff$revision, "double")
  # All rows are inner-joined: should be present in both vintages
  expect_true(all(!is.na(diff$value_a)))
  expect_true(all(!is.na(diff$value_b)))
  # revision must equal value_b - value_a
  expect_equal(diff$revision, diff$value_b - diff$value_a)
  # Should cover Net borrowing across at least one period
  expect_true("Net borrowing" %in% diff$series)
  expect_true(all(diff$period_type == "fiscal_year"))
  expect_true(all(diff$unit == "gbp_bn"))
})

test_that("obr_compare_vintages('inflation') returns metric_type-aware diff", {
  skip_on_cran()
  skip_if_offline()

  diff <- obr_compare_vintages("October 2024", "March 2026", what = "inflation")

  expect_s3_class(diff, "obr_tbl")
  expect_named(
    diff,
    c("period", "period_type", "series", "metric_type", "unit",
      "value_a", "value_b", "revision")
  )
  expect_true(all(diff$period_type == "quarter"))
  # CPI series should be present and tagged yoy_pct/pct
  cpi <- diff[diff$series == "CPI", ]
  expect_gt(nrow(cpi), 0)
  expect_true(all(cpi$metric_type == "yoy_pct"))
  expect_true(all(cpi$unit == "pct"))
})

test_that("obr_actual_vs_forecast('PSNB') joins HFD against PFD outturn", {
  skip_on_cran()
  skip_if_offline()

  eval <- obr_actual_vs_forecast("PSNB")

  expect_s3_class(eval, "obr_tbl")
  expect_named(
    eval,
    c("forecast_date", "period", "period_type", "series", "unit",
      "value_forecast", "value_actual", "error")
  )
  expect_type(eval$value_forecast, "double")
  expect_type(eval$value_actual,   "double")
  expect_type(eval$error,          "double")
  expect_equal(eval$error, eval$value_forecast - eval$value_actual)
  # All series are PSNB
  expect_true(all(eval$series == "PSNB"))
  expect_true(all(eval$unit == "gbp_bn"))
  expect_gt(nrow(eval), 50)
  # 2020-21 should have many forecasts paired against the (very large) outturn
  e_covid <- eval[eval$period == "2020-21", ]
  expect_gt(nrow(e_covid), 5)
})

test_that("obr_actual_vs_forecast() preserves HFD provenance", {
  skip_on_cran()
  skip_if_offline()

  eval <- obr_actual_vs_forecast("PSNB")
  prov <- obr_provenance(eval)
  expect_equal(prov$publication, "HFD")
  expect_match(prov$notes, "Outturn source")
})

test_that("obr_compare_vintages() invariant: revision == value_b - value_a", {
  # Level 3 invariant test. Holds by construction but worth pinning.
  skip_on_cran()
  skip_if_offline()

  diff <- obr_compare_vintages("October 2024", "March 2026", what = "fiscal")
  expect_equal(diff$revision, diff$value_b - diff$value_a, tolerance = 1e-10)
})

test_that("obr_compare_vintages() with same vintage returns zero revisions", {
  # Self-comparison invariant: if a == b, every revision must be exactly 0.
  skip_on_cran()
  skip_if_offline()

  diff <- obr_compare_vintages("March 2026", "March 2026", what = "fiscal")
  expect_gt(nrow(diff), 0)
  expect_true(all(diff$revision == 0))
  expect_equal(diff$value_a, diff$value_b)
})

test_that("obr_actual_vs_forecast() invariant: error == forecast - actual", {
  # Level 3 invariant. Pin it.
  skip_on_cran()
  skip_if_offline()

  eval <- obr_actual_vs_forecast("PSNB")
  expect_equal(eval$error,
               eval$value_forecast - eval$value_actual,
               tolerance = 1e-10)
})

test_that("obr_compare_vintages() accepts catalogue table ids and rejects junk", {
  expect_error(obr_compare_vintages("March 2025", "March 2026", what = "99.99"),
               regexp = "Unknown .*what")
  expect_error(obr_compare_vintages("March 2025", "March 2026", what = 6.5),
               regexp = "single character")
  # A catalogue id resolves to a fetcher without error (no network here)
  fn <- .compare_fn("6.13")
  expect_true(is.function(fn))
})

# Vintage layer for the obr package.
#
# `obr_efo_vintages()` returns a frame of every EFO published since June 2010,
# with publication date and URL slug. `obr_as_of(date)` resolves the EFO that
# was current on a given calendar date. `obr_pin()` / `obr_unpin()` set a
# session-scoped EFO override that flows into get_efo_fiscal() and
# get_efo_economy() unless those functions are called with an explicit
# `vintage =` argument.

# Hardcoded EFO publication calendar. Dates are the publication date of the
# EFO alongside its associated Budget / Statement, taken from obr.uk/efo/.
# Used both by obr_efo_vintages() and by URL construction for pinned vintages.
efo_vintage_table <- function() {
  vintage <- c(
    "June 2010", "November 2010",
    "March 2011", "November 2011",
    "March 2012", "December 2012",
    "March 2013", "December 2013",
    "March 2014", "December 2014",
    "March 2015", "July 2015", "November 2015",
    "March 2016", "November 2016",
    "March 2017", "November 2017",
    "March 2018", "October 2018",
    "March 2019",
    "March 2020", "November 2020",
    "March 2021", "October 2021",
    "March 2022", "November 2022",
    "March 2023", "November 2023",
    "March 2024", "October 2024",
    "March 2025", "November 2025",
    "March 2026"
  )
  date <- as.Date(c(
    "2010-06-22", "2010-11-29",
    "2011-03-23", "2011-11-29",
    "2012-03-21", "2012-12-05",
    "2013-03-20", "2013-12-05",
    "2014-03-19", "2014-12-03",
    "2015-03-18", "2015-07-08", "2015-11-25",
    "2016-03-16", "2016-11-23",
    "2017-03-08", "2017-11-22",
    "2018-03-13", "2018-10-29",
    "2019-03-13",
    "2020-03-11", "2020-11-25",
    "2021-03-03", "2021-10-27",
    "2022-03-23", "2022-11-17",
    "2023-03-15", "2023-11-22",
    "2024-03-06", "2024-10-30",
    "2025-03-26", "2025-11-26",
    "2026-03-26"
  ))
  slug <- vapply(vintage, function(v) {
    parts <- strsplit(tolower(v), " ", fixed = TRUE)[[1]]
    paste0(parts[1], "-", parts[2], "-economic-and-fiscal-outlook")
  }, character(1), USE.NAMES = FALSE)
  data.frame(
    publication = "EFO",
    vintage     = vintage,
    date        = date,
    slug        = slug,
    stringsAsFactors = FALSE
  )
}

# TRUE if a label looks like an EFO vintage ("November 2026") even when it
# is not (yet) in the hardcoded calendar. Lets users pin a brand-new EFO on
# publication day, before a package release adds it to the table.
is_wellformed_vintage <- function(vintage) {
  is.character(vintage) && length(vintage) == 1L && !is.na(vintage) &&
    grepl(paste0("^(January|February|March|April|May|June|July|August|",
                 "September|October|November|December) [0-9]{4}$"), vintage)
}

# Build a download URL for a given EFO vintage and table suffix. Vintages
# outside the hardcoded calendar are accepted if well-formed: the slug is
# constructed by the OBR's naming convention, with a warning.
efo_url_for_vintage <- function(vintage, suffix) {
  v <- efo_vintage_table()
  match <- v[v$vintage == vintage, , drop = FALSE]
  if (nrow(match) == 0L) {
    if (!is_wellformed_vintage(vintage)) {
      cli::cli_abort(c(
        "Unknown EFO vintage {.val {vintage}}.",
        "i" = "Run {.fn obr_efo_vintages} to see all known vintages, or pass a label like {.val November 2026}."
      ))
    }
    cli::cli_warn(c(
      "{.val {vintage}} is not in this version's EFO calendar.",
      "i" = "Constructing the download URL from the OBR's slug convention. This works on publication day for a brand-new EFO, but will 404 if no such EFO exists."
    ))
    slug <- paste0(tolower(gsub(" ", "-", vintage)),
                   "-economic-and-fiscal-outlook")
    return(sprintf("https://obr.uk/download/%s-%s/", slug, suffix))
  }
  sprintf("https://obr.uk/download/%s-%s/", match$slug[1L], suffix)
}

# Cache filename component derived from a vintage string ("March 2026" → "march_2026").
vintage_cache_tag <- function(vintage) {
  gsub("[^a-z0-9]+", "_", tolower(vintage))
}

# Resolve which vintage is currently in force for EFO calls. Precedence:
#   1. explicit `vintage` argument to the user-facing function
#   2. pin set via obr_pin()
#   3. NULL (= latest, via dynamic resolver)
resolve_efo_vintage <- function(vintage = NULL) {
  if (!is.null(vintage)) return(vintage)
  pinned <- getOption("obr.efo_vintage", default = NULL)
  if (!is.null(pinned) && !is.na(pinned) && nzchar(pinned)) return(pinned)
  NULL
}

#' List known OBR Economic and Fiscal Outlook vintages
#'
#' Returns a data frame of every EFO published since the OBR was created in
#' June 2010, with publication dates and the URL slug used in the OBR's
#' download links. Use this to look up which vintages can be pinned via
#' [obr_pin()] or passed to `vintage =` arguments on [get_efo_fiscal()] and
#' [get_efo_economy()].
#'
#' @return A data frame with columns:
#' \describe{
#'   \item{publication}{Always `"EFO"`.}
#'   \item{vintage}{Vintage label, e.g. `"March 2026"`.}
#'   \item{date}{Publication date of the EFO (Date).}
#'   \item{slug}{URL slug used by the OBR for that vintage's download pages.}
#' }
#'
#' @examples
#' head(obr_efo_vintages())
#' tail(obr_efo_vintages(), 5)
#'
#' @family vintage
#' @export
obr_efo_vintages <- function() {
  efo_vintage_table()
}

#' Find the OBR publication that was current on a given date
#'
#' Returns the most recent EFO that had been published on or before `date`.
#' Useful for reproducing analyses as they would have looked at a given point
#' in time, before subsequent forecast revisions.
#'
#' @param date A Date, or anything coercible to one (e.g. `"2024-11-15"`).
#' @param publication Currently only `"EFO"` is supported.
#'
#' @return A length-one character vector containing the vintage label,
#'   e.g. `"October 2024"`.
#'
#' @examples
#' obr_as_of("2024-11-15")
#' obr_as_of(Sys.Date())
#'
#' @family vintage
#' @export
obr_as_of <- function(date, publication = "EFO") {
  publication <- match.arg(publication, choices = "EFO")
  date_d <- tryCatch(as.Date(date),
                     error = function(e) {
                       cli::cli_abort("{.arg date} could not be coerced to a {.cls Date}.")
                     })
  if (is.na(date_d)) {
    cli::cli_abort("{.arg date} could not be coerced to a {.cls Date}.")
  }
  v <- efo_vintage_table()
  past <- v[v$date <= date_d, , drop = FALSE]
  if (nrow(past) == 0L) {
    cli::cli_abort(c(
      "No EFO had been published on or before {format(date_d)}.",
      "i" = "The OBR was established in June 2010; the first EFO was published 22 June 2010."
    ))
  }
  past$vintage[which.max(past$date)]
}

#' Pin a session-wide OBR EFO vintage
#'
#' Sets the EFO vintage that [get_efo_fiscal()] and [get_efo_economy()] will
#' use when called without an explicit `vintage =` argument. The pin is stored
#' as the option `obr.efo_vintage` and lasts for the R session unless
#' overwritten or removed via [obr_unpin()].
#'
#' @param vintage Vintage label such as `"October 2024"`. See
#'   [obr_efo_vintages()] for the full list. A well-formed label that is not
#'   yet in the package's EFO calendar (e.g. a brand-new EFO published after
#'   this package version was released) is accepted with a warning: download
#'   URLs are then constructed from the OBR's slug convention, so a new EFO
#'   can be pinned on publication day. If `NULL`, this function clears
#'   the pin (equivalent to calling [obr_unpin()]).
#'
#' @return Invisibly returns the pinned vintage string, or `NULL` after
#'   clearing.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   obr_pin("October 2024")
#'   obr_pinned()
#'   obr_unpin()
#' })
#' options(op)
#' }
#'
#' @family vintage
#' @export
obr_pin <- function(vintage = NULL) {
  if (is.null(vintage)) {
    return(invisible(obr_unpin()))
  }
  v <- efo_vintage_table()
  if (!vintage %in% v$vintage) {
    if (!is_wellformed_vintage(vintage)) {
      cli::cli_abort(c(
        "Unknown EFO vintage {.val {vintage}}.",
        "i" = "Run {.fn obr_efo_vintages} to see all known vintages."
      ))
    }
    cli::cli_warn(c(
      "{.val {vintage}} is not in this version's EFO calendar.",
      "i" = "Pinning it anyway; download URLs will be constructed from the OBR's slug convention. Useful on publication day for a brand-new EFO."
    ))
  }
  options(obr.efo_vintage = vintage)
  cli::cli_inform(c("v" = "Pinned EFO to {.val {vintage}}."))
  invisible(vintage)
}

#' Clear the pinned OBR EFO vintage
#'
#' Removes any pin set by [obr_pin()]. After unpinning, [get_efo_fiscal()]
#' and [get_efo_economy()] revert to resolving the most recent live EFO via
#' the dynamic URL resolver.
#'
#' @return Invisibly returns the previously pinned vintage (or `NULL` if no
#'   pin was set).
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   obr_pin("March 2025")
#'   obr_unpin()
#' })
#' options(op)
#' }
#'
#' @family vintage
#' @export
obr_unpin <- function() {
  prev <- getOption("obr.efo_vintage", default = NULL)
  options(obr.efo_vintage = NULL)
  if (!is.null(prev)) {
    cli::cli_inform(c("v" = "Unpinned EFO (was {.val {prev}})."))
  }
  invisible(prev)
}

#' Show the currently pinned OBR EFO vintage
#'
#' Returns the vintage string currently active via [obr_pin()], or `NULL`
#' if no pin is set.
#'
#' @return Either a length-one character vector or `NULL`.
#'
#' @examples
#' obr_pinned()
#'
#' @family vintage
#' @export
obr_pinned <- function() {
  pinned <- getOption("obr.efo_vintage", default = NULL)
  if (is.null(pinned) || (length(pinned) == 1L && (is.na(pinned) || !nzchar(pinned)))) {
    return(NULL)
  }
  pinned
}

# Vintage layer tests. Mostly offline; one integration test for pinning.

test_that("obr_efo_vintages() returns a structured table", {
  v <- obr_efo_vintages()
  expect_s3_class(v, "data.frame")
  expect_named(v, c("publication", "vintage", "date", "slug"))
  expect_true(all(v$publication == "EFO"))
  expect_s3_class(v$date, "Date")
  expect_match(v$slug, "-economic-and-fiscal-outlook$", all = TRUE)

  # 33 EFOs from June 2010 to March 2026 inclusive
  expect_gte(nrow(v), 30L)
  expect_true("June 2010"   %in% v$vintage)
  expect_true("October 2024" %in% v$vintage)
  expect_true("March 2026"  %in% v$vintage)
  # Dates are unique and ordered ascending in the table source
  expect_equal(anyDuplicated(v$vintage), 0L)
  expect_equal(anyDuplicated(v$date),    0L)
})

test_that("obr_as_of() returns the most recent EFO published on or before a date", {
  expect_equal(obr_as_of("2024-10-31"), "October 2024")
  expect_equal(obr_as_of("2024-10-29"), "March 2024")
  expect_equal(obr_as_of("2025-01-01"), "October 2024")
  expect_equal(obr_as_of("2026-04-25"), "March 2026")

  # Date object also accepted
  expect_equal(obr_as_of(as.Date("2024-10-30")), "October 2024")
})

test_that("obr_as_of() errors when no EFO had been published yet", {
  expect_error(obr_as_of("2010-05-01"), regexp = "No EFO had been published")
})

test_that("obr_as_of() errors on un-coercible date input", {
  expect_error(obr_as_of("not-a-date"), regexp = "Date")
})

test_that("obr_pin() and obr_unpin() round-trip the global option", {
  on.exit(options(obr.efo_vintage = NULL), add = TRUE)

  expect_null(obr_pinned())

  res <- suppressMessages(obr_pin("October 2024"))
  expect_equal(res, "October 2024")
  expect_equal(obr_pinned(), "October 2024")
  expect_equal(getOption("obr.efo_vintage"), "October 2024")

  prev <- suppressMessages(obr_unpin())
  expect_equal(prev, "October 2024")
  expect_null(obr_pinned())
  expect_null(getOption("obr.efo_vintage"))
})

test_that("obr_pin(NULL) is equivalent to obr_unpin()", {
  on.exit(options(obr.efo_vintage = NULL), add = TRUE)
  suppressMessages(obr_pin("March 2025"))
  expect_equal(obr_pinned(), "March 2025")
  suppressMessages(obr_pin(NULL))
  expect_null(obr_pinned())
})

test_that("obr_pin() rejects unknown vintages", {
  on.exit(options(obr.efo_vintage = NULL), add = TRUE)
  expect_error(obr_pin("Banana 2024"), regexp = "Unknown EFO vintage")
})

test_that("efo_url_for_vintage() builds the standard OBR slug pattern", {
  url <- efo_url_for_vintage("March 2026", "detailed-forecast-tables-aggregates")
  expect_equal(
    url,
    "https://obr.uk/download/march-2026-economic-and-fiscal-outlook-detailed-forecast-tables-aggregates/"
  )
  url2 <- efo_url_for_vintage("October 2024", "detailed-forecast-tables-economy")
  expect_equal(
    url2,
    "https://obr.uk/download/october-2024-economic-and-fiscal-outlook-detailed-forecast-tables-economy/"
  )
})

test_that("vintage_cache_tag() produces filename-safe slugs", {
  expect_equal(vintage_cache_tag("March 2026"),    "march_2026")
  expect_equal(vintage_cache_tag("October 2024"),  "october_2024")
  expect_equal(vintage_cache_tag("July 2015"),     "july_2015")
})

test_that("resolve_efo_vintage() respects precedence: arg > pin > NULL", {
  on.exit(options(obr.efo_vintage = NULL), add = TRUE)

  # Both NULL
  options(obr.efo_vintage = NULL)
  expect_null(resolve_efo_vintage(NULL))

  # Only pin set
  options(obr.efo_vintage = "October 2024")
  expect_equal(resolve_efo_vintage(NULL), "October 2024")

  # Explicit arg overrides pin
  expect_equal(resolve_efo_vintage("March 2025"), "March 2025")
})

test_that("get_efo_fiscal(vintage = ...) downloads the pinned vintage", {
  skip_on_cran()
  skip_if_offline()
  op <- options(obr.cache_dir = tempfile("obr_vintage_test_"))
  on.exit(options(op), add = TRUE)

  res <- get_efo_fiscal(vintage = "March 2026")
  expect_s3_class(res, "obr_tbl")
  prov <- obr_provenance(res)
  expect_equal(prov$publication, "EFO")
  expect_equal(prov$vintage, "March 2026")
  expect_match(prov$source_url, "march-2026-economic-and-fiscal-outlook")
})

test_that("efo_url_for_vintage() constructs slugs for well-formed unknown vintages", {
  expect_warning(
    url <- efo_url_for_vintage("November 2026", "detailed-forecast-tables-aggregates"),
    regexp = "not in this version's EFO calendar"
  )
  expect_equal(
    url,
    "https://obr.uk/download/november-2026-economic-and-fiscal-outlook-detailed-forecast-tables-aggregates/"
  )
  expect_error(
    efo_url_for_vintage("Nonsense 2026", "detailed-forecast-tables-aggregates"),
    regexp = "Unknown EFO vintage"
  )
})

test_that("obr_pin() accepts a well-formed unknown vintage with a warning", {
  on.exit(options(obr.efo_vintage = NULL), add = TRUE)
  expect_warning(obr_pin("November 2026"),
                 regexp = "not in this version's EFO calendar")
  expect_equal(obr_pinned(), "November 2026")
  obr_unpin()
  expect_error(obr_pin("NotAMonth 2026"), regexp = "Unknown EFO vintage")
})

# obr_tbl: a data.frame with attached OBR provenance.
# All data-returning user functions wrap their output in this class so that
# users (and OBR researchers in particular) can audit which OBR publication
# produced which numbers, when they were retrieved, and from what URL.

# Internal constructor. Not exported.
new_obr_tbl <- function(data,
                        publication,
                        source_url,
                        retrieved   = Sys.time(),
                        vintage     = NA_character_,
                        file_md5    = NA_character_,
                        notes       = NULL) {
  if (!is.data.frame(data)) {
    cli::cli_abort("{.arg data} must be a {.cls data.frame}, not a {.cls {class(data)[1]}}.")
  }
  obj <- as.data.frame(data, stringsAsFactors = FALSE)
  rownames(obj) <- NULL

  pkg_version <- tryCatch(
    as.character(utils::packageVersion("obr")),
    error = function(e) NA_character_
  )

  attr(obj, "obr_provenance") <- list(
    publication     = publication,
    vintage         = vintage,
    source_url      = source_url,
    retrieved       = retrieved,
    file_md5        = file_md5,
    package_version = pkg_version,
    notes           = notes
  )
  class(obj) <- c("obr_tbl", "data.frame")
  obj
}

# Map a publication code to a human-readable label for printing.
obr_publication_label <- function(code) {
  switch(code %||% "",
    PFD = "Public Finances Databank",
    HFD = "Historical Official Forecasts Database",
    EFO = "Economic and Fiscal Outlook",
    "EFO-MP" = "EFO Monthly Profiles",
    WTR = "Welfare Trends Report",
    FSR = "Fiscal Risks and Sustainability Report",
    PMD = "Policy Measures Database",
    FRD = "Forecast Revisions Database",
    code
  )
}

# Extract a "Month YYYY" vintage label from an OBR download slug, if present.
# OBR slugs follow patterns like "march-2026-economic-and-fiscal-outlook-..."
obr_url_vintage <- function(url) {
  if (is.null(url) || is.na(url)) return(NA_character_)
  pattern <- "(january|february|march|april|may|june|july|august|september|october|november|december)-([0-9]{4})"
  m <- regmatches(url, regexpr(pattern, url, ignore.case = TRUE))
  if (length(m) == 0L) return(NA_character_)
  parts <- strsplit(m[1], "-", fixed = TRUE)[[1]]
  month <- paste0(toupper(substring(parts[1], 1, 1)),
                  tolower(substring(parts[1], 2)))
  paste(month, parts[2])
}

# Null-coalesce
`%||%` <- function(a, b) if (is.null(a) || (length(a) == 1L && is.na(a))) b else a

#' Print an obr_tbl
#'
#' Prints the data with a provenance header that names the OBR publication,
#' vintage, source URL, retrieval time, and underlying file fingerprint.
#'
#' @param x An `obr_tbl` object.
#' @param n Number of rows to display (default 10).
#' @param ... Further arguments passed to [print.data.frame()].
#'
#' @return The input, returned invisibly.
#' @export
print.obr_tbl <- function(x, n = 10L, ...) {
  prov <- attr(x, "obr_provenance")
  cat(sprintf("# obr_tbl: %d rows x %d cols\n", nrow(x), ncol(x)))
  if (!is.null(prov)) {
    pub_label <- obr_publication_label(prov$publication)
    vintage   <- prov$vintage %||% NA_character_
    head_line <- if (!is.na(vintage)) {
      sprintf("# Source:       OBR %s, %s", pub_label, vintage)
    } else {
      sprintf("# Source:       OBR %s", pub_label)
    }
    cat(head_line, "\n", sep = "")
    if (!is.null(prov$source_url) && !is.na(prov$source_url)) {
      cat(sprintf("# URL:          %s\n", prov$source_url))
    }
    if (!is.null(prov$retrieved) && !is.na(prov$retrieved)) {
      cat(sprintf("# Retrieved:    %s\n",
                  format(prov$retrieved, "%Y-%m-%d %H:%M:%S %Z")))
    }
    if (!is.null(prov$file_md5) && !is.na(prov$file_md5) && nzchar(prov$file_md5)) {
      cat(sprintf("# File MD5:     %s\n", substr(prov$file_md5, 1L, 12L)))
    }
    if (!is.null(prov$package_version) && !is.na(prov$package_version)) {
      cat(sprintf("# Package:      obr %s\n", prov$package_version))
    }
    if (!is.null(prov$notes) && nzchar(prov$notes)) {
      cat(sprintf("# Note:         %s\n", prov$notes))
    }
  }
  cat("\n")
  df <- as.data.frame(x)
  if (nrow(df) > n) {
    print(utils::head(df, n = n), ...)
    cat(sprintf("# ... with %d more row%s\n", nrow(df) - n,
                if (nrow(df) - n == 1L) "" else "s"))
  } else {
    print(df, ...)
  }
  invisible(x)
}

#' Summary of an obr_tbl
#'
#' Returns the full provenance card and a structural summary of the data.
#'
#' @param object An `obr_tbl` object.
#' @param ... Unused.
#'
#' @return Invisibly returns the provenance list.
#' @export
summary.obr_tbl <- function(object, ...) {
  prov <- attr(object, "obr_provenance")
  cat("obr_tbl provenance\n")
  cat("------------------\n")
  if (!is.null(prov)) {
    fields <- c(
      "publication"     = "Publication",
      "vintage"         = "Vintage",
      "source_url"      = "Source URL",
      "retrieved"       = "Retrieved",
      "file_md5"        = "File MD5",
      "package_version" = "Package",
      "notes"           = "Notes"
    )
    for (key in names(fields)) {
      val <- prov[[key]]
      if (is.null(val) || (length(val) == 1L && is.na(val))) next
      if (!nzchar(as.character(val))) next
      label <- fields[[key]]
      if (key == "publication") {
        val <- sprintf("%s (%s)", val, obr_publication_label(val))
      }
      if (inherits(val, "POSIXt")) {
        val <- format(val, "%Y-%m-%d %H:%M:%S %Z")
      }
      cat(sprintf("  %-13s %s\n", paste0(label, ":"), val))
    }
  } else {
    cat("  (no provenance attached)\n")
  }
  cat("\nData\n")
  cat("----\n")
  cat(sprintf("  Rows:    %d\n", nrow(object)))
  cat(sprintf("  Columns: %d (%s)\n", ncol(object),
              paste(names(object), collapse = ", ")))
  invisible(prov)
}

#' Subset an obr_tbl, preserving provenance
#'
#' Standard data.frame subsetting drops the `obr_tbl` class. This method
#' preserves it (and the provenance attributes) when the result is still a
#' data.frame.
#'
#' @param x An `obr_tbl`.
#' @param ... Subsetting arguments passed to [`[.data.frame`].
#'
#' @return An `obr_tbl` if the subset is two-dimensional, otherwise the
#'   underlying vector.
#' @export
`[.obr_tbl` <- function(x, ...) {
  prov <- attr(x, "obr_provenance")
  out <- NextMethod()
  if (is.data.frame(out)) {
    attr(out, "obr_provenance") <- prov
    class(out) <- c("obr_tbl", "data.frame")
  }
  out
}

#' Coerce an obr_tbl to a plain data.frame
#'
#' Strips the provenance attributes and the `obr_tbl` class so the result
#' interacts smoothly with packages that dispatch on `data.frame` directly.
#'
#' @param x An `obr_tbl`.
#' @param ... Unused.
#'
#' @return A plain `data.frame` with no provenance.
#' @export
as.data.frame.obr_tbl <- function(x, ...) {
  attr(x, "obr_provenance") <- NULL
  class(x) <- "data.frame"
  x
}

#' Extract OBR provenance metadata
#'
#' Returns the provenance list attached to an `obr_tbl`: the OBR publication
#' it came from, the publication vintage, the source URL, retrieval time,
#' file fingerprint, and package version.
#'
#' @param x An `obr_tbl` (or any object; returns `NULL` if none attached).
#'
#' @return A named list of provenance fields, or `NULL` if no provenance is
#'   attached. Fields:
#' \describe{
#'   \item{publication}{Short code: `"PFD"`, `"HFD"`, `"EFO"`, `"WTR"`, `"FSR"`, `"PMD"`.}
#'   \item{vintage}{Publication vintage label, e.g. `"March 2026"`.}
#'   \item{source_url}{Canonical OBR download URL the data came from.}
#'   \item{retrieved}{`POSIXct` timestamp of when the file was downloaded or
#'     last validated in the cache.}
#'   \item{file_md5}{MD5 fingerprint of the underlying spreadsheet.}
#'   \item{package_version}{`obr` package version that produced the object.}
#'   \item{notes}{Optional free-text notes, or `NULL`.}
#' }
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   psnb <- get_psnb()
#'   obr_provenance(psnb)
#' })
#' options(op)
#' }
#'
#' @family provenance
#' @export
obr_provenance <- function(x) {
  attr(x, "obr_provenance")
}

# Tests for the obr_tbl class itself. Offline so they run on CRAN.

make_test_tbl <- function() {
  new_obr_tbl(
    data        = data.frame(year = c("2024-25", "2025-26"),
                             value = c(100, 110),
                             stringsAsFactors = FALSE),
    publication = "PFD",
    vintage     = "March 2026",
    source_url  = "https://obr.uk/download/public-finances-databank/",
    retrieved   = as.POSIXct("2026-04-25 12:00:00", tz = "UTC"),
    file_md5    = "abcdef0123456789"
  )
}

test_that("new_obr_tbl() returns an obr_tbl with expected attributes", {
  x <- make_test_tbl()
  expect_s3_class(x, "obr_tbl")
  expect_s3_class(x, "data.frame")

  prov <- attr(x, "obr_provenance")
  expect_false(is.null(prov))
  expect_equal(prov$publication, "PFD")
  expect_equal(prov$vintage, "March 2026")
  expect_equal(prov$source_url,
               "https://obr.uk/download/public-finances-databank/")
  expect_s3_class(prov$retrieved, "POSIXt")
  expect_equal(prov$file_md5, "abcdef0123456789")
})

test_that("obr_provenance() returns the attached metadata", {
  x <- make_test_tbl()
  prov <- obr_provenance(x)
  expect_type(prov, "list")
  expect_equal(prov$publication, "PFD")
  expect_equal(prov$vintage, "March 2026")
})

test_that("obr_provenance() returns NULL on an unwrapped data.frame", {
  expect_null(obr_provenance(data.frame(x = 1)))
})

test_that("subsetting an obr_tbl preserves the class and provenance", {
  x <- make_test_tbl()
  y <- x[x$value > 100, ]
  expect_s3_class(y, "obr_tbl")
  expect_equal(obr_provenance(y)$publication, "PFD")
  expect_equal(nrow(y), 1L)
})

test_that("as.data.frame() strips the obr_tbl class and provenance", {
  x <- make_test_tbl()
  y <- as.data.frame(x)
  expect_s3_class(y, "data.frame")
  expect_false(inherits(y, "obr_tbl"))
  expect_null(attr(y, "obr_provenance"))
})

test_that("print.obr_tbl() prints a provenance header", {
  x <- make_test_tbl()
  out <- capture.output(print(x))
  expect_true(any(grepl("obr_tbl", out)))
  expect_true(any(grepl("Public Finances Databank", out)))
  expect_true(any(grepl("March 2026", out)))
  expect_true(any(grepl("public-finances-databank", out)))
})

test_that("summary.obr_tbl() prints a provenance card", {
  x <- make_test_tbl()
  out <- capture.output(summary(x))
  expect_true(any(grepl("Publication:", out)))
  expect_true(any(grepl("Vintage:", out)))
  expect_true(any(grepl("Source URL:", out)))
  expect_true(any(grepl("Rows:", out)))
})

test_that("obr_url_vintage() extracts month-year from EFO slug", {
  expect_equal(
    obr_url_vintage(
      "https://obr.uk/download/march-2026-economic-and-fiscal-outlook-detailed-forecast-tables-aggregates/"
    ),
    "March 2026"
  )
  expect_equal(
    obr_url_vintage(
      "https://obr.uk/download/october-2024-economic-and-fiscal-outlook/"
    ),
    "October 2024"
  )
  expect_true(is.na(obr_url_vintage(
    "https://obr.uk/download/public-finances-databank/"
  )))
})

test_that("obr_publication_label() resolves the codes used in this package", {
  expect_equal(obr_publication_label("PFD"), "Public Finances Databank")
  expect_equal(obr_publication_label("HFD"),
               "Historical Official Forecasts Database")
  expect_equal(obr_publication_label("EFO"),
               "Economic and Fiscal Outlook")
  expect_equal(obr_publication_label("WTR"), "Welfare Trends Report")
  expect_equal(obr_publication_label("FSR"),
               "Fiscal Risks and Sustainability Report")
})

test_that("strip_footnote() preserves names that legitimately end in a digit", {
  expect_equal(strip_footnote("Income tax2"), "Income tax")
  expect_equal(strip_footnote("VAT3"), "VAT")
  # Year-bearing labels (not used in OBR receipts but worth guarding)
  expect_equal(strip_footnote("2024"), "2024")
  expect_equal(strip_footnote("Income tax"), "Income tax")
})

test_that("get_forecast_revisions() errors on invalid unit", {
  expect_error(get_forecast_revisions(unit = "kgs"),
               regexp = "should be one of")
})

test_that("get_forecast_revisions() returns long-format obr_tbl", {
  skip_on_cran()
  skip_if_offline()

  res <- get_forecast_revisions(unit = "gbp_bn")
  expect_s3_class(res, "obr_tbl")
  expect_named(res, c("forecast_date", "component", "fiscal_year", "value"))
  expect_type(res$value, "double")

  # Top-level components present
  components <- unique(res$component)
  expect_true("Total"      %in% components)
  expect_true("Policy"     %in% components)
  expect_true("Underlying" %in% components)
  expect_true(any(grepl("Classification", components)))

  # Forecast dates look right
  expect_true(all(grepl("^[A-Z][a-z]+ [0-9]{4}$", res$forecast_date)))
  # Fiscal years look right
  expect_true(all(grepl("^[0-9]{4}-[0-9]{2}$", res$fiscal_year)))

  prov <- obr_provenance(res)
  expect_equal(prov$publication, "FRD")
  expect_match(prov$source_url, "forecast-revisions-database")
})

test_that("get_forecast_revisions(unit='pct_gdp') returns sensible magnitudes", {
  skip_on_cran()
  skip_if_offline()

  res <- get_forecast_revisions(unit = "pct_gdp")
  expect_s3_class(res, "obr_tbl")
  # PSNB revisions across vintages can be large in extreme periods
  # (e.g. COVID reclassifications), but should stay well within +/- 50% of GDP.
  expect_true(all(abs(res$value) < 50))
  # Median revision for any single vintage-year-component is small
  expect_lt(stats::median(abs(res$value)), 5)
})

# EFO Monthly Profiles.
# About a month after each EFO, the OBR publishes a workbook of monthly
# profiles for receipts, spending, and the CGNCR, consistent with that EFO's
# full-year forecast. The ONS/HMT monthly public sector finances release is
# then judged against these profiles ("borrowing so far this year vs the
# OBR profile"), making this the most-watched OBR product between fiscal
# events, and the reference point for the run-up to a Budget.

MONTHLY_PROFILES_FALLBACK <- "https://obr.uk/download/march-2026-economic-and-fiscal-outlook-monthly-profiles/"
MONTHLY_PROFILES_FILENAME <- "efo_monthly_profiles.xlsx"

monthly_profiles_source <- function(refresh = FALSE, vintage = NULL) {
  vintage <- resolve_efo_vintage(vintage)
  if (!is.null(vintage)) {
    return(efo_pinned_source(
      vintage = vintage,
      suffix  = "monthly-profiles",
      stem    = "efo_monthly_profiles",
      refresh = refresh
    ))
  }
  obr_get_xlsx(
    candidates = efo_url_candidates("monthly-profiles"),
    fallback   = MONTHLY_PROFILES_FALLBACK,
    filename   = MONTHLY_PROFILES_FILENAME,
    refresh    = refresh,
    label      = "EFO Monthly Profiles"
  )
}

MONTH_ABBREVS <- c("Apr", "May", "Jun", "Jul", "Aug", "Sep",
                   "Oct", "Nov", "Dec", "Jan", "Feb", "Mar")

# Convert a month abbreviation to a "YYYY-MM" period string given the start
# year of the fiscal year it belongs to (Apr-Dec = start year, Jan-Mar = +1).
month_to_period <- function(month_abbrev, fy_start_year) {
  idx  <- match(month_abbrev, MONTH_ABBREVS)
  mnum <- ((idx + 2L) %% 12L) + 1L                # Apr -> 4, ..., Mar -> 3
  year <- ifelse(mnum >= 4L, fy_start_year, fy_start_year + 1L)
  sprintf("%d-%02d", year, mnum)
}

# Parse one monthly-profiles sheet. Layout: series names in col 1 (with
# sparse "of which:" markers), the full-year EFO forecast in the column
# headed "EFO Forecast", and monthly values under an Apr..Mar header row.
parse_monthly_profiles_sheet <- function(path, sheet) {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")

  month_regex <- paste0("^(", paste(MONTH_ABBREVS, collapse = "|"), ")$")

  # Month header row: first row with at least six month abbreviations.
  month_row <- NA_integer_
  for (i in seq_len(nrow(raw))) {
    cells <- trimws(as.character(unlist(raw[i, ])))
    if (sum(grepl(month_regex, cells), na.rm = TRUE) >= 6L) {
      month_row <- i
      break
    }
  }
  if (is.na(month_row)) return(NULL)

  month_cells <- trimws(as.character(unlist(raw[month_row, ])))
  month_cols  <- which(grepl(month_regex, month_cells))
  months      <- month_cells[month_cols]

  # Fiscal year: first "YYYY-YY" anywhere at or above the month header row.
  fiscal_year <- NA_character_
  for (i in seq_len(month_row)) {
    cells <- as.character(unlist(raw[i, ]))
    m <- regmatches(cells, regexpr("[0-9]{4}-[0-9]{2}", cells))
    m <- m[!is.na(m) & nzchar(m)]
    if (length(m) > 0L) { fiscal_year <- m[1L]; break }
  }
  if (is.na(fiscal_year)) return(NULL)
  fy_start_year <- as.integer(substr(fiscal_year, 1L, 4L))

  # Full-year forecast column: the column whose header mentions "Forecast"
  # above the month row (e.g. "Mar26 EFO Forecast"). Restricted to the
  # columns between the series names (col 1, whose sheet title also says
  # "forecast") and the first month column.
  fy_col <- NA_integer_
  fy_search_cols <- setdiff(seq_len(min(month_cols) - 1L), 1L)
  for (i in seq_len(month_row)) {
    if (length(fy_search_cols) == 0L) break
    cells <- as.character(unlist(raw[i, fy_search_cols]))
    hit <- which(grepl("Forecast", cells, ignore.case = TRUE))
    if (length(hit) > 0L) { fy_col <- fy_search_cols[hit[1L]]; break }
  }

  periods <- month_to_period(months, fy_start_year)

  results <- list()
  if (month_row >= nrow(raw)) return(NULL)
  for (i in (month_row + 1L):nrow(raw)) {
    series <- trimws(as.character(unlist(raw[i, 1L])))
    if (is.na(series) || !nzchar(series)) next
    if (grepl("^of which", series, ignore.case = TRUE)) next
    if (grepl("^Note", series, ignore.case = TRUE)) next

    vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[i, month_cols])))
    )
    fy_val <- if (!is.na(fy_col)) {
      suppressWarnings(as.numeric(as.character(unlist(raw[i, fy_col]))))
    } else {
      NA_real_
    }
    if (all(is.na(vals)) && is.na(fy_val)) next

    part <- obr_long(
      period      = periods,
      period_type = "month",
      series      = series,
      value       = vals,
      unit        = "gbp_bn",
      metric_type = "level"
    )
    if (!is.na(fy_val)) {
      part <- rbind(part, obr_long(
        period      = fiscal_year,
        period_type = "fiscal_year",
        series      = series,
        value       = fy_val,
        unit        = "gbp_bn",
        metric_type = "level"
      ))
    }
    results[[length(results) + 1L]] <- part
  }

  if (length(results) == 0L) return(NULL)
  out <- do.call(rbind, results)
  out[!is.na(out$value), ]
}

#' Get the OBR monthly profiles for the public finances
#'
#' Downloads (and caches) the monthly profiles workbook the OBR publishes
#' alongside each \emph{Economic and Fiscal Outlook}. The profiles apportion
#' the full-year EFO forecast for receipts, spending, and the central
#' government net cash requirement (CGNCR) across the twelve months of the
#' fiscal year, so that each month's ONS/HMT public sector finances outturn
#' can be judged against the path implied by the OBR's forecast.
#'
#' This is the reference point used every month in the run-up to a fiscal
#' event ("borrowing so far this year vs the OBR profile"). The OBR itself
#' publishes a monthly commentary against these profiles; this function
#' provides the underlying numbers in tidy long format. Pair with monthly
#' outturn data (e.g. from the ONS public sector finances release) to
#' compute in-year deviations from profile.
#'
#' The profiles workbook is typically published a few weeks after the EFO
#' itself. The OBR describes the profiles as broad-brush and illustrative;
#' see the Notes sheet of the source workbook.
#'
#' @param sheet Which profile table to return. `"profiles"` (the default)
#'   returns the receipts and spending profiles; `"cgncr"` returns the
#'   central government net cash requirement breakdown.
#' @param vintage Optional EFO vintage label such as `"March 2026"`. If
#'   `NULL` (the default), uses any pin set via [obr_pin()] or resolves the
#'   latest live profiles workbook.
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard schema columns (`period`,
#' `period_type`, `series`, `metric_type`, `value`, `unit`). Monthly rows
#' have `period_type = "month"` and `period` in `"YYYY-MM"` format; each
#' series also carries one `period_type = "fiscal_year"` row holding the
#' full-year EFO forecast the profile sums to. All values are GBP billion.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#'
#' mp <- tryCatch(get_monthly_profiles(), error = function(e) NULL)
#' if (!is.null(mp)) {
#'   # Monthly profile for HMRC cash receipts
#'   mp[mp$series == "HMRC cash receipts" & mp$period_type == "month", ]
#' }
#'
#' # CGNCR breakdown by month
#' cg <- tryCatch(get_monthly_profiles("cgncr"), error = function(e) NULL)
#'
#' options(op)
#' }
#'
#' @family EFO
#' @export
get_monthly_profiles <- function(sheet = c("profiles", "cgncr"),
                                 vintage = NULL,
                                 refresh = FALSE) {
  sheet <- match.arg(sheet)

  src <- monthly_profiles_source(refresh = refresh, vintage = vintage)

  sheet_name <- if (sheet == "profiles") {
    resolve_sheet_name(src$path, "Monthly Profiles", "[Pp]rofiles")
  } else {
    resolve_sheet_name(src$path, "CGNCR Breakdown", "CGNCR")
  }

  data <- parse_monthly_profiles_sheet(src$path, sheet_name)
  if (is.null(data)) {
    cli::cli_abort(c(
      "Could not parse sheet {.val {sheet_name}} of the monthly profiles workbook.",
      "!" = "The OBR may have changed the layout. Please file an issue at https://github.com/charlescoverdale/obr/issues."
    ))
  }

  new_obr_tbl(
    data        = data,
    publication = "EFO-MP",
    vintage     = obr_url_vintage(src$url),
    source_url  = src$url,
    retrieved   = src$retrieved,
    file_md5    = src$file_md5,
    notes       = "Monthly profiles are broad-brush and illustrative (OBR). Each series also carries a fiscal_year row with the full-year EFO forecast."
  )
}

# Monthly profiles tests. Unit tests are offline; parsing tests hit the
# live workbook and are skipped on CRAN.

v04_long_cols <- c("period", "period_type", "series",
                   "metric_type", "value", "unit")

test_that("month_to_period() maps the fiscal year correctly", {
  expect_equal(month_to_period("Apr", 2026), "2026-04")
  expect_equal(month_to_period("Dec", 2026), "2026-12")
  expect_equal(month_to_period("Jan", 2026), "2027-01")
  expect_equal(month_to_period("Mar", 2026), "2027-03")
  # Vectorised
  expect_equal(month_to_period(c("Apr", "Jan"), 2026),
               c("2026-04", "2027-01"))
})

test_that("get_monthly_profiles() validates the sheet argument", {
  expect_error(get_monthly_profiles("nonsense"), regexp = "arg")
})

test_that("get_monthly_profiles() returns tidy monthly profiles", {
  skip_on_cran()
  skip_if_offline()
  op <- options(obr.cache_dir = tempdir())
  on.exit(options(op), add = TRUE)

  mp <- get_monthly_profiles()
  expect_s3_class(mp, "obr_tbl")
  expect_true(all(v04_long_cols %in% names(mp)))
  expect_true(all(mp$period_type %in% c("month", "fiscal_year")))
  expect_true(all(mp$unit == "gbp_bn"))

  monthly <- mp[mp$period_type == "month", ]
  expect_match(monthly$period, "^[0-9]{4}-[0-9]{2}$", all = TRUE)
  expect_true("HMRC cash receipts" %in% monthly$series)

  # The 12 monthly values sum to the full-year forecast (small tolerance
  # for rounding in the published workbook).
  s <- "HMRC cash receipts"
  m_sum <- sum(monthly$value[monthly$series == s])
  fy    <- mp$value[mp$series == s & mp$period_type == "fiscal_year"]
  expect_equal(length(fy), 1L)
  expect_lt(abs(m_sum - fy), 1.0)

  prov <- obr_provenance(mp)
  expect_equal(prov$publication, "EFO-MP")
  expect_match(prov$source_url, "monthly-profiles")
})

test_that("get_monthly_profiles('cgncr') returns the CGNCR breakdown", {
  skip_on_cran()
  skip_if_offline()
  op <- options(obr.cache_dir = tempdir())
  on.exit(options(op), add = TRUE)

  cg <- get_monthly_profiles("cgncr")
  expect_s3_class(cg, "obr_tbl")
  expect_true("CGNCR" %in% cg$series)
  expect_true(any(cg$period_type == "month"))
})

# OBR Forecast Revisions Database (FRD).
# A workbook decomposing each EFO-to-EFO change in the headline PSNB
# forecast into policy, classification, and underlying (economic
# determinants) components. Two parallel sheets: "Revisions (Per cent of
# GDP)" and "Revisions (£ billion)". Within each sheet, every forecast
# vintage occupies a block of rows: the vintage row carries the total
# revision, then sub-rows label Policy / Classifications and one-offs /
# Underlying, each with optional sub-components.

FRD_FALLBACK <- "https://obr.uk/download/forecast-revisions-database-march-2025/"
FRD_FILENAME <- "forecast_revisions.xlsx"

frd_url_candidates <- function() {
  current_year <- as.integer(format(Sys.Date(), "%Y"))
  months <- c("march", "november", "october", "july")
  candidates <- character(0)
  for (yr in seq(current_year, current_year - 3L)) {
    for (mn in months) {
      candidates <- c(candidates, paste0(
        "https://obr.uk/download/forecast-revisions-database-", mn, "-", yr, "/"
      ))
    }
  }
  candidates
}

frd_source <- function(refresh = FALSE) {
  obr_get_xlsx(
    candidates = frd_url_candidates(),
    fallback   = FRD_FALLBACK,
    filename   = FRD_FILENAME,
    refresh    = refresh,
    label      = "Forecast Revisions Database"
  )
}

frd_obr_tbl <- function(data, src, notes = NULL) {
  new_obr_tbl(
    data        = data,
    publication = "FRD",
    vintage     = obr_url_vintage(src$final_url %||% src$url),
    source_url  = src$url,
    retrieved   = src$retrieved,
    file_md5    = src$file_md5,
    notes       = notes
  )
}

# Parse the "Revisions (Per cent of GDP)" or "Revisions (£ billion)" sheet
# into long format: one row per (forecast_date, component, fiscal_year).
# Top-level components are "Total", "Policy", "Classifications and one-offs",
# and "Underlying"; sub-components carry the OBR's labels (e.g.
# "of which: receipts").
parse_frd_revisions <- function(path, unit) {
  sheet_name <- switch(unit,
    pct_gdp = "Revisions (Per cent of GDP)",
    gbp_bn  = "Revisions (\u00a3 billion)",
    cli::cli_abort("Unknown FRD unit {.val {unit}}.")
  )
  raw <- readxl::read_excel(path, sheet = sheet_name,
                            col_names = FALSE, .name_repair = "minimal")

  yr_header <- as.character(unlist(raw[2, ]))
  year_cols <- which(grepl("^[0-9]{4}-[0-9]{2}$", yr_header))
  if (length(year_cols) == 0L) {
    cli::cli_abort(c(
      "Could not find fiscal-year headers in FRD sheet {.val {sheet_name}}.",
      "i" = "OBR may have changed the layout. Please file an issue."
    ))
  }
  fiscal_years <- yr_header[year_cols]

  col1 <- as.character(unlist(raw[, 1]))
  is_vintage   <- !is.na(col1) & grepl("^[A-Z][a-z]+ [0-9]{4}$", col1)
  vintage_rows <- which(is_vintage)
  if (length(vintage_rows) == 0L) {
    cli::cli_abort(c(
      "No forecast vintage rows found in FRD sheet {.val {sheet_name}}.",
      "i" = "Expected first-column entries like {.val March 2024}."
    ))
  }

  results <- list()
  for (v_idx in seq_along(vintage_rows)) {
    v_row         <- vintage_rows[v_idx]
    forecast_date <- col1[v_row]
    end_row <- if (v_idx < length(vintage_rows)) {
      vintage_rows[v_idx + 1L] - 1L
    } else {
      nrow(raw)
    }

    total_vals <- suppressWarnings(
      as.numeric(as.character(unlist(raw[v_row, year_cols])))
    )
    if (any(!is.na(total_vals))) {
      results[[length(results) + 1L]] <- data.frame(
        forecast_date = forecast_date,
        component     = "Total",
        fiscal_year   = fiscal_years,
        value         = total_vals,
        stringsAsFactors = FALSE
      )
    }

    if (v_row < end_row) {
      for (i in (v_row + 1L):end_row) {
        if (i > nrow(raw)) break
        label <- col1[i]
        if (is.na(label) || !nzchar(label)) next
        if (label == "of which:") next
        # Skip if this is the start of the next vintage (defensive: covered
        # by the end_row calculation, but cheap insurance).
        if (grepl("^[A-Z][a-z]+ [0-9]{4}$", label)) next
        vals <- suppressWarnings(
          as.numeric(as.character(unlist(raw[i, year_cols])))
        )
        if (all(is.na(vals))) next
        results[[length(results) + 1L]] <- data.frame(
          forecast_date = forecast_date,
          component     = label,
          fiscal_year   = fiscal_years,
          value         = vals,
          stringsAsFactors = FALSE
        )
      }
    }
  }

  result <- do.call(rbind, results)
  result <- result[!is.na(result$value), ]
  rownames(result) <- NULL
  result
}

#' Get OBR forecast revisions
#'
#' Downloads (and caches) the OBR Forecast Revisions Database, which
#' decomposes each EFO-to-EFO revision in the headline Public Sector Net
#' Borrowing forecast into three top-level components - policy,
#' classifications and one-offs, and underlying (economic determinants) -
#' with sub-components for each.
#'
#' This is the dataset behind the "what changed?" attribution charts in
#' OBR fiscal commentary and IFS Green Budget chapters.
#'
#' @param unit Character. Either `"gbp_bn"` (GBP billion, the default) or
#'   `"pct_gdp"` (per cent of GDP).
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with columns:
#' \describe{
#'   \item{forecast_date}{Forecast vintage, e.g. `"November 2024"` (character).}
#'   \item{component}{Revision component. Top-level rows are `"Total"`,
#'     `"Policy"`, `"Classifications and one-offs"`, and `"Underlying"`.
#'     Sub-components carry the OBR's `"of which: ..."` labels (character).}
#'   \item{fiscal_year}{Fiscal year being revised, e.g. `"2024-25"` (character).}
#'   \item{value}{Revision value, in GBP billion or per cent of GDP per
#'     `unit` (numeric). A positive value indicates an upward revision to
#'     PSNB.}
#' }
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   rev <- get_forecast_revisions()
#'   # Top-level revisions only
#'   rev[rev$component %in% c("Total", "Policy",
#'   "Classifications and one-offs", "Underlying"), ]
#' })
#' options(op)
#' }
#'
#' @family forecasts
#' @export
get_forecast_revisions <- function(unit = c("gbp_bn", "pct_gdp"),
                                   refresh = FALSE) {
  unit <- match.arg(unit)
  src  <- frd_source(refresh)
  data <- parse_frd_revisions(src$path, unit)
  notes <- sprintf("Decomposition of PSNB forecast revisions in %s.",
                   if (unit == "gbp_bn") "GBP billion" else "per cent of GDP")
  frd_obr_tbl(data, src, notes = notes)
}

# OBR Policy Measures Database.
# A single Excel workbook with two main sheets ("Tax Measures" and
# "Spending Measures"), one row per individual policy measure since 1970
# (tax) or 2010 (spending). Each row carries the fiscal event that scored
# the measure, a description, the Treasury head, and the Exchequer effect
# in GBP million for each fiscal year out to year five.

PMD_FALLBACK <- "https://obr.uk/download/policy-measures-database-march-2025/"
PMD_FILENAME <- "policy_measures_database.xlsx"

# Build PMD URL candidates for recent fiscal events. OBR re-uses an older
# slug across vintages, so we still include the known-stable slug as a
# fallback after probing the most recent slugs.
pmd_url_candidates <- function() {
  current_year <- as.integer(format(Sys.Date(), "%Y"))
  months <- c("november", "october", "july", "march")
  candidates <- character(0)
  for (yr in seq(current_year, current_year - 3L)) {
    for (mn in months) {
      candidates <- c(candidates, paste0(
        "https://obr.uk/download/policy-measures-database-", mn, "-", yr, "/"
      ))
    }
  }
  candidates
}

pmd_source <- function(refresh = FALSE) {
  obr_get_xlsx(
    candidates = pmd_url_candidates(),
    fallback   = PMD_FALLBACK,
    filename   = PMD_FILENAME,
    refresh    = refresh,
    label      = "Policy Measures Database"
  )
}

pmd_obr_tbl <- function(data, src) {
  new_obr_tbl(
    data        = data,
    publication = "PMD",
    vintage     = obr_url_vintage(src$final_url %||% src$url),
    source_url  = src$url,
    retrieved   = src$retrieved,
    file_md5    = src$file_md5
  )
}

# Parse a "Tax Measures" or "Spending Measures" sheet into long format.
# Layout assumption: row 1 = sheet title, row 2 = blank, row 3 = headers
# ("Event", "Measure description", "Tax head" / "Spending head", then
# fiscal-year columns), row 4+ = data. Trailing footnote rows where the
# event column is NA are dropped automatically.
parse_pmd_measures <- function(path, sheet) {
  raw <- readxl::read_excel(path, sheet = sheet,
                            col_names = FALSE, .name_repair = "minimal")
  hdr <- as.character(unlist(raw[3, ]))
  year_cols <- which(grepl("^[0-9]{4}-[0-9]{2}$", hdr))
  if (length(year_cols) == 0L) {
    cli::cli_abort(c(
      "No fiscal-year header columns found in sheet {.val {sheet}}.",
      "i" = "OBR may have changed the PMD layout. Please file an issue at https://github.com/charlescoverdale/obr/issues."
    ))
  }
  fiscal_years <- hdr[year_cols]

  body <- raw[4:nrow(raw), ]
  events  <- as.character(unlist(body[, 1]))
  descs   <- as.character(unlist(body[, 2]))
  heads   <- as.character(unlist(body[, 3]))

  result <- do.call(rbind, lapply(seq_along(year_cols), function(i) {
    yr_col <- year_cols[i]
    data.frame(
      event       = events,
      measure     = descs,
      head        = heads,
      fiscal_year = fiscal_years[i],
      value_mn    = suppressWarnings(
        as.numeric(as.character(unlist(body[, yr_col])))
      ),
      stringsAsFactors = FALSE
    )
  }))

  result <- result[!is.na(result$event) & !is.na(result$value_mn), ]
  rownames(result) <- NULL
  result
}

#' Get OBR policy measures by fiscal event
#'
#' Downloads (and caches) the OBR Policy Measures Database, a single workbook
#' that lists every tax or spending measure scored at a UK fiscal event,
#' with the Exchequer effect in GBP million for each year of the forecast
#' horizon. Tax measures cover 1970 to date; spending measures cover 2010
#' to date.
#'
#' This is the only programmatic access to the PMD: the OBR otherwise
#' distributes it as a single Excel file.
#'
#' @param type Character vector. Which measures to return: `"tax"`,
#'   `"spending"`, or both (the default).
#' @param search Optional character. A regular expression used to filter
#'   `measure` and `head` (case-insensitive). For example,
#'   `search = "alcohol"` keeps any measure whose description or head
#'   mentions alcohol.
#' @param since Optional fiscal-year cut-off (e.g. `"2020-21"`). Rows whose
#'   `fiscal_year` is earlier are dropped.
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy
#'   exists. Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with columns:
#' \describe{
#'   \item{event}{Fiscal event that scored the measure, e.g. `"Budget 2024"` (character)}
#'   \item{measure}{Plain-English description of the measure (character)}
#'   \item{head}{Tax head (for tax measures) or spending head (character)}
#'   \item{fiscal_year}{Fiscal year being scored, e.g. `"2024-25"` (character)}
#'   \item{value_mn}{Exchequer effect in GBP million. For tax measures, a
#'     positive value raises revenue; for spending measures, a positive
#'     value increases spending (numeric)}
#' }
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#'
#' # The OBR CDN occasionally rate-limits or returns 403 to repeated
#' # requests. tryCatch so the example degrades gracefully if that happens
#' # during the CRAN check.
#' oct24 <- tryCatch(get_policy_measures(),
#'                   error = function(e) NULL)
#' if (!is.null(oct24)) {
#'   head(oct24[grepl("2024", oct24$event) &
#'              oct24$fiscal_year == "2025-26", ])
#'
#'   # All alcohol-duty measures since 2010
#'   tryCatch(
#'     get_policy_measures(type = "tax", search = "alcohol",
#'                         since = "2010-11"),
#'     error = function(e) NULL
#'   )
#' }
#'
#' options(op)
#' }
#'
#' @family policy measures
#' @export
get_policy_measures <- function(type = c("tax", "spending"),
                                search = NULL,
                                since  = NULL,
                                refresh = FALSE) {
  type <- match.arg(type, several.ok = TRUE)
  if (!is.null(search) && (!is.character(search) || length(search) != 1L)) {
    cli::cli_abort("{.arg search} must be a single character string.")
  }
  if (!is.null(since) &&
      (!is.character(since) || length(since) != 1L ||
       !grepl("^[0-9]{4}-[0-9]{2}$", since))) {
    cli::cli_abort("{.arg since} must be a fiscal-year string like {.val 2020-21}.")
  }

  src <- pmd_source(refresh)

  parts <- list()
  if ("tax" %in% type) {
    tx <- parse_pmd_measures(src$path, "Tax Measures")
    tx$type <- "tax"
    parts$tax <- tx
  }
  if ("spending" %in% type) {
    sp <- parse_pmd_measures(src$path, "Spending Measures")
    sp$type <- "spending"
    parts$spending <- sp
  }

  data <- do.call(rbind, parts)
  rownames(data) <- NULL
  data <- data[, c("type", "event", "measure", "head", "fiscal_year", "value_mn")]

  if (!is.null(search)) {
    keep <- grepl(search, data$measure, ignore.case = TRUE) |
            grepl(search, data$head,    ignore.case = TRUE)
    data <- data[keep, ]
  }
  if (!is.null(since)) {
    data <- data[data$fiscal_year >= since, ]
  }
  rownames(data) <- NULL

  pmd_obr_tbl(data, src)
}

#' Summarise policy measures by fiscal event
#'
#' Aggregates the measures returned by [get_policy_measures()] to give the
#' net Exchequer effect (positive = revenue-raising / spending-reducing for
#' tax, spending-increasing for spending) by fiscal event and fiscal year.
#'
#' @param x An `obr_tbl` returned by [get_policy_measures()].
#'
#' @return An `obr_tbl` with columns:
#' \describe{
#'   \item{type}{`"tax"` or `"spending"`}
#'   \item{event}{Fiscal event}
#'   \item{fiscal_year}{Fiscal year}
#'   \item{value_mn}{Sum of the Exchequer effect across all measures
#'     scored at that event, in GBP million}
#' }
#' Provenance is preserved.
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   pm <- get_policy_measures(type = "tax", since = "2024-25")
#'   policy_measures_summary(pm)
#' })
#' options(op)
#' }
#'
#' @family policy measures
#' @export
policy_measures_summary <- function(x) {
  if (!inherits(x, "obr_tbl")) {
    cli::cli_abort("{.arg x} must be an {.cls obr_tbl} from {.fn get_policy_measures}.")
  }
  prov <- attr(x, "obr_provenance")
  df <- as.data.frame(x)
  required <- c("type", "event", "fiscal_year", "value_mn")
  missing <- setdiff(required, names(df))
  if (length(missing) > 0L) {
    cli::cli_abort(c(
      "{.arg x} is missing required column{?s} {.val {missing}}.",
      "i" = "Pass the result of {.fn get_policy_measures} directly."
    ))
  }
  agg <- aggregate(value_mn ~ type + event + fiscal_year, data = df, FUN = sum)
  agg <- agg[order(agg$event, agg$type, agg$fiscal_year), ]
  rownames(agg) <- NULL
  out <- new_obr_tbl(
    data        = agg,
    publication = "PMD",
    vintage     = prov$vintage,
    source_url  = prov$source_url,
    retrieved   = prov$retrieved,
    file_md5    = prov$file_md5,
    notes       = "Summed across measures by event and fiscal year."
  )
  out
}

# Network tests for the Policy Measures Database wrapper.

test_that("get_policy_measures() errors on invalid type", {
  expect_error(
    get_policy_measures(type = "bananas"),
    regexp = "should be one of"
  )
})

test_that("get_policy_measures() errors on bad search/since", {
  expect_error(get_policy_measures(search = c("a", "b")),
               regexp = "single character string")
  expect_error(get_policy_measures(since = "20-21"),
               regexp = "fiscal-year string")
})

test_that("get_policy_measures(type='tax') returns expected structure", {
  skip_on_cran()
  skip_if_offline()

  res <- get_policy_measures(type = "tax", since = "2024-25")
  expect_s3_class(res, "obr_tbl")
  expect_named(res, c("type", "event", "measure", "head",
                      "fiscal_year", "value_mn"))
  expect_true(all(res$type == "tax"))
  expect_type(res$value_mn, "double")
  expect_true(all(grepl("^[0-9]{4}-[0-9]{2}$", res$fiscal_year)))
  expect_gte(nrow(res), 50L)

  prov <- obr_provenance(res)
  expect_equal(prov$publication, "PMD")
  expect_match(prov$source_url, "policy-measures-database")
})

test_that("get_policy_measures() defaults return both tax and spending", {
  skip_on_cran()
  skip_if_offline()

  res <- get_policy_measures(since = "2023-24")
  expect_true(all(c("tax", "spending") %in% unique(res$type)))
})

test_that("get_policy_measures(search = ...) filters case-insensitively", {
  skip_on_cran()
  skip_if_offline()

  res <- get_policy_measures(type = "tax", search = "alcohol",
                             since = "2010-11")
  expect_gt(nrow(res), 0L)
  expect_true(all(
    grepl("alcohol", res$measure, ignore.case = TRUE) |
    grepl("alcohol", res$head,    ignore.case = TRUE)
  ))
})

test_that("policy_measures_summary() aggregates by event and fiscal_year", {
  skip_on_cran()
  skip_if_offline()

  raw <- get_policy_measures(type = "tax", since = "2023-24")
  agg <- policy_measures_summary(raw)
  expect_s3_class(agg, "obr_tbl")
  expect_named(agg, c("type", "event", "fiscal_year", "value_mn"))
  expect_lte(nrow(agg), nrow(raw))
  # Same provenance fields preserved
  expect_equal(obr_provenance(agg)$publication, "PMD")
  # No duplicate (event, type, fiscal_year) rows
  key <- paste(agg$event, agg$type, agg$fiscal_year, sep = "|")
  expect_equal(anyDuplicated(key), 0L)
})

test_that("policy_measures_summary() rejects non-obr_tbl input", {
  expect_error(policy_measures_summary(data.frame(x = 1)),
               regexp = "obr_tbl")
})

# Catalogue of every detailed-forecast table the package can fetch from the
# EFO Aggregates and EFO Economy workbooks. Each row says where the table
# lives, what shape it has, and what units / metric type to default to.
#
# Layouts:
#   quarterly_wide   - cols: blank, period (e.g. "2008Q1"), one column per
#                      series. Headers may span multiple rows.
#   quarterly_single - cols: blank, period, single value. Used for one-series
#                      tables like Nominal GDP (1.4), output gap (1.14).
#   annual_year_wide - cols: blank, blank, year columns ("2020", "2021"...).
#                      Series names in col 2, values in cols 3+.
#   fiscal_year_wide - cols: blank, blank, fiscal-year columns ("2024-25"...).
#                      Series names in col 2 (flat or with "of which:" sub-
#                      indentation in col 3+). The parser flattens hierarchy
#                      so each row carries the full path as the series name.
#   cross_reference  - sheet exists but redirects to a previous EFO. Returns
#                      NULL (with an informative warning) rather than data.

efo_catalogue_table <- function() {
  # Hand-curated metadata. The unit / metric_type values follow OBR's own
  # sheet-title units; the parser may override on a per-row basis when a
  # series name strongly signals otherwise (e.g. "deflator", "Index").
  rows <- list(
    # ------------------------------------------------------------------
    # Economy workbook (sections 1.x)
    # ------------------------------------------------------------------
    list("1.1",  "economy", "GDP",        "GDP expenditure components (chain-linked volumes)",          "quarterly_wide",   "level",   "gbp_bn"),
    list("1.2",  "economy", "GDP",        "GDP expenditure components (current prices)",                "quarterly_wide",   "level",   "gbp_bn"),
    list("1.3",  "economy", "GDP",        "GDP income components",                                      "quarterly_wide",   "level",   "gbp_bn"),
    list("1.4",  "economy", "GDP",        "Nominal GDP (non-seasonally adjusted)",                      "quarterly_single", "level",   "gbp_bn"),
    list("1.5",  "economy", "GDP",        "Per capita (age 16+)",                                       "quarterly_wide",   "level",   NA_character_),
    list("1.6",  "economy", "Labour",     "Labour market",                                              "quarterly_wide",   NA,        NA_character_),
    list("1.7",  "economy", "Inflation",  "Inflation",                                                  "quarterly_wide",   "yoy_pct", "pct"),
    list("1.8",  "economy", "Trade",      "Balance of payments",                                        "quarterly_wide",   "level",   NA_character_),
    list("1.9",  "economy", "Markets",    "Market-derived assumptions (rates, FX, oil, equities)",      "quarterly_wide",   "level",   NA_character_),
    list("1.10", "economy", "Sectors",    "Financial balances by sector",                               "quarterly_wide",   "pct",     "pct"),
    list("1.11", "economy", "Households", "Household balance sheets and lending",                       "quarterly_wide",   "level",   NA_character_),
    list("1.11b","economy", "Households", "Household balance sheet (unsecured debt)",                   "quarterly_wide",   "level",   NA_character_),
    list("1.12", "economy", "Households", "Household disposable income",                                "quarterly_wide",   "level",   NA_character_),
    list("1.13", "economy", "Wages",      "National Minimum Wage and National Living Wage (annual)",    "annual_year_wide", "level",   "gbp_per_hour"),
    list("1.14", "economy", "Cycle",      "OBR central estimate of the output gap",                    "quarterly_single", "pct",     "pct"),
    list("1.15", "economy", "Cycle",      "Potential output forecast",                                  "quarterly_wide",   "level",   NA_character_),
    list("1.16", "economy", "Housing",    "Housing market",                                             "quarterly_wide",   "level",   NA_character_),
    list("1.17", "economy", "Households", "Household debt-servicing costs",                             "quarterly_wide",   "level",   NA_character_),
    list("1.18", "economy", "Housing",    "Eligible rent growth assumptions",                           "fiscal_year_wide",   "yoy_pct", "pct"),
    list("1.19", "economy", "Inflation",  "CPI category inflation",                                     "quarterly_wide",     "yoy_pct", "pct"),
    list("1.19b","economy", "Inflation",  "CPI category weights",                                       "annual_period_wide", "level",   "weight"),
    list("1.20", "economy", "Markets",    "Electricity price forecast",                                 "quarterly_single", "level",   "pence_per_mwh"),
    # ------------------------------------------------------------------
    # Aggregates workbook (sections 6.x)
    # ------------------------------------------------------------------
    list("6.1",  "aggregates", "Expenditure", "Breakdown of expenditure forecast by sector and economic category", "fiscal_year_wide", "level", "gbp_bn"),
    list("6.2",  "aggregates", "Receipts",    "Breakdown of receipts forecast by sector and economic category",    "fiscal_year_wide", "level", "gbp_bn"),
    list("6.3",  "aggregates", "Aggregates",  "General government transactions by economic category",              "fiscal_year_wide", "level", "gbp_bn"),
    list("6.4",  "aggregates", "Aggregates",  "Public sector transactions by sub-sector and economic category",    "subsector_matrix", "level", "gbp_bn"),
    list("6.5",  "aggregates", "Aggregates",  "Components of net borrowing",                                        "fiscal_year_wide", "level", "gbp_bn"),
    list("6.6",  "aggregates", "APF",         "Asset Purchase Facility annual runoff assumptions",                  "fiscal_year_wide", "level", "gbp_bn"),
    list("6.7",  "aggregates", "ONS",         "Inconsistencies between OBR forecasts and ONS outturns",             "fiscal_year_wide", "level", "gbp_bn"),
    list("6.8",  "aggregates", "Students",    "Student entrant borrowers forecast",                                 "fiscal_year_wide", "level", NA_character_),
    list("6.9",  "aggregates", "Students",    "Net flow of student loans and repayments",                           "fiscal_year_wide", "level", "gbp_bn"),
    list("6.10", "aggregates", "Students",    "Proxy for prevailing market rates (student loan interest cap)",      "quarterly_indented", "pct", "pct"),
    list("6.11", "aggregates", "Debt",        "Public sector net debt year-on-year changes",                        "cross_reference",  NA,      NA_character_),
    list("6.12", "aggregates", "Debt",        "Total gross financing",                                              "fiscal_year_wide", "level", "gbp_bn"),
    list("6.13", "aggregates", "Debt",        "Composition of public sector net debt",                              "fiscal_year_wide", "pct",   "pct"),
    list("6.14", "aggregates", "Debt",        "Composition of public sector net worth",                             "fiscal_year_wide", "level", "gbp_bn"),
    list("6.15", "aggregates", "Debt",        "Reconciliation of PSNCR and CGNCR",                                  "cross_reference",  NA,    NA_character_),
    list("6.16", "aggregates", "Debt",        "Central government debt interest by financing component",           "fiscal_year_wide", "level", "gbp_bn"),
    list("6.17", "aggregates", "Debt",        "Outstanding stocks, debt interest payments and effective rates",     "fiscal_year_wide", "level", "gbp_bn")
  )

  cols <- c("table_id", "file", "section", "title",
            "layout", "default_metric_type", "default_unit")
  out <- as.data.frame(do.call(rbind, lapply(rows, function(r) {
    stats::setNames(as.data.frame(r, stringsAsFactors = FALSE), cols)
  })))
  rownames(out) <- NULL
  out
}

#' List the EFO detailed-forecast tables this package can fetch
#'
#' Returns a data frame describing every Detailed Forecast Table in the OBR
#' Economic and Fiscal Outlook (Aggregates and Economy workbooks) that
#' [get_efo_table()] knows how to parse.
#'
#' @details
#' Use this catalogue to discover which tables are available, what each
#' contains, and the default `metric_type` / `unit` `get_efo_table()` will
#' attach. Pass any `table_id` to [get_efo_table()].
#'
#' Coverage: 17 fiscal aggregates tables (Section 6) plus 22 macro economy
#' tables (Section 1). One sheet (6.11) is currently a cross-reference to
#' a previous EFO and returns `NULL` with a warning rather than data.
#'
#' @return A data frame with columns:
#' \describe{
#'   \item{table_id}{The EFO table identifier (e.g. `"6.5"`, `"1.7"`).}
#'   \item{file}{Which EFO workbook the table sits in: `"aggregates"` or
#'     `"economy"`.}
#'   \item{section}{Theme tag (e.g. `"GDP"`, `"Labour"`, `"Debt"`).}
#'   \item{title}{Human-readable title taken from the OBR contents page.}
#'   \item{layout}{Layout family the parser uses: `"quarterly_wide"`,
#'     `"quarterly_single"`, `"annual_year_wide"`, `"fiscal_year_wide"`,
#'     or `"cross_reference"`.}
#'   \item{default_metric_type}{Default `metric_type` applied to series
#'     whose name does not signal otherwise. `NA` lets the heuristic decide
#'     per row.}
#'   \item{default_unit}{Default `unit` for the same.}
#' }
#'
#' @examples
#' head(obr_efo_catalogue())
#'
#' # All tables in the Debt section
#' cat <- obr_efo_catalogue()
#' cat[cat$section == "Debt", c("table_id", "title")]
#'
#' @family EFO
#' @export
obr_efo_catalogue <- function() {
  efo_catalogue_table()
}

# Internal: look up a single row of the catalogue by table id.
efo_catalogue_lookup <- function(table_id) {
  cat <- efo_catalogue_table()
  hit <- cat[cat$table_id == table_id, , drop = FALSE]
  if (nrow(hit) == 0L) {
    cli::cli_abort(c(
      "Unknown EFO table id {.val {table_id}}.",
      "i" = "Run {.fn obr_efo_catalogue} to see all known tables."
    ))
  }
  hit[1L, ]
}

# Historical Official Forecasts Database.
# OBR sometimes re-uses an older slug across vintages, so the package tries
# recent fiscal events first, then falls back to the known-stable slug.
FORECASTS_FALLBACK <- "https://obr.uk/download/historical-official-forecasts-database-march-2025/"
FORECASTS_FILENAME <- "historical_forecasts.xlsx"

forecasts_source <- function(refresh = FALSE) {
  obr_get_xlsx(
    candidates = forecasts_url_candidates(),
    fallback   = FORECASTS_FALLBACK,
    filename   = FORECASTS_FILENAME,
    refresh    = refresh,
    label      = "Historical Forecasts Database"
  )
}

# User-friendly series name -> primary Excel sheet name.
# The literal Pound symbol in some OBR sheet names is fragile; the tolerant
# resolver in resolve_sheet_name() retries with a regex if the primary name
# is missing.
SERIES_MAP <- c(
  "PSNB"            = "\u00a3PSNB",
  "PSNB_pct"        = "PSNB",
  "PSND"            = "PSND",
  "receipts"        = "\u00a3PSCR",
  "receipts_pct"    = "PSCR",
  "expenditure"     = "\u00a3TME",
  "expenditure_pct" = "TME",
  "GDP"             = "NGDP",
  "real_GDP"        = "UKGDP",
  "CPI"             = "CPI"
)

# Regex fallbacks for series whose primary sheet name uses a non-ASCII
# currency symbol (Pound). For ASCII-only sheet names, the primary lookup
# is enough and no fallback is needed.
SERIES_FALLBACK <- c(
  "PSNB"        = "^.PSNB$",
  "receipts"    = "^.PSCR$",
  "expenditure" = "^.TME$"
)

# v0.4.0 schema metadata for each series in SERIES_MAP. Used to tag every
# returned row with metric_type and unit so callers can rbind() across series
# and know what the value column means.
SERIES_META <- list(
  "PSNB"            = list(metric_type = "level",   unit = "gbp_bn"),
  "PSNB_pct"        = list(metric_type = "pct",     unit = "pct"),
  "PSND"            = list(metric_type = "pct",     unit = "pct"),
  "receipts"        = list(metric_type = "level",   unit = "gbp_bn"),
  "receipts_pct"    = list(metric_type = "pct",     unit = "pct"),
  "expenditure"     = list(metric_type = "level",   unit = "gbp_bn"),
  "expenditure_pct" = list(metric_type = "pct",     unit = "pct"),
  "GDP"             = list(metric_type = "yoy_pct", unit = "pct"),
  "real_GDP"        = list(metric_type = "yoy_pct", unit = "pct"),
  "CPI"             = list(metric_type = "yoy_pct", unit = "pct")
)

#' List available forecast series
#'
#' Returns a data frame showing the series names accepted by
#' [get_forecasts()], the corresponding Excel sheet in the OBR
#' Historical Official Forecasts Database, and a plain-English description.
#'
#' @return A data frame with columns `series`, `sheet`, and `description`.
#'
#' @examples
#' list_forecast_series()
#'
#' @family forecasts
#' @export
list_forecast_series <- function() {
  data.frame(
    series = names(SERIES_MAP),
    sheet  = unname(SERIES_MAP),
    description = c(
      "Public sector net borrowing (\u00a3bn)",
      "Public sector net borrowing (% of GDP)",
      "Public sector net debt (% of GDP)",
      "Public sector current receipts (\u00a3bn)",
      "Public sector current receipts (% of GDP)",
      "Total managed expenditure (\u00a3bn)",
      "Total managed expenditure (% of GDP)",
      "Nominal GDP growth (%)",
      "Real GDP growth (%)",
      "CPI inflation (%)"
    ),
    stringsAsFactors = FALSE
  )
}

#' Get OBR forecast history for a fiscal series
#'
#' Downloads (and caches) the OBR Historical Official Forecasts Database and
#' returns a tidy long-format data frame showing every forecast the OBR has
#' ever published for a given series. Each row is one forecast for one fiscal
#' year, made at one fiscal event.
#'
#' This is useful for visualising how OBR forecasts have evolved over time,
#' and for comparing forecasts against outturns. The vintage of the underlying
#' database is recorded in the returned object's provenance.
#'
#' @param series Character. The series to return. Use
#'   [list_forecast_series()] to see all options. Defaults to
#'   `"PSNB"` (Public Sector Net Borrowing in \enc{£}{GBP} billion).
#' @param refresh Logical. If `TRUE`, re-download even if a cached copy exists.
#'   Defaults to `FALSE`.
#'
#' @return An `obr_tbl` with the standard v0.4.0 schema plus a
#' `forecast_date` column (so each row identifies one fiscal-year value at
#' one fiscal event):
#' \describe{
#'   \item{forecast_date}{When the forecast was published, e.g. `"March 2024"`}
#'   \item{period}{Fiscal year being forecast, e.g. `"2024-25"`}
#'   \item{period_type}{Always `"fiscal_year"`}
#'   \item{series}{Series name as supplied (character)}
#'   \item{metric_type}{One of `"level"`, `"pct"`, `"yoy_pct"`, set per
#'     series. CPI / GDP / real_GDP are `"yoy_pct"` (rates of change);
#'     PSNB / receipts / expenditure are `"level"`; the `_pct` variants and
#'     PSND are `"pct"` (% of GDP).}
#'   \item{value}{Numeric forecast value in the unit described by `unit`}
#'   \item{unit}{`"gbp_bn"` for level series, `"pct"` for percentage series}
#' }
#'
#' @examples
#' \donttest{
#' op <- options(obr.cache_dir = tempdir())
#' try({
#'   get_forecasts("PSNB")
#'
#'   psnb <- get_forecasts("PSNB")
#'   psnb[psnb$period == "2024-25", ]
#' })
#' options(op)
#' }
#'
#' @family forecasts
#' @export
get_forecasts <- function(series = "PSNB", refresh = FALSE) {

  series <- match.arg(series, choices = names(SERIES_MAP))

  src <- forecasts_source(refresh)
  primary  <- SERIES_MAP[[series]]
  fallback <- if (series %in% names(SERIES_FALLBACK)) {
    SERIES_FALLBACK[[series]]
  } else {
    NULL
  }
  sheet_name <- resolve_sheet_name(src$path, primary, fallback)

  raw  <- readxl::read_excel(src$path, sheet = sheet_name,
                              col_names = FALSE, .name_repair = "minimal")

  fiscal_years   <- as.character(unlist(raw[4, 2:ncol(raw)]))
  forecast_dates <- as.character(unlist(raw[5:nrow(raw), 1]))
  data_matrix    <- as.data.frame(raw[5:nrow(raw), 2:ncol(raw)],
                                  stringsAsFactors = FALSE)

  valid <- grepl("^[A-Za-z]+ [0-9]{4}", forecast_dates)
  forecast_dates <- forecast_dates[valid]
  data_matrix    <- data_matrix[valid, ]

  n_dates <- length(forecast_dates)
  n_years <- length(fiscal_years)

  meta <- SERIES_META[[series]]

  result <- data.frame(
    forecast_date = rep(forecast_dates, times  = n_years),
    period        = rep(fiscal_years,   each   = n_dates),
    period_type   = "fiscal_year",
    series        = series,
    metric_type   = meta$metric_type,
    value         = suppressWarnings(as.numeric(unlist(data_matrix))),
    unit          = meta$unit,
    stringsAsFactors = FALSE
  )

  result <- result[!is.na(result$value), ]
  rownames(result) <- NULL

  new_obr_tbl(
    data        = result,
    publication = "HFD",
    vintage     = obr_url_vintage(src$url),
    source_url  = src$url,
    retrieved   = src$retrieved,
    file_md5    = src$file_md5
  )
}

test_that("list_forecast_series() works without network", {
  result <- list_forecast_series()

  expect_s3_class(result, "data.frame")
  expect_named(result, c("series", "sheet", "description"))
  expect_equal(nrow(result), 10)
  expect_true("PSNB" %in% result$series)
  expect_true("CPI"  %in% result$series)
})

test_that("get_forecasts() errors on invalid series", {
  expect_error(
    get_forecasts("NOT_A_SERIES"),
    regexp = "should be one of"
  )
})

# v0.4.0 schema for HFD output: forecast_date plus the standard tidy long cols.
hfd_cols <- c("forecast_date", "period", "period_type",
              "series", "metric_type", "value", "unit")

test_that("get_forecasts('PSNB') returns the v0.4.0 HFD schema", {
  skip_on_cran()
  skip_if_offline()

  result <- get_forecasts("PSNB")

  expect_s3_class(result, "data.frame")
  expect_named(result, hfd_cols)
  expect_type(result$series,        "character")
  expect_type(result$forecast_date, "character")
  expect_type(result$period,        "character")
  expect_type(result$value,         "double")
  expect_gt(nrow(result), 100)
  expect_true(all(result$series == "PSNB"))
  expect_true(all(result$period_type == "fiscal_year"))
  expect_true(all(result$metric_type == "level"))
  expect_true(all(result$unit == "gbp_bn"))
})

test_that("get_forecasts('PSNB') returns data for 2024-25", {
  skip_on_cran()
  skip_if_offline()

  result <- get_forecasts("PSNB")
  forecasts_2425 <- result[result$period == "2024-25", ]

  expect_gt(nrow(forecasts_2425), 5)
  expect_true(all(forecasts_2425$value > 0))
})

test_that("get_forecasts('CPI') is tagged as a yoy_pct rate", {
  skip_on_cran()
  skip_if_offline()

  result <- get_forecasts("CPI")

  expect_s3_class(result, "data.frame")
  expect_named(result, hfd_cols)
  expect_true(all(result$series == "CPI"))
  expect_gt(nrow(result), 50)
  expect_true(all(result$metric_type == "yoy_pct"))
  expect_true(all(result$unit == "pct"))
})

test_that("get_forecasts() returns obr_tbl with HFD provenance", {
  skip_on_cran()
  skip_if_offline()

  res  <- get_forecasts("PSNB")
  expect_s3_class(res, "obr_tbl")
  prov <- obr_provenance(res)
  expect_equal(prov$publication, "HFD")
  expect_match(prov$source_url, "historical-official-forecasts-database")
  expect_match(prov$vintage, "^[A-Z][a-z]+ [0-9]{4}$")
})

---
title: "Tracking fiscal policy measures since 1970"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tracking fiscal policy measures since 1970}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>"
)
```

The OBR's Policy Measures Database (PMD) lists every individual tax measure scored at a UK fiscal event since 1970, and every spending measure since 2010, with the Exchequer effect in GBP million for each year of the forecast horizon. It is the dataset behind virtually every "Chancellor's Budget worth GBP X" headline in the financial press.

The OBR distributes the PMD as a single Excel workbook with a non-trivial layout. `obr` is the only programmatic wrapper for it that exists.

The chunks below that download data are shown as code only; for a quick offline preview, run any chunk in an R session with internet access.

## Pulling the database

```{r, eval = FALSE}
library(obr)

pmd <- get_policy_measures()      # tax + spending, all years
obr_provenance(pmd)$vintage       # which PMD vintage was downloaded
```

Returned columns are `type`, `event`, `measure`, `head`, `fiscal_year`, `value_mn`. Positive values are revenue-raising for tax measures and spending-increasing for spending measures.

## Filtering by fiscal event

`type`, `search`, and `since` filter at the source so you don't have to subset a 100k-row data frame manually.

```{r, eval = FALSE}
# Every tax measure scored at the October 2024 Budget,
# ordered by 2025-26 effect
oct24 <- get_policy_measures(type = "tax", since = "2025-26")
oct24 <- oct24[grepl("October 2024", oct24$event) &
               oct24$fiscal_year == "2025-26", ]
oct24 <- oct24[order(-oct24$value_mn), ]
head(oct24[, c("measure", "head", "value_mn")])
```

## Searching by theme

`search` is a regular expression (case-insensitive) applied to both the measure description and the Treasury head.

```{r, eval = FALSE}
# Every alcohol-duty measure since 2010
alc <- get_policy_measures(type = "tax", search = "alcohol", since = "2010-11")
unique(alc$event)
```

This lets you build thematic time series for empirical work: e.g. cumulative tax raised from energy measures over a Parliament, or the share of personal-tax revenue from threshold freezes.

## Aggregating to the Budget scorecard

`policy_measures_summary()` collapses the long format to net Exchequer effect by event and fiscal year. This is what fits in an IFS Green Budget chart.

```{r, eval = FALSE}
pm  <- get_policy_measures(type = "tax", since = "2024-25")
agg <- policy_measures_summary(pm)
agg[agg$event == "Budget October 2024", ]
```

Provenance is preserved through aggregation: the returned `obr_tbl` still records which PMD vintage produced the numbers.

```{r, eval = FALSE}
obr_provenance(agg)$vintage
```

## Worked example: contribution of each fiscal event

For attribution work ("which Budget did most of the lifting?"), aggregate by event for a fixed target year.

```{r, eval = FALSE}
pm   <- get_policy_measures(type = "tax")
parl <- pm[pm$fiscal_year == "2027-28" &
           grepl("(October 2024|2025|2026)", pm$event), ]
agg  <- policy_measures_summary(parl)
agg[order(agg$event), c("event", "value_mn")]
```

The pattern - filter once, summarise once - mirrors how analysts read the OBR's own scorecard tables in the EFO.

## Cross-vintage comparison

Because each PMD release re-prints the full history with any reclassifications, you can compare measure-level numbers across PMD vintages by downloading two and joining on `event` + `measure`. (At the time of writing, the OBR re-uses the `march-2025` slug across vintages for the PMD; the package's URL resolver finds whichever file is currently live and records its MD5 fingerprint, so cross-vintage comparison is hash-verifiable.)

---
title: "Pinning to a specific OBR publication"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Pinning to a specific OBR publication}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>"
)
NOT_CRAN <- identical(Sys.getenv("NOT_CRAN"), "true")
```

The OBR revises every forecast at every Budget and Statement. A number that is correct in March can be wrong in October. For research, audit, and reproducibility, you need a way to say "this analysis used the October 2024 forecast" and have the package reliably return the same numbers six months later.

`obr` calls this the **vintage layer**. This vignette walks through the four ways to use it. The chunks that do not require a network connection are run inline; the chunks that download from the OBR are shown as code only.

## The vintage table

```{r}
library(obr)

efos <- obr_efo_vintages()
nrow(efos)
head(efos, 3)
tail(efos, 3)
```

`obr_efo_vintages()` is offline: the publication calendar is hardcoded into the package and refreshed at each release.

## Resolving an EFO by date

```{r}
obr_as_of("2024-12-15")
obr_as_of("2010-05-01") |> tryCatch(error = function(e) conditionMessage(e))
```

The first call returns the most recent EFO published on or before 15 December 2024. The second errors because no EFO existed before June 2010.

## Pinning a session-wide vintage

```{r}
obr_pin("October 2024")
obr_pinned()
obr_unpin()
obr_pinned()
```

When pinned, the EFO functions default to that vintage. `obr_unpin()` clears the pin and returns to the dynamic resolver.

## Downloading a specific vintage

These chunks require a network connection and are shown as code only.

```{r, eval = FALSE}
# Pull two vintages of the same fiscal table to compare forecast revisions
oct24 <- get_efo_fiscal(vintage = "October 2024")
mar26 <- get_efo_fiscal(vintage = "March 2026")

# Net borrowing forecast for 2027-28 from each vintage. v0.4.0 schema
# uses `period` for the time column (with `period_type = "fiscal_year"`).
oct24[oct24$series == "Net borrowing" & oct24$period == "2027-28", ]
mar26[mar26$series == "Net borrowing" & mar26$period == "2027-28", ]
```

Cached files are vintage-tagged, so different vintages do not overwrite each other in the cache. The `vintage = ...` argument always overrides any pin set with `obr_pin()`.

## What's recorded as provenance

Every `obr_tbl` carries enough metadata for an OBR analyst to audit which publication produced any number.

```{r, eval = FALSE}
obr_provenance(get_efo_fiscal(vintage = "October 2024"))
# $publication: "EFO"
# $vintage:     "October 2024"
# $source_url:  https://obr.uk/download/october-2024-economic-and-fiscal-outlook-...
# $retrieved:   timestamp of download
# $file_md5:    MD5 fingerprint of the underlying spreadsheet
# $package_version: obr version that produced the object
```

The MD5 fingerprint lets two analyses verify they ran against an identical underlying file, even if downloaded weeks apart.

## How vintage interacts with `get_forecasts()`

The Historical Forecasts Database is a single workbook covering every OBR vintage in one file, so it does not take a `vintage =` argument. To filter to a specific forecast date, subset the long-format result.

```{r, eval = FALSE}
fc  <- get_forecasts("PSNB")
fc[fc$forecast_date == "October 2024", ]
```

For reproducibility, the HFD itself is itself versioned: the OBR releases a new HFD at each fiscal event, so the underlying file MD5 changes over time even if the older forecasts within it do not. `obr_provenance()` records the HFD vintage on every result.

---
title: "Working with EFO forecasts"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with EFO forecasts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>"
)
```

The OBR's *Economic and Fiscal Outlook* contains the official UK fiscal forecast: five-year projections for borrowing, debt, receipts, expenditure, and the underlying economy. `obr` exposes this through `get_efo_fiscal()` (Table 6.5 aggregates) and `get_efo_economy()` (sheets 1.6 labour, 1.7 inflation, 1.14 output gap).

This vignette covers what changed in the v0.4.0 schema and how to use it.

## The standard schema

From v0.4.0, every data-fetching function in `obr` returns the same six columns:

| Column        | Type      | Values                                                     |
|---------------|-----------|------------------------------------------------------------|
| `period`      | character | `"2024-25"`, `"2025Q1"`, `"2025"` (depends on `period_type`)|
| `period_type` | character | `"fiscal_year"`, `"quarter"`, `"calendar_year"`            |
| `series`      | character | The variable name, e.g. `"CPI"`, `"Net borrowing"`         |
| `metric_type` | character | `"level"`, `"yoy_pct"`, `"index"`, `"pct"`, `"pct_pts"`    |
| `value`       | double    | Numeric value                                              |
| `unit`        | character | `"gbp_bn"`, `"pct"`, `"pct_pts"`, `"index"`, `"count_k"`, etc. |

`get_forecasts()` adds `forecast_date` as a leading column.

The schema is consistent enough that you can `rbind()` outputs from different functions and still know what each value means.

## Index vs YoY: the v0.4.0 fix

Before v0.4.0, `get_efo_economy("inflation")` returned CPI / RPI / etc. as series with values in a single `value` column, with no machine-readable indication of whether a row was a level (index points) or a year-on-year growth rate (per cent). The `metric_type` and `unit` columns now make the distinction explicit.

```{r, eval = FALSE}
library(obr)

inf <- get_efo_economy("inflation")
table(inf$metric_type, inf$unit)
#         index  pct
#   index   372    0
#   yoy_pct   0 1844
```

To get just the year-on-year inflation rates:

```{r, eval = FALSE}
inf_yoy <- inf[inf$metric_type == "yoy_pct", ]
head(inf_yoy[inf_yoy$series == "CPI", ])
```

To get just any index series (e.g. GDP deflator if the source publishes it as an index):

```{r, eval = FALSE}
inf_idx <- inf[inf$metric_type == "index", ]
```

The classifier defaults bare names like "CPI" / "RPI" to `yoy_pct` because the OBR inflation sheet reports them as growth rates by convention. Series matching `Index`, `(2015=100)`, or similar literal patterns are tagged as `index` instead. You can override per-row by post-processing the returned frame.

## Combining EFO with outturn from PFD

Because the schema is uniform across publications, you can compare a fiscal forecast (EFO) against the realised outturn (PFD) without column mapping.

```{r, eval = FALSE}
forecast <- get_efo_fiscal()                # 5-year forecast, gbp_bn
forecast <- forecast[forecast$series == "Net borrowing", ]

outturn <- get_psnb()                       # historical outturn, series = "PSNB", gbp_bn
outturn <- outturn[outturn$period >= "2020-21", ]

# Both have period (fiscal_year), value (gbp_bn), unit. Stack them.
combined <- rbind(
  data.frame(source = "outturn",  outturn[,  c("period", "value", "unit")]),
  data.frame(source = "forecast", forecast[, c("period", "value", "unit")])
)
```

## Comparing two vintages

Pin a vintage explicitly to compare how the OBR's view changed between fiscal events.

```{r, eval = FALSE}
oct24 <- get_efo_fiscal(vintage = "October 2024")
mar26 <- get_efo_fiscal(vintage = "March 2026")

# Net borrowing forecast for 2027-28 from each vintage
oct24[oct24$series == "Net borrowing" & oct24$period == "2027-28", "value"]
mar26[mar26$series == "Net borrowing" & mar26$period == "2027-28", "value"]
```

See `vignette("vintages")` for the full vintage layer.

## What `unit` covers

| Unit code | Meaning                                            | Typical series                          |
|-----------|----------------------------------------------------|-----------------------------------------|
| `gbp_bn`  | Pounds sterling, billions                          | PSNB, PSND, TME, receipts forecasts     |
| `pct`     | Percentage (rate, share, growth rate)              | Inflation, unemployment rate, debt/GDP  |
| `pct_pts` | Percentage points                                  | Output gap (pp of potential)            |
| `index`   | Index level (typically rebased to 100 at a year)   | CPI Index, GDP deflator (when level)    |
| `count_k` | Count, thousands                                   | Incapacity benefit claimants            |

`unit` is plain character. It does not enforce arithmetic safety: if you sum a `gbp_bn` value with a `pct` value, R will compute it without complaint. The column is there to remind you and to allow programmatic filtering.

## Provenance

Every returned object carries provenance metadata describing which OBR publication and vintage produced it.

```{r, eval = FALSE}
obr_provenance(get_efo_fiscal())
# $publication: "EFO"
# $vintage:     "March 2026"
# $source_url:  ...
# $retrieved:   timestamp
# $file_md5:    fingerprint of the underlying spreadsheet
# $package_version: obr version
```

This lets you audit which OBR publication produced any number in your analysis. See `vignette("vintages")`.

---
title: "Anatomy of a forecast revision"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Anatomy of a forecast revision}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>"
)
```

Every EFO revises the previous forecast. The OBR's *Forecast Evaluation Report* and IFS Green Budget routinely decompose those revisions into three components:

* **Policy** - measures announced at the fiscal event itself.
* **Classifications and one-offs** - reclassifications by ONS, valuation adjustments, court rulings.
* **Underlying** - everything else, mostly changes to the macroeconomic determinants (real GDP, inflation, employment).

The OBR publishes this decomposition as the **Forecast Revisions Database (FRD)**. `obr` exposes it through `get_forecast_revisions()`. This vignette walks through how to read it and how it pairs with the wide forecast panel for forecast-evaluation work.

The chunks that download data are shown as code only.

## The decomposition

```{r, eval = FALSE}
library(obr)

rev <- get_forecast_revisions(unit = "gbp_bn")
unique(rev$component)
# "Total"
# "Policy"
# "of which: receipts"
# "of which: spending"
# "Classifications and one-offs"
# "Underlying"
# "of which: receipts"
# "of which: debt interest"
# "of which: non-interest spending"
```

`Total` for a given (`forecast_date`, `fiscal_year`) cell equals the sum of `Policy` + `Classifications and one-offs` + `Underlying` (up to rounding). The `of which:` rows further decompose Policy and Underlying.

## Top-level revisions only

For the standard "what changed?" chart, filter to top-level components.

```{r, eval = FALSE}
top <- rev[rev$component %in% c(
  "Total", "Policy",
  "Classifications and one-offs", "Underlying"
), ]

# Most recent vintage's revision for each fiscal year
latest <- max(top$forecast_date)
top[top$forecast_date == latest, ]
```

Plot this with `ggplot2`'s `geom_col(position = "stack")` and you reproduce the OBR's "drivers of revision" chart.

## Per cent of GDP

Pass `unit = "pct_gdp"` for the same decomposition expressed as a share of GDP. This is the right scaling for comparing revisions across fiscal events with materially different nominal-GDP levels.

```{r, eval = FALSE}
rev_pct <- get_forecast_revisions(unit = "pct_gdp")
```

## Pairing with the forecast panel

For forecast-evaluation work it helps to see the levels alongside the revisions. `obr_forecast_panel()` pivots the Historical Forecasts Database from long to wide: rows are forecast vintages (chronologically ordered), columns are fiscal years.

```{r, eval = FALSE}
panel <- obr_forecast_panel("PSNB")
# Every PSNB forecast for 2027-28
panel[, c("forecast_date", "2027-28")]
```

Each row of the panel is one EFO; each column is a fiscal year. Reading down a column gives you the time-series of forecasts the OBR made for the same target year - exactly the input to a Diebold-Mariano test or a forecast-error decomposition.

## Worked example: PSNB 2024-25 across the full vintage history

```{r, eval = FALSE}
panel <- obr_forecast_panel("PSNB")

# Forecast value for FY 2024-25 across every vintage
psnb_2425 <- panel[, c("forecast_date", "2024-25")]
psnb_2425 <- psnb_2425[!is.na(psnb_2425[["2024-25"]]), ]
psnb_2425
```

For each vintage that forecast 2024-25 (March 2022 onward, when 2024-25 entered the five-year horizon), you get one row. The first vintage forecast PSNB at around GBP 37bn; by November 2025, the last forecast before the outturn arrives, the projection had reached around GBP 150bn. The FRD attributes most of that drift to the underlying component (weaker tax base) with a smaller contribution from policy.

## Provenance for cited results

For a paper or report, cite the vintage and MD5 of the underlying file so readers can reproduce.

```{r, eval = FALSE}
prov <- obr_provenance(rev)
sprintf("OBR Forecast Revisions Database, vintage %s. File MD5: %s. Retrieved %s.",
        prov$vintage,
        prov$file_md5,
        format(prov$retrieved, "%Y-%m-%d"))
```

Two researchers running the same code six months apart on the same vintage will see different `retrieved` timestamps but identical `file_md5` values - which is the audit-trail property OBR analysts care about.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/public_finances.R
\name{get_psnd}
\alias{get_psnd}
\title{Get Public Sector Net Debt (PSND)}
\usage{
get_psnd(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy exists.
Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema. \code{series} is
\code{"PSND"}, \code{metric_type} is \code{"level"}, \code{unit} is \code{"gbp_bn"}. See
\code{\link[=get_public_finances]{get_public_finances()}} for column definitions.
}
\description{
Downloads (and caches) the OBR Public Finances Databank and returns
annual Public Sector Net Debt in \enc{£}{GBP} billion.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  psnd <- get_psnd()
  tail(psnd)
})
options(op)
}

}
\seealso{
Other public finances: 
\code{\link{get_expenditure}()},
\code{\link{get_psnb}()},
\code{\link{get_public_finances}()},
\code{\link{get_receipts}()}
}
\concept{public finances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monthly_profiles.R
\name{get_monthly_profiles}
\alias{get_monthly_profiles}
\title{Get the OBR monthly profiles for the public finances}
\usage{
get_monthly_profiles(
  sheet = c("profiles", "cgncr"),
  vintage = NULL,
  refresh = FALSE
)
}
\arguments{
\item{sheet}{Which profile table to return. \code{"profiles"} (the default)
returns the receipts and spending profiles; \code{"cgncr"} returns the
central government net cash requirement breakdown.}

\item{vintage}{Optional EFO vintage label such as \code{"March 2026"}. If
\code{NULL} (the default), uses any pin set via \code{\link[=obr_pin]{obr_pin()}} or resolves the
latest live profiles workbook.}

\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard schema columns (\code{period},
\code{period_type}, \code{series}, \code{metric_type}, \code{value}, \code{unit}). Monthly rows
have \code{period_type = "month"} and \code{period} in \code{"YYYY-MM"} format; each
series also carries one \code{period_type = "fiscal_year"} row holding the
full-year EFO forecast the profile sums to. All values are GBP billion.
}
\description{
Downloads (and caches) the monthly profiles workbook the OBR publishes
alongside each \emph{Economic and Fiscal Outlook}. The profiles apportion
the full-year EFO forecast for receipts, spending, and the central
government net cash requirement (CGNCR) across the twelve months of the
fiscal year, so that each month's ONS/HMT public sector finances outturn
can be judged against the path implied by the OBR's forecast.
}
\details{
This is the reference point used every month in the run-up to a fiscal
event ("borrowing so far this year vs the OBR profile"). The OBR itself
publishes a monthly commentary against these profiles; this function
provides the underlying numbers in tidy long format. Pair with monthly
outturn data (e.g. from the ONS public sector finances release) to
compute in-year deviations from profile.

The profiles workbook is typically published a few weeks after the EFO
itself. The OBR describes the profiles as broad-brush and illustrative;
see the Notes sheet of the source workbook.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())

mp <- tryCatch(get_monthly_profiles(), error = function(e) NULL)
if (!is.null(mp)) {
  # Monthly profile for HMRC cash receipts
  mp[mp$series == "HMRC cash receipts" & mp$period_type == "month", ]
}

# CGNCR breakdown by month
cg <- tryCatch(get_monthly_profiles("cgncr"), error = function(e) NULL)

options(op)
}

}
\seealso{
Other EFO: 
\code{\link{get_efo_economy}()},
\code{\link{get_efo_fiscal}()},
\code{\link{get_efo_table}()},
\code{\link{list_efo_economy_measures}()},
\code{\link{obr_efo_catalogue}()}
}
\concept{EFO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efo.R
\name{list_efo_economy_measures}
\alias{list_efo_economy_measures}
\title{List available EFO economy measures}
\usage{
list_efo_economy_measures()
}
\value{
A data frame with columns \code{measure}, \code{sheet}, and \code{description}.
}
\description{
Returns a data frame of the economy measures available via
\code{\link[=get_efo_economy]{get_efo_economy()}}, showing the \code{measure} name to pass and a
short description of what each covers.
}
\examples{
list_efo_economy_measures()

}
\seealso{
Other EFO: 
\code{\link{get_efo_economy}()},
\code{\link{get_efo_fiscal}()},
\code{\link{get_efo_table}()},
\code{\link{get_monthly_profiles}()},
\code{\link{obr_efo_catalogue}()}
}
\concept{EFO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{as.data.frame.obr_tbl}
\alias{as.data.frame.obr_tbl}
\title{Coerce an obr_tbl to a plain data.frame}
\usage{
\method{as.data.frame}{obr_tbl}(x, ...)
}
\arguments{
\item{x}{An \code{obr_tbl}.}

\item{...}{Unused.}
}
\value{
A plain \code{data.frame} with no provenance.
}
\description{
Strips the provenance attributes and the \code{obr_tbl} class so the result
interacts smoothly with packages that dispatch on \code{data.frame} directly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/welfare.R
\name{get_welfare_spending}
\alias{get_welfare_spending}
\title{Get working-age welfare spending}
\usage{
get_welfare_spending(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema (columns:
\code{period}, \code{period_type}, \code{series}, \code{metric_type}, \code{value}, \code{unit}).
Values are spending as a percentage of GDP; \code{metric_type} is \code{"pct"},
\code{unit} is \code{"pct"}.
}
\description{
Downloads (and caches) the OBR Welfare Trends Report charts and tables
workbook and returns annual working-age welfare spending as a share of
GDP, split into incapacity-related and non-incapacity spending.
}
\details{
Data cover fiscal years from 1978-79 through the current forecast horizon.
The exact vintage is recorded in the returned object's provenance.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  welfare <- get_welfare_spending()
  welfare[welfare$series == "Working-age incapacity benefits spending" &
  welfare$period >= "2000-01", ]
})
options(op)
}

}
\seealso{
Other welfare: 
\code{\link{get_incapacity_caseloads}()},
\code{\link{get_incapacity_spending}()}
}
\concept{welfare}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vintages.R
\name{obr_efo_vintages}
\alias{obr_efo_vintages}
\title{List known OBR Economic and Fiscal Outlook vintages}
\usage{
obr_efo_vintages()
}
\value{
A data frame with columns:
\describe{
\item{publication}{Always \code{"EFO"}.}
\item{vintage}{Vintage label, e.g. \code{"March 2026"}.}
\item{date}{Publication date of the EFO (Date).}
\item{slug}{URL slug used by the OBR for that vintage's download pages.}
}
}
\description{
Returns a data frame of every EFO published since the OBR was created in
June 2010, with publication dates and the URL slug used in the OBR's
download links. Use this to look up which vintages can be pinned via
\code{\link[=obr_pin]{obr_pin()}} or passed to \verb{vintage =} arguments on \code{\link[=get_efo_fiscal]{get_efo_fiscal()}} and
\code{\link[=get_efo_economy]{get_efo_economy()}}.
}
\examples{
head(obr_efo_vintages())
tail(obr_efo_vintages(), 5)

}
\seealso{
Other vintage: 
\code{\link{obr_as_of}()},
\code{\link{obr_pin}()},
\code{\link{obr_pinned}()},
\code{\link{obr_unpin}()}
}
\concept{vintage}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_measures.R
\name{policy_measures_summary}
\alias{policy_measures_summary}
\title{Summarise policy measures by fiscal event}
\usage{
policy_measures_summary(x)
}
\arguments{
\item{x}{An \code{obr_tbl} returned by \code{\link[=get_policy_measures]{get_policy_measures()}}.}
}
\value{
An \code{obr_tbl} with columns:
\describe{
\item{type}{\code{"tax"} or \code{"spending"}}
\item{event}{Fiscal event}
\item{fiscal_year}{Fiscal year}
\item{value_mn}{Sum of the Exchequer effect across all measures
scored at that event, in GBP million}
}
Provenance is preserved.
}
\description{
Aggregates the measures returned by \code{\link[=get_policy_measures]{get_policy_measures()}} to give the
net Exchequer effect (positive = revenue-raising / spending-reducing for
tax, spending-increasing for spending) by fiscal event and fiscal year.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  pm <- get_policy_measures(type = "tax", since = "2024-25")
  policy_measures_summary(pm)
})
options(op)
}

}
\seealso{
Other policy measures: 
\code{\link{get_policy_measures}()}
}
\concept{policy measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fiscal_sustainability.R
\name{get_pension_projections}
\alias{get_pension_projections}
\title{Get long-run state pension spending projections (deprecated)}
\usage{
get_pension_projections(refresh = FALSE)
}
\arguments{
\item{refresh}{Ignored. Retained so existing calls do not error.}
}
\value{
\code{NULL}, invisibly.
}
\description{
\strong{Deprecated since obr 0.5.1.}

In July 2026 the OBR restructured the Fiscal Risks and Sustainability
Report workbook set. The executive-summary sheet this function read
(\code{C1.2}, state pension spending split into demographic and triple-lock
scenarios) no longer exists, so the function could no longer return data.

The equivalent series is now published in the FSR Chapter 3 "Long-term
spending projections" workbook, as Chart 3.11 "State pension spending
under alternative uprating assumptions", with a different scenario
structure (triple-lock, CPI, and average-earnings uprating rather than the
old demographic vs triple-lock split). See
\url{https://obr.uk/frs/fiscal-risks-and-sustainability-july-2026/}.

This stub is retained so existing scripts do not error. It emits a
deprecation warning and returns \code{NULL}. It will be removed in a future
release.
}
\examples{
# Deprecated since 0.5.1: emits a warning and returns NULL.
suppressWarnings(get_pension_projections())

}
\concept{long-term fiscal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{clear_cache}
\alias{clear_cache}
\title{Clear cached OBR files}
\usage{
clear_cache()
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Deletes all files downloaded and cached by the obr package. The next
function call will re-download fresh data from the OBR website.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
clear_cache()
options(op)
}

}
\concept{data access}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecast_revisions.R
\name{get_forecast_revisions}
\alias{get_forecast_revisions}
\title{Get OBR forecast revisions}
\usage{
get_forecast_revisions(unit = c("gbp_bn", "pct_gdp"), refresh = FALSE)
}
\arguments{
\item{unit}{Character. Either \code{"gbp_bn"} (GBP billion, the default) or
\code{"pct_gdp"} (per cent of GDP).}

\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with columns:
\describe{
\item{forecast_date}{Forecast vintage, e.g. \code{"November 2024"} (character).}
\item{component}{Revision component. Top-level rows are \code{"Total"},
\code{"Policy"}, \code{"Classifications and one-offs"}, and \code{"Underlying"}.
Sub-components carry the OBR's \code{"of which: ..."} labels (character).}
\item{fiscal_year}{Fiscal year being revised, e.g. \code{"2024-25"} (character).}
\item{value}{Revision value, in GBP billion or per cent of GDP per
\code{unit} (numeric). A positive value indicates an upward revision to
PSNB.}
}
}
\description{
Downloads (and caches) the OBR Forecast Revisions Database, which
decomposes each EFO-to-EFO revision in the headline Public Sector Net
Borrowing forecast into three top-level components - policy,
classifications and one-offs, and underlying (economic determinants) -
with sub-components for each.
}
\details{
This is the dataset behind the "what changed?" attribution charts in
OBR fiscal commentary and IFS Green Budget chapters.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  rev <- get_forecast_revisions()
  # Top-level revisions only
  rev[rev$component \%in\% c("Total", "Policy",
  "Classifications and one-offs", "Underlying"), ]
})
options(op)
}

}
\seealso{
Other forecasts: 
\code{\link{get_forecasts}()},
\code{\link{list_forecast_series}()},
\code{\link{obr_actual_vs_forecast}()},
\code{\link{obr_compare_vintages}()},
\code{\link{obr_forecast_panel}()}
}
\concept{forecasts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efo.R
\name{get_efo_economy}
\alias{get_efo_economy}
\title{Get EFO economy projections}
\usage{
get_efo_economy(
  measure = c("inflation", "labour", "output_gap"),
  refresh = FALSE,
  vintage = NULL
)
}
\arguments{
\item{measure}{Character. Which economy table to return. One of
\code{"inflation"}, \code{"labour"}, or \code{"output_gap"}. Defaults to
\code{"inflation"}.}

\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}

\item{vintage}{Optional EFO vintage label such as \code{"October 2024"}. If
supplied, the function downloads the file for that specific EFO. If
\code{NULL} (the default), the function uses any vintage set via \code{\link[=obr_pin]{obr_pin()}},
or falls back to the latest live EFO via the dynamic URL resolver.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema (columns:
\code{period}, \code{period_type}, \code{series}, \code{metric_type}, \code{value}, \code{unit}):
\describe{
\item{period}{Calendar quarter, e.g. \code{"2025Q1"} (character)}
\item{period_type}{Always \code{"quarter"} for this function (character)}
\item{series}{Variable name, e.g. \code{"CPI"} (character)}
\item{metric_type}{One of \code{"index"}, \code{"yoy_pct"}, \code{"pct"}, \code{"level"},
classified from the series name. This is the v0.4.0 fix for the v0.3.x
issue where, e.g., CPI Index values and CPI YoY values shared a single
\code{value} column with no machine-readable distinction.}
\item{value}{Numeric value in units described by \code{unit}}
\item{unit}{One of \code{"index"}, \code{"pct"}, etc., paired with \code{metric_type}}
}
}
\description{
Downloads (and caches) the OBR \emph{Economic and Fiscal Outlook} Detailed
Forecast Tables - Economy file and returns quarterly economic projections
for a chosen measure in tidy long format.
}
\details{
Data run from 2008 Q1 through the current forecast horizon. Use
\code{\link[=list_efo_economy_measures]{list_efo_economy_measures()}} to see all available measures.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  inf <- get_efo_economy("inflation")
  inf[inf$series == "CPI", ]

  lab <- get_efo_economy("labour")

  # Compare CPI projections from two different EFOs
  inf_oct24 <- get_efo_economy("inflation", vintage = "October 2024")
  inf_mar26 <- get_efo_economy("inflation", vintage = "March 2026")
})
options(op)
}

}
\seealso{
Other EFO: 
\code{\link{get_efo_fiscal}()},
\code{\link{get_efo_table}()},
\code{\link{get_monthly_profiles}()},
\code{\link{list_efo_economy_measures}()},
\code{\link{obr_efo_catalogue}()}
}
\concept{EFO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/public_finances.R
\name{get_receipts}
\alias{get_receipts}
\title{Get public sector receipts by tax type}
\usage{
get_receipts(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy exists.
Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema (columns:
\code{period}, \code{period_type}, \code{series}, \code{metric_type}, \code{value}, \code{unit}).
\code{series} is the tax or receipt category, \code{metric_type} is \code{"level"},
\code{unit} is \code{"gbp_bn"}. See \code{\link[=get_public_finances]{get_public_finances()}} for full column docs.
}
\description{
Downloads (and caches) the OBR Public Finances Databank and returns
public sector current receipts broken down by individual tax type,
in tidy long format. Coverage begins in 1999-00.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  receipts <- get_receipts()
  receipts[grepl("income tax", receipts$series, ignore.case = TRUE), ]
})
options(op)
}

}
\seealso{
Other public finances: 
\code{\link{get_expenditure}()},
\code{\link{get_psnb}()},
\code{\link{get_psnd}()},
\code{\link{get_public_finances}()}
}
\concept{public finances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/public_finances.R
\name{get_expenditure}
\alias{get_expenditure}
\title{Get Total Managed Expenditure}
\usage{
get_expenditure(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy exists.
Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema. \code{series} is
\code{"TME"}, \code{metric_type} is \code{"level"}, \code{unit} is \code{"gbp_bn"}. See
\code{\link[=get_public_finances]{get_public_finances()}} for column definitions.
}
\description{
Downloads (and caches) the OBR Public Finances Databank and returns
annual Total Managed Expenditure (TME) in \enc{£}{GBP} billion. TME is
the broadest measure of UK government spending.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  tme <- get_expenditure()
  tail(tme)
})
options(op)
}

}
\seealso{
Other public finances: 
\code{\link{get_psnb}()},
\code{\link{get_psnd}()},
\code{\link{get_public_finances}()},
\code{\link{get_receipts}()}
}
\concept{public finances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{print.obr_tbl}
\alias{print.obr_tbl}
\title{Print an obr_tbl}
\usage{
\method{print}{obr_tbl}(x, n = 10L, ...)
}
\arguments{
\item{x}{An \code{obr_tbl} object.}

\item{n}{Number of rows to display (default 10).}

\item{...}{Further arguments passed to \code{\link[=print.data.frame]{print.data.frame()}}.}
}
\value{
The input, returned invisibly.
}
\description{
Prints the data with a provenance header that names the OBR publication,
vintage, source URL, retrieval time, and underlying file fingerprint.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/welfare.R
\name{get_incapacity_spending}
\alias{get_incapacity_spending}
\title{Get incapacity benefits spending by type}
\usage{
get_incapacity_spending(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema. \code{series} is the
benefit name, values are spending as a percentage of GDP, \code{metric_type}
is \code{"pct"}, \code{unit} is \code{"pct"}. See \code{\link[=get_public_finances]{get_public_finances()}} for full
column docs.
}
\description{
Downloads (and caches) the OBR Welfare Trends Report charts and tables
workbook and returns annual spending on each incapacity benefit as a
share of GDP, from 1978-79 to the current forecast horizon.
}
\details{
Series include: Invalidity Benefit, Incapacity Benefit, Employment and
Support Allowance (ESA), Sickness Benefit, and Severe Disablement
Allowance.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  ib <- get_incapacity_spending()
  unique(ib$series)
})
options(op)
}

}
\seealso{
Other welfare: 
\code{\link{get_incapacity_caseloads}()},
\code{\link{get_welfare_spending}()}
}
\concept{welfare}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/policy_measures.R
\name{get_policy_measures}
\alias{get_policy_measures}
\title{Get OBR policy measures by fiscal event}
\usage{
get_policy_measures(
  type = c("tax", "spending"),
  search = NULL,
  since = NULL,
  refresh = FALSE
)
}
\arguments{
\item{type}{Character vector. Which measures to return: \code{"tax"},
\code{"spending"}, or both (the default).}

\item{search}{Optional character. A regular expression used to filter
\code{measure} and \code{head} (case-insensitive). For example,
\code{search = "alcohol"} keeps any measure whose description or head
mentions alcohol.}

\item{since}{Optional fiscal-year cut-off (e.g. \code{"2020-21"}). Rows whose
\code{fiscal_year} is earlier are dropped.}

\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with columns:
\describe{
\item{event}{Fiscal event that scored the measure, e.g. \code{"Budget 2024"} (character)}
\item{measure}{Plain-English description of the measure (character)}
\item{head}{Tax head (for tax measures) or spending head (character)}
\item{fiscal_year}{Fiscal year being scored, e.g. \code{"2024-25"} (character)}
\item{value_mn}{Exchequer effect in GBP million. For tax measures, a
positive value raises revenue; for spending measures, a positive
value increases spending (numeric)}
}
}
\description{
Downloads (and caches) the OBR Policy Measures Database, a single workbook
that lists every tax or spending measure scored at a UK fiscal event,
with the Exchequer effect in GBP million for each year of the forecast
horizon. Tax measures cover 1970 to date; spending measures cover 2010
to date.
}
\details{
This is the only programmatic access to the PMD: the OBR otherwise
distributes it as a single Excel file.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())

# The OBR CDN occasionally rate-limits or returns 403 to repeated
# requests. tryCatch so the example degrades gracefully if that happens
# during the CRAN check.
oct24 <- tryCatch(get_policy_measures(),
                  error = function(e) NULL)
if (!is.null(oct24)) {
  head(oct24[grepl("2024", oct24$event) &
             oct24$fiscal_year == "2025-26", ])

  # All alcohol-duty measures since 2010
  tryCatch(
    get_policy_measures(type = "tax", search = "alcohol",
                        since = "2010-11"),
    error = function(e) NULL
  )
}

options(op)
}

}
\seealso{
Other policy measures: 
\code{\link{policy_measures_summary}()}
}
\concept{policy measures}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/public_finances.R
\name{get_public_finances}
\alias{get_public_finances}
\title{Get all Public Finances Databank aggregates}
\usage{
get_public_finances(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy exists.
Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} (a \code{data.frame} with attached provenance) with the
standard v0.4.0 schema (columns: \code{period}, \code{period_type}, \code{series},
\code{metric_type}, \code{value}, \code{unit}):
\describe{
\item{period}{Fiscal year (character, e.g. \code{"2024-25"})}
\item{period_type}{Always \code{"fiscal_year"} for this function}
\item{series}{Series name (character)}
\item{metric_type}{Usually \code{"level"}; ratio or index series get a more
specific value derived from the series name}
\item{value}{Numeric value in units described by \code{unit}}
\item{unit}{Usually \code{"gbp_bn"}; ratios and indices override}
}
Use \code{\link[=obr_provenance]{obr_provenance()}} to extract source URL, vintage, and retrieval time.
}
\description{
Downloads (and caches) the OBR Public Finances Databank and returns all
aggregate fiscal series in tidy long format. Covers outturn from 1946-47
and OBR projections through the current forecast horizon.
}
\details{
Series include: Public sector net borrowing, Public sector net debt, Total
managed expenditure, Public sector current receipts, Nominal GDP, GDP
deflator, and more.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  pf <- get_public_finances()
  unique(pf$series)
  obr_provenance(pf)
})
options(op)
}

}
\seealso{
Other public finances: 
\code{\link{get_expenditure}()},
\code{\link{get_psnb}()},
\code{\link{get_psnd}()},
\code{\link{get_receipts}()}
}
\concept{public finances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/headroom.R
\name{obr_headroom}
\alias{obr_headroom}
\title{Compute the stability-rule margin from the EFO forecast}
\usage{
obr_headroom(vintage = NULL, target_year = NULL, refresh = FALSE)
}
\arguments{
\item{vintage}{Optional EFO vintage label such as \code{"March 2026"}. If
\code{NULL} (the default), uses any pin set via \code{\link[=obr_pin]{obr_pin()}} or the latest
live EFO.}

\item{target_year}{Optional fiscal-year string (e.g. \code{"2029-30"}). If
supplied, an \code{is_target_year} column flags that year.}

\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard schema columns (\code{period},
\code{period_type}, \code{series}, \code{metric_type}, \code{value}, \code{unit}), where \code{series}
is \code{"Current budget surplus"} and \code{value} is in GBP billion (positive =
surplus = headroom under the stability rule). If \code{target_year} is
supplied, an additional logical \code{is_target_year} column is included.
}
\description{
Returns the current budget surplus for every year of an EFO forecast,
derived from Table 6.5 (Components of net borrowing). Under the Charter
for Budget Responsibility's stability rule the current budget must be in
balance or surplus by the target year, so the value in the target year
is the margin, or "headroom", against the rule: positive means the rule
is met with room to spare, negative means it is missed.
}
\details{
The value returned is \verb{-1 *} the published "Current budget deficit"
series, i.e. a surplus is positive. Pass \code{target_year} to flag the year
the rule currently bites on; the function does not guess it, because the
Charter's target-year convention changes over time (see
\code{\link[=obr_fiscal_rules]{obr_fiscal_rules()}} for the rules as encoded at release).

Note the OBR's own published headroom figure at a fiscal event can
differ slightly from the Table 6.5 arithmetic (rounding, and any
rule-specific adjustments described in the EFO text). Treat this as the
published forecast path for the rule metric, not a reproduction of the
OBR's press-notice headroom number.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())

hr <- tryCatch(obr_headroom(), error = function(e) NULL)
if (!is.null(hr)) hr

# Flag the target year and read off the margin
hr <- tryCatch(obr_headroom(target_year = "2029-30"),
               error = function(e) NULL)
if (!is.null(hr)) hr[hr$is_target_year, ]

options(op)
}

}
\seealso{
Other fiscal rules: 
\code{\link{obr_fiscal_rules}()}
}
\concept{fiscal rules}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efo_dispatcher.R
\name{get_efo_table}
\alias{get_efo_table}
\title{Get any EFO detailed-forecast table by id}
\usage{
get_efo_table(table_id, vintage = NULL, refresh = FALSE)
}
\arguments{
\item{table_id}{Character. The EFO table identifier, e.g. \code{"6.5"},
\code{"1.7"}, \code{"6.13"}. See \code{\link[=obr_efo_catalogue]{obr_efo_catalogue()}} for the full list.}

\item{vintage}{Optional EFO vintage label (e.g. \code{"October 2024"}). If
\code{NULL}, uses any pin set via \code{\link[=obr_pin]{obr_pin()}} or falls back to the latest
live EFO.}

\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema columns
(\code{period}, \code{period_type}, \code{series}, \code{metric_type}, \code{value}, \code{unit}).
Returns \code{NULL} (with a warning) for cross-reference sheets.
}
\description{
Generic fetcher that takes an EFO table identifier (e.g. \code{"6.5"},
\code{"1.7"}, \code{"4.1"}) and returns the parsed contents in the standard
v0.4.0 schema. Use \code{\link[=obr_efo_catalogue]{obr_efo_catalogue()}} to discover which tables
are available.
}
\details{
Internally, this function looks up \code{table_id} in the catalogue, fetches
the right workbook (Aggregates or Economy), dispatches to a layout-
specific parser, and tags every row with \code{metric_type} and \code{unit}
according to the catalogue's defaults plus per-row classification.

Coverage today: 17 fiscal Aggregates tables + 22 macro Economy tables.
One sheet (6.11 PSND year-on-year changes) is a cross-reference to a
previous EFO and returns \code{NULL} with a warning rather than data; OBR
itself directs users to the previous EFO for that table.

Headline functions \code{\link[=get_efo_fiscal]{get_efo_fiscal()}} and \code{\link[=get_efo_economy]{get_efo_economy()}} are kept
as thin wrappers over this dispatcher.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({

  # Net borrowing components (same data as get_efo_fiscal())
  get_efo_table("6.5")

  # CPI category inflation by year
  get_efo_table("1.19")

  # Composition of public sector net debt
  get_efo_table("6.13")

  # Pin to a specific vintage
  get_efo_table("6.5", vintage = "October 2024")
})
options(op)
}

}
\seealso{
Other EFO: 
\code{\link{get_efo_economy}()},
\code{\link{get_efo_fiscal}()},
\code{\link{get_monthly_profiles}()},
\code{\link{list_efo_economy_measures}()},
\code{\link{obr_efo_catalogue}()}
}
\concept{EFO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vintages.R
\name{obr_unpin}
\alias{obr_unpin}
\title{Clear the pinned OBR EFO vintage}
\usage{
obr_unpin()
}
\value{
Invisibly returns the previously pinned vintage (or \code{NULL} if no
pin was set).
}
\description{
Removes any pin set by \code{\link[=obr_pin]{obr_pin()}}. After unpinning, \code{\link[=get_efo_fiscal]{get_efo_fiscal()}}
and \code{\link[=get_efo_economy]{get_efo_economy()}} revert to resolving the most recent live EFO via
the dynamic URL resolver.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  obr_pin("March 2025")
  obr_unpin()
})
options(op)
}

}
\seealso{
Other vintage: 
\code{\link{obr_as_of}()},
\code{\link{obr_efo_vintages}()},
\code{\link{obr_pin}()},
\code{\link{obr_pinned}()}
}
\concept{vintage}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{obr_actual_vs_forecast}
\alias{obr_actual_vs_forecast}
\title{Pair OBR forecasts with PFD outturn}
\usage{
obr_actual_vs_forecast(
  series = c("PSNB", "PSND", "expenditure"),
  refresh = FALSE
)
}
\arguments{
\item{series}{Forecast series. One of \code{"PSNB"}, \code{"PSND"}, \code{"expenditure"}.
Defaults to \code{"PSNB"}.}

\item{refresh}{Logical. If \code{TRUE}, re-download underlying files.}
}
\value{
An \code{obr_tbl} with columns \code{forecast_date}, \code{period} (fiscal year
being forecast), \code{period_type}, \code{series}, \code{unit}, \code{value_forecast}
(from HFD), \code{value_actual} (from PFD outturn), and \code{error}
(\code{value_forecast - value_actual}). Provenance points at the HFD;
the PFD source URL is recorded in \code{notes}.
}
\description{
Joins the long-format Historical Forecasts Database for a given series
against the Public Finances Databank outturn for the same series.
Returns one row per (forecast vintage, fiscal year) where both an OBR
forecast value and an ONS outturn value exist, with the forecast error
(\code{value_forecast - value_actual}) computed.
}
\details{
Useful for forecast-evaluation studies, similar in shape to the OBR's
own Forecast Evaluation Report decomposition.

Currently supports series for which a clean \code{gbp_bn} outturn function
exists in this package: \code{"PSNB"}, \code{"PSND"}, \code{"expenditure"}. Other
series (CPI, GDP, percentages of GDP) need outturn from external
packages and will error.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  eval <- obr_actual_vs_forecast("PSNB")

  # 1-year-ahead forecast errors only:
  # take the forecast vintage closest to the start of each fiscal year
  eval2425 <- eval[eval$period == "2024-25", ]
  eval2425[order(eval2425$forecast_date), ]
})
options(op)
}

}
\seealso{
Other forecasts: 
\code{\link{get_forecast_revisions}()},
\code{\link{get_forecasts}()},
\code{\link{list_forecast_series}()},
\code{\link{obr_compare_vintages}()},
\code{\link{obr_forecast_panel}()}
}
\concept{forecasts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vintages.R
\name{obr_pin}
\alias{obr_pin}
\title{Pin a session-wide OBR EFO vintage}
\usage{
obr_pin(vintage = NULL)
}
\arguments{
\item{vintage}{Vintage label such as \code{"October 2024"}. See
\code{\link[=obr_efo_vintages]{obr_efo_vintages()}} for the full list. A well-formed label that is not
yet in the package's EFO calendar (e.g. a brand-new EFO published after
this package version was released) is accepted with a warning: download
URLs are then constructed from the OBR's slug convention, so a new EFO
can be pinned on publication day. If \code{NULL}, this function clears
the pin (equivalent to calling \code{\link[=obr_unpin]{obr_unpin()}}).}
}
\value{
Invisibly returns the pinned vintage string, or \code{NULL} after
clearing.
}
\description{
Sets the EFO vintage that \code{\link[=get_efo_fiscal]{get_efo_fiscal()}} and \code{\link[=get_efo_economy]{get_efo_economy()}} will
use when called without an explicit \verb{vintage =} argument. The pin is stored
as the option \code{obr.efo_vintage} and lasts for the R session unless
overwritten or removed via \code{\link[=obr_unpin]{obr_unpin()}}.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  obr_pin("October 2024")
  obr_pinned()
  obr_unpin()
})
options(op)
}

}
\seealso{
Other vintage: 
\code{\link{obr_as_of}()},
\code{\link{obr_efo_vintages}()},
\code{\link{obr_pinned}()},
\code{\link{obr_unpin}()}
}
\concept{vintage}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efo.R
\name{get_efo_fiscal}
\alias{get_efo_fiscal}
\title{Get EFO fiscal projections (net borrowing components)}
\usage{
get_efo_fiscal(refresh = FALSE, vintage = NULL)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}

\item{vintage}{Optional EFO vintage label such as \code{"October 2024"}. If
supplied, the function downloads the file for that specific EFO. If
\code{NULL} (the default), the function uses any vintage set via \code{\link[=obr_pin]{obr_pin()}},
or falls back to the latest live EFO via the dynamic URL resolver. See
\code{\link[=obr_efo_vintages]{obr_efo_vintages()}} for the full list of supported vintages.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema (columns:
\code{period}, \code{period_type}, \code{series}, \code{metric_type}, \code{value}, \code{unit}):
\describe{
\item{period}{Fiscal year being forecast, e.g. \code{"2025-26"} (character)}
\item{period_type}{Always \code{"fiscal_year"} for this function (character)}
\item{series}{Component name, e.g. \code{"Net borrowing"} (character)}
\item{metric_type}{Always \code{"level"} for this function (character)}
\item{value}{Projected value (numeric)}
\item{unit}{Always \code{"gbp_bn"} for this function (character)}
}
}
\description{
Downloads (and caches) the OBR \emph{Economic and Fiscal Outlook} Detailed
Forecast Tables - Aggregates file and returns the components of net
borrowing (Table 6.5) in tidy long format.
}
\details{
Covers the five-year forecast horizon published at the most recent
fiscal event. Key series include current receipts, current expenditure,
depreciation, net investment, and net borrowing (PSNB). The exact
vintage is recorded in the returned object's provenance attribute and
visible in the printed header.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  efo <- get_efo_fiscal()
  efo[efo$series == "Net borrowing", ]
  obr_provenance(efo)$vintage

  # Pin to a specific EFO for reproducibility
  october_2024 <- get_efo_fiscal(vintage = "October 2024")
})
options(op)
}

}
\seealso{
Other EFO: 
\code{\link{get_efo_economy}()},
\code{\link{get_efo_table}()},
\code{\link{get_monthly_profiles}()},
\code{\link{list_efo_economy_measures}()},
\code{\link{obr_efo_catalogue}()}
}
\concept{EFO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecast_panel.R
\name{obr_forecast_panel}
\alias{obr_forecast_panel}
\title{Build a wide real-time panel of OBR forecasts}
\usage{
obr_forecast_panel(
  series = "PSNB",
  refresh = FALSE,
  order_chronologically = TRUE
)
}
\arguments{
\item{series}{Character. Series to return; see \code{\link[=list_forecast_series]{list_forecast_series()}}.
Defaults to \code{"PSNB"}.}

\item{refresh}{Logical. If \code{TRUE}, re-download the underlying database.}

\item{order_chronologically}{Logical. If \code{TRUE} (the default), rows are
ordered from earliest to latest forecast vintage. If \code{FALSE}, rows
appear in the order returned by \code{\link[=get_forecasts]{get_forecasts()}}.}
}
\value{
An \code{obr_tbl} whose first column (\code{forecast_date}) is character
and whose remaining columns are numeric forecast values, one per
fiscal year. Provenance is inherited from the underlying call to
\code{\link[=get_forecasts]{get_forecasts()}} and a \code{notes} field describes the panel.
}
\description{
Takes the long-format Historical Forecasts Database output of
\code{\link[=get_forecasts]{get_forecasts()}} and pivots it to a wide panel with one row per
forecast vintage (e.g. \code{"March 2024"}) and one column per fiscal year
being forecast (e.g. \code{"2024-25"}). The first column is the forecast
vintage; remaining columns are numeric forecast values.
}
\details{
This format mirrors how the OBR's own Historical Forecasts Database is
laid out and how forecast-error studies (e.g. the OBR's own Forecast
Evaluation Report) decompose performance: each row is one forecast
published at a fiscal event, each column is the target year being
forecast, and the diagonal-like structure lets you read the h-step
ahead forecast for any vintage.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  panel <- obr_forecast_panel("PSNB")
  # PSNB forecast for 2024-25 across every vintage
  panel[, c("forecast_date", "2024-25")]
})
options(op)
}

}
\seealso{
Other forecasts: 
\code{\link{get_forecast_revisions}()},
\code{\link{get_forecasts}()},
\code{\link{list_forecast_series}()},
\code{\link{obr_actual_vs_forecast}()},
\code{\link{obr_compare_vintages}()}
}
\concept{forecasts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{obr_compare_vintages}
\alias{obr_compare_vintages}
\title{Compare two EFO vintages}
\usage{
obr_compare_vintages(vintage_a, vintage_b, what = "fiscal", refresh = FALSE)
}
\arguments{
\item{vintage_a, vintage_b}{EFO vintage labels (e.g. \code{"October 2024"},
\code{"March 2026"}). Use \code{\link[=obr_efo_vintages]{obr_efo_vintages()}} to see all valid labels.}

\item{what}{Which EFO table to compare. Either one of the named
shortcuts \code{"fiscal"} (Table 6.5, the default), \code{"inflation"} (sheet
1.7), \code{"labour"} (sheet 1.6), \code{"output_gap"} (sheet 1.14), or any
table id from \code{\link[=obr_efo_catalogue]{obr_efo_catalogue()}} (e.g. \code{"6.13"}, \code{"1.19"}), so all
detailed-forecast tables can be diffed across vintages.}

\item{refresh}{Logical. If \code{TRUE}, re-download even if cached files
exist. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema columns
(\code{period}, \code{period_type}, \code{series}, \code{metric_type}, \code{unit}) plus
\code{value_a}, \code{value_b}, and \code{revision} (\code{value_b - value_a}).
Provenance points at the second vintage; the first vintage URL is
recorded in the \code{notes} field.
}
\description{
Pulls the same EFO table from two vintages and returns a tidy diff with
a revision column (\code{value_b - value_a}). Useful for quantifying how the
OBR's view changed between fiscal events.
}
\details{
Rows are the \strong{inner join} of the two vintages on the schema keys
(\code{period}, \code{period_type}, \code{series}, \code{metric_type}, \code{unit}). Periods or
series that are present in only one vintage are silently dropped. If
you need to see what was added or removed between vintages, compare
\code{obr_efo_vintages()} row counts or call the underlying functions
directly with each vintage and \code{setdiff()} on the keys.

Calling the function with \code{vintage_a == vintage_b} is allowed and
returns an all-zero \code{revision} column. There is no special handling
beyond that.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  diff <- obr_compare_vintages("October 2024", "March 2026")
  diff[diff$series == "Net borrowing", ]

  # Compare the inflation forecast across two vintages
  inf_diff <- obr_compare_vintages("October 2024", "March 2026",
  what = "inflation")

  # Any catalogue table works too, e.g. debt interest (Table 6.16)
  di_diff <- obr_compare_vintages("November 2025", "March 2026",
  what = "6.16")
})
options(op)
}

}
\seealso{
Other forecasts: 
\code{\link{get_forecast_revisions}()},
\code{\link{get_forecasts}()},
\code{\link{list_forecast_series}()},
\code{\link{obr_actual_vs_forecast}()},
\code{\link{obr_forecast_panel}()}
}
\concept{forecasts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{[.obr_tbl}
\alias{[.obr_tbl}
\title{Subset an obr_tbl, preserving provenance}
\usage{
\method{[}{obr_tbl}(x, ...)
}
\arguments{
\item{x}{An \code{obr_tbl}.}

\item{...}{Subsetting arguments passed to [\verb{[.data.frame}].}
}
\value{
An \code{obr_tbl} if the subset is two-dimensional, otherwise the
underlying vector.
}
\description{
Standard data.frame subsetting drops the \code{obr_tbl} class. This method
preserves it (and the provenance attributes) when the result is still a
data.frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vintages.R
\name{obr_pinned}
\alias{obr_pinned}
\title{Show the currently pinned OBR EFO vintage}
\usage{
obr_pinned()
}
\value{
Either a length-one character vector or \code{NULL}.
}
\description{
Returns the vintage string currently active via \code{\link[=obr_pin]{obr_pin()}}, or \code{NULL}
if no pin is set.
}
\examples{
obr_pinned()

}
\seealso{
Other vintage: 
\code{\link{obr_as_of}()},
\code{\link{obr_efo_vintages}()},
\code{\link{obr_pin}()},
\code{\link{obr_unpin}()}
}
\concept{vintage}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecasts.R
\name{get_forecasts}
\alias{get_forecasts}
\title{Get OBR forecast history for a fiscal series}
\usage{
get_forecasts(series = "PSNB", refresh = FALSE)
}
\arguments{
\item{series}{Character. The series to return. Use
\code{\link[=list_forecast_series]{list_forecast_series()}} to see all options. Defaults to
\code{"PSNB"} (Public Sector Net Borrowing in \enc{£}{GBP} billion).}

\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy exists.
Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema plus a
\code{forecast_date} column (so each row identifies one fiscal-year value at
one fiscal event):
\describe{
\item{forecast_date}{When the forecast was published, e.g. \code{"March 2024"}}
\item{period}{Fiscal year being forecast, e.g. \code{"2024-25"}}
\item{period_type}{Always \code{"fiscal_year"}}
\item{series}{Series name as supplied (character)}
\item{metric_type}{One of \code{"level"}, \code{"pct"}, \code{"yoy_pct"}, set per
series. CPI / GDP / real_GDP are \code{"yoy_pct"} (rates of change);
PSNB / receipts / expenditure are \code{"level"}; the \verb{_pct} variants and
PSND are \code{"pct"} (\% of GDP).}
\item{value}{Numeric forecast value in the unit described by \code{unit}}
\item{unit}{\code{"gbp_bn"} for level series, \code{"pct"} for percentage series}
}
}
\description{
Downloads (and caches) the OBR Historical Official Forecasts Database and
returns a tidy long-format data frame showing every forecast the OBR has
ever published for a given series. Each row is one forecast for one fiscal
year, made at one fiscal event.
}
\details{
This is useful for visualising how OBR forecasts have evolved over time,
and for comparing forecasts against outturns. The vintage of the underlying
database is recorded in the returned object's provenance.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  get_forecasts("PSNB")

  psnb <- get_forecasts("PSNB")
  psnb[psnb$period == "2024-25", ]
})
options(op)
}

}
\seealso{
Other forecasts: 
\code{\link{get_forecast_revisions}()},
\code{\link{list_forecast_series}()},
\code{\link{obr_actual_vs_forecast}()},
\code{\link{obr_compare_vintages}()},
\code{\link{obr_forecast_panel}()}
}
\concept{forecasts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{obr_provenance}
\alias{obr_provenance}
\title{Extract OBR provenance metadata}
\usage{
obr_provenance(x)
}
\arguments{
\item{x}{An \code{obr_tbl} (or any object; returns \code{NULL} if none attached).}
}
\value{
A named list of provenance fields, or \code{NULL} if no provenance is
attached. Fields:
\describe{
\item{publication}{Short code: \code{"PFD"}, \code{"HFD"}, \code{"EFO"}, \code{"WTR"}, \code{"FSR"}, \code{"PMD"}.}
\item{vintage}{Publication vintage label, e.g. \code{"March 2026"}.}
\item{source_url}{Canonical OBR download URL the data came from.}
\item{retrieved}{\code{POSIXct} timestamp of when the file was downloaded or
last validated in the cache.}
\item{file_md5}{MD5 fingerprint of the underlying spreadsheet.}
\item{package_version}{\code{obr} package version that produced the object.}
\item{notes}{Optional free-text notes, or \code{NULL}.}
}
}
\description{
Returns the provenance list attached to an \code{obr_tbl}: the OBR publication
it came from, the publication vintage, the source URL, retrieval time,
file fingerprint, and package version.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  psnb <- get_psnb()
  obr_provenance(psnb)
})
options(op)
}

}
\concept{provenance}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efo_catalogue.R
\name{obr_efo_catalogue}
\alias{obr_efo_catalogue}
\title{List the EFO detailed-forecast tables this package can fetch}
\usage{
obr_efo_catalogue()
}
\value{
A data frame with columns:
\describe{
\item{table_id}{The EFO table identifier (e.g. \code{"6.5"}, \code{"1.7"}).}
\item{file}{Which EFO workbook the table sits in: \code{"aggregates"} or
\code{"economy"}.}
\item{section}{Theme tag (e.g. \code{"GDP"}, \code{"Labour"}, \code{"Debt"}).}
\item{title}{Human-readable title taken from the OBR contents page.}
\item{layout}{Layout family the parser uses: \code{"quarterly_wide"},
\code{"quarterly_single"}, \code{"annual_year_wide"}, \code{"fiscal_year_wide"},
or \code{"cross_reference"}.}
\item{default_metric_type}{Default \code{metric_type} applied to series
whose name does not signal otherwise. \code{NA} lets the heuristic decide
per row.}
\item{default_unit}{Default \code{unit} for the same.}
}
}
\description{
Returns a data frame describing every Detailed Forecast Table in the OBR
Economic and Fiscal Outlook (Aggregates and Economy workbooks) that
\code{\link[=get_efo_table]{get_efo_table()}} knows how to parse.
}
\details{
Use this catalogue to discover which tables are available, what each
contains, and the default \code{metric_type} / \code{unit} \code{get_efo_table()} will
attach. Pass any \code{table_id} to \code{\link[=get_efo_table]{get_efo_table()}}.

Coverage: 17 fiscal aggregates tables (Section 6) plus 22 macro economy
tables (Section 1). One sheet (6.11) is currently a cross-reference to
a previous EFO and returns \code{NULL} with a warning rather than data.
}
\examples{
head(obr_efo_catalogue())

# All tables in the Debt section
cat <- obr_efo_catalogue()
cat[cat$section == "Debt", c("table_id", "title")]

}
\seealso{
Other EFO: 
\code{\link{get_efo_economy}()},
\code{\link{get_efo_fiscal}()},
\code{\link{get_efo_table}()},
\code{\link{get_monthly_profiles}()},
\code{\link{list_efo_economy_measures}()}
}
\concept{EFO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/obr-package.R
\docType{package}
\name{obr-package}
\alias{obr}
\alias{obr-package}
\title{obr: Access 'Office for Budget Responsibility' Data}
\description{
Provides clean, tidy access to data published by the 'Office for Budget Responsibility' ('OBR'), the UK's independent fiscal watchdog. Covers the Public Finances Databank (outturn for PSNB, PSND, receipts, and expenditure since 1946), the Historical Official Forecasts Database (every 'OBR' forecast since 2010), the Economic and Fiscal Outlook detailed forecast tables (five-year projections from the latest Budget), the monthly profiles for the public finances (the 'OBR' forecast apportioned across the months of the fiscal year), and the Welfare Trends Report (incapacity benefit spending and caseloads). All returned objects carry provenance metadata recording the source URL, publication vintage, retrieval time, and file fingerprint, so analyses can be audited and reproduced. Data is downloaded from the 'OBR' on first use and cached locally for subsequent calls. Data is sourced from the 'OBR' website \url{https://obr.uk}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://charlescoverdale.github.io/obr/}
  \item \url{https://github.com/charlescoverdale/obr}
  \item Report bugs at \url{https://github.com/charlescoverdale/obr/issues}
}

}
\author{
\strong{Maintainer}: Charles Coverdale \email{charlesfcoverdale@gmail.com}

}
\concept{OBR}
\concept{Office for Budget Responsibility}
\concept{PSNB}
\concept{UK budget}
\concept{economic forecasts}
\concept{fiscal policy}
\concept{national debt}
\concept{public finances}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vintages.R
\name{obr_as_of}
\alias{obr_as_of}
\title{Find the OBR publication that was current on a given date}
\usage{
obr_as_of(date, publication = "EFO")
}
\arguments{
\item{date}{A Date, or anything coercible to one (e.g. \code{"2024-11-15"}).}

\item{publication}{Currently only \code{"EFO"} is supported.}
}
\value{
A length-one character vector containing the vintage label,
e.g. \code{"October 2024"}.
}
\description{
Returns the most recent EFO that had been published on or before \code{date}.
Useful for reproducing analyses as they would have looked at a given point
in time, before subsequent forecast revisions.
}
\examples{
obr_as_of("2024-11-15")
obr_as_of(Sys.Date())

}
\seealso{
Other vintage: 
\code{\link{obr_efo_vintages}()},
\code{\link{obr_pin}()},
\code{\link{obr_pinned}()},
\code{\link{obr_unpin}()}
}
\concept{vintage}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecasts.R
\name{list_forecast_series}
\alias{list_forecast_series}
\title{List available forecast series}
\usage{
list_forecast_series()
}
\value{
A data frame with columns \code{series}, \code{sheet}, and \code{description}.
}
\description{
Returns a data frame showing the series names accepted by
\code{\link[=get_forecasts]{get_forecasts()}}, the corresponding Excel sheet in the OBR
Historical Official Forecasts Database, and a plain-English description.
}
\examples{
list_forecast_series()

}
\seealso{
Other forecasts: 
\code{\link{get_forecast_revisions}()},
\code{\link{get_forecasts}()},
\code{\link{obr_actual_vs_forecast}()},
\code{\link{obr_compare_vintages}()},
\code{\link{obr_forecast_panel}()}
}
\concept{forecasts}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/welfare.R
\name{get_incapacity_caseloads}
\alias{get_incapacity_caseloads}
\title{Get incapacity benefit caseloads}
\usage{
get_incapacity_caseloads(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy
exists. Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema. The two series
(\code{"Claimants"} and \code{"Share of working age population"}) carry different
units: claimants are in thousands and the share is a percentage. After
calling, the caller may want to overwrite \code{unit} to \code{"count_k"} for the
claimants series, since the heuristic classifier cannot infer the
"thousands" denomination from the series name alone.
}
\description{
Downloads (and caches) the OBR Welfare Trends Report charts and tables
workbook and returns the combined incapacity benefit caseload since
2008-09, in both absolute terms (thousands of claimants) and as a share
of the working-age population.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  cases <- get_incapacity_caseloads()
  cases[cases$series == "Claimants", ]
})
options(op)
}

}
\seealso{
Other welfare: 
\code{\link{get_incapacity_spending}()},
\code{\link{get_welfare_spending}()}
}
\concept{welfare}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fiscal_rules.R
\name{obr_fiscal_rules}
\alias{obr_fiscal_rules}
\title{Get the current UK fiscal rules}
\usage{
obr_fiscal_rules()
}
\value{
A data frame with columns:
\describe{
\item{rule}{Short rule code: \code{"stability"}, \code{"investment"}, \code{"welfare_cap"}.}
\item{metric}{The metric the rule binds on.}
\item{target_description}{Plain-English target.}
\item{target_year_rule}{How the target year is set under the Charter.}
\item{direction_of_pass}{Sign convention for headroom.}
\item{source_charter}{Charter version that defines the rule.}
\item{source_act}{Act of Parliament backing the rule.}
}
}
\description{
Returns a data frame describing the Charter for Budget Responsibility
fiscal rules currently in force, as encoded in the package at the time
of release. Numerical headroom against each rule is \emph{not} shipped as a
constant because it is updated at every fiscal event and would go stale
within months; use \code{\link[=obr_headroom]{obr_headroom()}} to derive the stability-rule margin
from the current EFO, or consult the OBR's EFO press release for the
relevant vintage.
}
\details{
The current Charter (Autumn 2024, with an Autumn 2025 update) sets three
numerical rules:
\itemize{
\item \strong{Stability rule}: current budget in balance or surplus by
the target year. The target year is the 5th forecast year and rolls
to the 3rd forecast year from 2026-27 onwards.
\item \strong{Investment rule}: Public Sector Net Financial Liabilities
(PSNFL) as a share of GDP falling year-on-year by the target year.
\item \strong{Welfare cap}: AME-capped welfare spending below the
statutory cap level by the target year.
}
The Budget Responsibility Act 2024 also adds a non-numerical
\strong{fiscal lock} requiring an OBR forecast for any fiscally
significant measure.
}
\examples{
obr_fiscal_rules()

}
\seealso{
Other fiscal rules: 
\code{\link{obr_headroom}()}
}
\concept{fiscal rules}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{summary.obr_tbl}
\alias{summary.obr_tbl}
\title{Summary of an obr_tbl}
\usage{
\method{summary}{obr_tbl}(object, ...)
}
\arguments{
\item{object}{An \code{obr_tbl} object.}

\item{...}{Unused.}
}
\value{
Invisibly returns the provenance list.
}
\description{
Returns the full provenance card and a structural summary of the data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/public_finances.R
\name{get_psnb}
\alias{get_psnb}
\title{Get Public Sector Net Borrowing (PSNB)}
\usage{
get_psnb(refresh = FALSE)
}
\arguments{
\item{refresh}{Logical. If \code{TRUE}, re-download even if a cached copy exists.
Defaults to \code{FALSE}.}
}
\value{
An \code{obr_tbl} with the standard v0.4.0 schema. \code{series} is
\code{"PSNB"}, \code{metric_type} is \code{"level"}, \code{unit} is \code{"gbp_bn"}. See
\code{\link[=get_public_finances]{get_public_finances()}} for column definitions.
}
\description{
Downloads (and caches) the OBR Public Finances Databank and returns
annual Public Sector Net Borrowing in \enc{£}{GBP} billion. A positive
value means the government is borrowing (deficit); a negative value means
a surplus.
}
\examples{
\donttest{
op <- options(obr.cache_dir = tempdir())
try({
  psnb <- get_psnb()
  tail(psnb)
})
options(op)
}

}
\seealso{
Other public finances: 
\code{\link{get_expenditure}()},
\code{\link{get_psnd}()},
\code{\link{get_public_finances}()},
\code{\link{get_receipts}()}
}
\concept{public finances}
